#include "LeicaCamera.h"
#include "utilities.h"

#include <vcl_limits.h>
#include <vcl_cstdio.h>
#include <vcl_fstream.h>

#include <vnl/vnl_math.h>
#include <vnl/algo/vnl_svd.h>

#include <vul/vul_file_iterator.h>
#include <vul/vul_file.h>

//*****************************************************************************

LeicaCamera::LeicaCamera()
{
  name = "";
  width = 0;  height = 0;
  rows = 0;   columns = 0;
  R.set_identity();  t.fill(0);  
  f = 1;   Su = 1;  Du = 1;  Dv = 1;
  u0 = 0;  v0 = 0;  k1 = 0;  k2 = 0;
  p1 = 0;  p2 = 0;
}

//*****************************************************************************

LeicaCamera::LeicaCamera(const LeicaCamera &cam) : Camera(cam)
{
  rows = cam.rows;   columns = cam.columns;   
  f = cam.f;    Su = cam.Su;  k1 = cam.k1;  k2 = cam.k2; 
  p1 = cam.p1;  p2 = cam.p2;  rk1 = cam.rk1;  rk2 = cam.rk2;
  rp1 = cam.rp1;  rp2 = cam.rp2;  rscale = cam.rscale;
  Du = cam.Du;   Dv = cam.Dv;   u0 = cam.u0;    v0 = cam.v0;
  width = cam.width;    height = cam.height;
}

//*****************************************************************************

void LeicaCamera::set_name(const char* n)
{
  this->name = n?n:"";
}

//*****************************************************************************

const char* LeicaCamera::get_name() const
{
  return this->name.c_str();
}

//*****************************************************************************

vcl_istream& LeicaCamera::read_parameters(vcl_istream& in)
{
  // Find the image size specification.
  vcl_string line;
  while(vcl_getline(in, line) && line.find("Image Width") == line.npos);
  if(!in)    
    return in;    
  if(vcl_sscanf(line.c_str(), "  Image Width = %d", &width) != 1)
  {
    in.setstate(vcl_ios::failbit);
    return in;
  }
  if(!vcl_getline(in, line))    
    return in;    
  if(vcl_sscanf(line.c_str(), "  Image Height = %d", &height) != 1)
  {
    in.setstate(vcl_ios::failbit);
    return in;
  }

  // Find the beginning of the calibration parameters.
  while(vcl_getline(in, line) && line.find("Calibration Parameters:") == line.npos);

  // Read each parameter.
  vcl_string unitRotationAngleX, unitRotationAngleY, unitRotationAngleZ;
  double FocalLength, HorizontalScale, ImageCenterX, ImageCenterY, PixelScaleX, PixelScaleY;
  double RadialFactor1, RadialFactor2, TangentialFactor1, TangentialFactor2;
  double TranslationX, TranslationY, TranslationZ, RotationAngleX, RotationAngleY,RotationAngleZ;
  double InverseParams1, InverseParams2, InverseParams3, InverseParams4, InverseParams5;
  double InverseParams6, InverseParams7, InverseParams8;
  double Rows, Columns;

  read(in, "FocalLength", &FocalLength);
  read(in, "HorizontalScale", &HorizontalScale);
  read(in, "ImageCenterX", &ImageCenterX);
  read(in, "ImageCenterY", &ImageCenterY);
  read(in, "PixelScaleX", &PixelScaleX);
  read(in, "PixelScaleY", &PixelScaleY);
  read(in, "RadialFactor1", &RadialFactor1);
  read(in, "RadialFactor2", &RadialFactor2);
  read(in, "TangentialFactor1", &TangentialFactor1);
  read(in, "TangentialFactor2", &TangentialFactor2);
  read(in, "TranslationX", &TranslationX);
  read(in, "TranslationY", &TranslationY);
  read(in, "TranslationZ", &TranslationZ);
  read(in, "RotationAngleX", &RotationAngleX, &unitRotationAngleX);
  read(in, "RotationAngleY", &RotationAngleY, &unitRotationAngleY);
  read(in, "RotationAngleZ", &RotationAngleZ, &unitRotationAngleZ);
  read(in, "InverseParams1", &InverseParams1);
  read(in, "InverseParams2", &InverseParams2);
  read(in, "InverseParams3", &InverseParams3);
  read(in, "InverseParams4", &InverseParams4);
  read(in, "InverseParams5", &InverseParams5);
  read(in, "InverseParams6", &InverseParams6);
  read(in, "InverseParams7", &InverseParams7);
  read(in, "InverseParams8", &InverseParams8);
  read(in, "Rows", &Rows);
  read(in, "Columns", &Columns);

  // Make sure everything read.
  if(!in)
    return in;    

  // Store logical image size.
  columns = static_cast<int>(Columns);
  rows = static_cast<int>(Rows);

  // Store internal parameters. 
  //cyclone focal length, its the actual f, scaled by inverse of the pixel size
  f = FocalLength;          Su = HorizontalScale; 
  u0 = ImageCenterX;        v0 = ImageCenterY; 
  Du = PixelScaleX;         Dv = PixelScaleY;
  k1 = RadialFactor1;       k2 = RadialFactor2; 
  p1 = TangentialFactor1;   p2 = TangentialFactor2;

  // Store camera origin (already in camera axes).
  t(0) = TranslationX;
  t(1) = TranslationY;
  t(2) = TranslationZ;

  // Compute camera axes.
  double omega = RotationAngleX;
  double phi = RotationAngleY;
  double kappa = RotationAngleZ;
  if(unitRotationAngleX == "deg") omega *= vnl_math::pi/180.0;
  if(unitRotationAngleY == "deg") phi *= vnl_math::pi/180.0;
  if(unitRotationAngleZ == "deg") kappa *= vnl_math::pi/180.0;
  R(0,0) = cos(phi)*cos(kappa);
  R(0,1) = sin(omega)*sin(phi)*cos(kappa)-cos(omega)*sin(kappa);
  R(0,2) = cos(omega)*sin(phi)*cos(kappa)+sin(omega)*sin(kappa);
  R(1,0) = cos(phi)*sin(kappa);
  R(1,1) = sin(omega)*sin(phi)*sin(kappa)+cos(omega)*cos(kappa);
  R(1,2) = cos(omega)*sin(phi)*sin(kappa)-sin(omega)*cos(kappa);
  R(2,0) = -sin(phi);
  R(2,1) = sin(omega)*cos(phi);
  R(2,2) = cos(omega)*cos(phi);

  //origin in world coords
  origin = - R.transpose() * t;

  return in;
}

//*****************************************************************************

void LeicaCamera::read(vcl_istream& in, const char* name,
                  double* value, vcl_string* unit)
{
  vcl_string line;
  vcl_string format = "  ";
  format += name;
  format += " = %lg %15s";
  if(vcl_getline(in, line))
    {
    char u[16];
    int nread = vcl_sscanf(line.c_str(), format.c_str(), value, u);
    switch(nread)
      {
      case 0: in.setstate(vcl_ios::failbit); break;
      case 1: if(unit) { *unit = ""; } break;
      case 2: if(unit) { *unit = u; } break;
      }
    }
}

//*****************************************************************************

bool LeicaCamera::get_xyz(vnl_double_2 pUV,
                     vnl_double_3& pXYZ) const
{
  double u1 = columns*pUV[0]/width - u0;
  double v1 = rows*pUV[1]/height - v0;

  double up = (u1/(Du*Su))/rscale;
  double vp = (v1/(Dv))/rscale;
  
  //undo distortion
  double squp = up*up, sqvp = vp * vp;
  double sumsq = squp + sqvp;
  up -= up*rk1*sumsq + up*rk2*sumsq*sumsq + 2*rp1*up*vp + rp2*(3*squp + sqvp);
  vp -= vp*rk1*sumsq + vp*rk2*sumsq*sumsq + rp1*(3*squp + sqvp) + 2*rp2*up*vp;

  up *= rscale;
  vp *= rscale;

  vnl_double_3 pLeicaCamera;
  pLeicaCamera[0] = up;
  pLeicaCamera[1] = vp;
  pLeicaCamera[2] = f;
  pXYZ = R.transpose()*(pLeicaCamera-t);
  return true;
}

//*****************************************************************************

bool LeicaCamera::project(vnl_double_3 const& pXYZ,
                     vnl_double_2& pUV,
                     double* dist2) const
{
  // Project point into camera coordinate system.
  vnl_double_3 pLeicaCamera = R*pXYZ + t;

  // Make sure point is in front of camera.
  if(pLeicaCamera[2] <= 0.1)
    return false;    

  // Compute approximate image coordinates.
  double uPrime = f*pLeicaCamera[0]/pLeicaCamera[2];
  double vPrime = f*pLeicaCamera[1]/pLeicaCamera[2];

  // If the point is too far away from the origin to be in the image
  // without extreme distortion, stop now.
  double r2 = uPrime*uPrime + vPrime*vPrime;
  if(r2 > columns*columns+rows*rows)    
    return false;    

  // Adjust image coordinates for radial and tangential distortion.
  double dUr = uPrime*(k1*r2 + k2*r2*r2);
  double dVr = vPrime*(k1*r2 + k2*r2*r2);
  double dUt = 2*p1*uPrime*vPrime + p2*(r2 + 2*uPrime*uPrime);
  double dVt = p1*(r2+2*vPrime*vPrime) + 2*p2*uPrime*vPrime;
  double u1 = Du*Su*(uPrime + dUr + dUt);
  double v1 = Dv*(vPrime + dVr + dVt);

  // Set distance squared from center if requested.
  if(dist2)    
    *dist2 = u1*u1+v1*v1;
    
  // Adjust image coordinates for image center.
  pUV[0] = width*(u1 + u0)/columns;
  pUV[1] = height*(v1 + v0)/rows;

  // Return whether the point is in the image.
  int u = static_cast<int>(vcl_floor(pUV[0]+0.5));
  int v = static_cast<int>(vcl_floor(pUV[1]+0.5));
  return (u >= 0 && u < width && v >= 0 && v < height);
}

//*****************************************************************************

vnl_double_3 LeicaCamera::get_origin() const
{
  return origin;
}

//*****************************************************************************

void LeicaCamera::setDistCorrectionParams(const vnl_vector_fixed<double,4> &params, double scale)
{
  rk1 = params[0];
  rk2 = params[1];
  rp1 = params[2]; 
  rp2 = params[3];
  rscale = scale;
}

//*****************************************************************************

void LeicaCamera::calcDistortionCorrection(LeicaCamera *cam, vnl_vector_fixed<double, 4> &params, double &scale)
{
  double k1 = cam->k1, k2 = cam->k2, p1 = cam->p1, p2 = cam->p2;

  //Number of samples for each row and each column, so sampsize^2 samples total 
  const int sampsize = 50;  

  //spacing between each sample (in pixels)
  double spacing = 20;

  double half = floor((double)sampsize/2.0);
  double scaledhalf = half * spacing;
  
  vcl_vector<double> uprimes, vprimes;

  //samples the pixel position + its distortion uniformly across the space
  for (double u = -scaledhalf; u <= scaledhalf; u += spacing)
  {
    double squ = u * u;
    for (double v = -scaledhalf; v <= scaledhalf; v += spacing)
    {
      double sqv = v * v;
      uprimes.push_back(u + u*(k1*(squ+sqv)) + u*(k2*(squ+sqv)*(squ+sqv)) + 2*p1*u*v + p2*(3*squ+sqv));
      vprimes.push_back(v + v*(k1*(squ+sqv)) + v*(k2*(squ+sqv)*(squ+sqv)) + p1*(3*squ+sqv) + 2*p2*u*v);
    }
  }

  //normalize the samples
  for (unsigned int i = 0; i < uprimes.size(); i++)
  {
    uprimes[i] /= scaledhalf;
    vprimes[i] /= scaledhalf;
  }

  vnl_matrix<double> A, b;
  unsigned int numrows;
  
  //find the number of rows in the A and b matrices
  if (sampsize % 2) 
    numrows = sampsize * sampsize * 2;
  else
    numrows = (sampsize+1) * (sampsize+1) * 2;

  A.set_size(numrows, 4);
  b.set_size(numrows, 1);

  unsigned int len = (int)half * 2 + 1;   //length of one row
  unsigned int row = 0;
  for (double u = -half; u <= half; u++)
  {
    for (double v = -half; v <= half; v++)
    {
      int index = (int)((u+half)*len + v+half);
      double up = uprimes[index];
      double vp = vprimes[index];
      double squp = up * up;
      double sqvp = vp * vp;

      vnl_vector_fixed<double, 4> Arow;
      vnl_vector_fixed<double, 1> brow;

      Arow[0] = (squp+sqvp)*up;
      Arow[1] = (up*(squp+sqvp)*(squp+sqvp));
      Arow[2] = 2*up*vp;
      Arow[3] = (3*squp + sqvp);
      A.set_row(row, Arow);

      Arow[0] = (squp+sqvp)*vp;
      Arow[1] = (vp*(squp+sqvp)*(squp+sqvp));
      Arow[2] = (3*squp+sqvp);
      Arow[3] = 2*up*vp;
      A.set_row(row+1, Arow);

      brow[0] = up - ((u*spacing)/scaledhalf);
      b.set_row(row, brow);
      brow[0] = vp - ((v*spacing)/scaledhalf);
      b.set_row(row+1, brow);

      row += 2;
    }
  }

  vnl_matrix<double> At = A.transpose();
  vnl_matrix<double> x = vnl_svd<double>(At * A).inverse() * (At * b);
  params = x.get_column(0);
  scale = scaledhalf;
}

//*****************************************************************************

double LeicaCamera::getDist2FromCenter(const vnl_double_2 &imgpt) const
{
  double u = columns*imgpt[0]/width - u0;
  double v = rows*imgpt[1]/height - v0;

  return u*u + v*v;
}

//*****************************************************************************

Camera *LeicaCamera::clone() const
{
  LeicaCamera *lc = new LeicaCamera(*this);
  return (Camera *)lc;
}

//*****************************************************************************