#include "LeicaRangeData.h"

#include <vcl_string.h>
#include <vcl_iostream.h>
#include <vcl_algorithm.h>
#include <vcl_fstream.h>
#include <vnl/vnl_math.h>
#include <vcl_utility.h>
#include <float.h>

#include <rsdl/rsdl_bins.h>

#include "vtkFloatArray.h"
#include "vtkPoints.h"
#include "vtkPolyData.h"
#include "vtkPointData.h"
#include "vtkSmartPointer.h"
#include "vtkXMLPolyDataWriter.h"
#include "utilities.h"
#include "parameters.h"

#include "LocationsAndNormals.h"

const double twoPi = 2*vnl_math::pi;

//*****************************************************************************

LeicaRangeData::LeicaRangeData() : gridWidth(0), gridHeight(0)
{

}

//*****************************************************************************

LeicaRangeData::~LeicaRangeData()
{

}

//*****************************************************************************

bool LeicaRangeData::processData(const char *filename)
{  
  vcl_ifstream infile(filename);
  if (!infile) return false;
  vcl_string line;
  // Read grid width and height.
  int tempwidth, tempheight;
  if(!(vcl_getline(infile, line) && sscanf(line.c_str(), "%d", &tempwidth) == 1))
    return false; 
  if(!(vcl_getline(infile, line) && sscanf(line.c_str(), "%d", &tempheight) == 1))
    return false;
  this->gridWidth = tempwidth;   this->gridHeight = tempheight;

  botPhi = -DBL_MAX; topPhi = DBL_MAX; 
  rmin = DBL_MAX;  rmax = 0.0;

  //Skip transform lines.  
  for(int i = 0; i < 8; i++)    
    vcl_getline(infile, line);

  const int n = 500;
  bool *theta_histogram = new bool[n];
  for (int i = 0; i < n; i++)
    theta_histogram[i] = false;

  unsigned int numpts = 0;

  //Read points.  Scan points are organized with bottom-to-top index
  //varying fastest, and then left-to-right.  Fill in point grid for normals/scales calc  
  //Also calculate an approximate spherical grid for the points to fit into
  vbl_array_2d< PointRecord > point_grid(this->gridHeight, this->gridWidth);
  for(int i = 0; i < this->gridWidth; ++i)
  {
    for(int j = this->gridHeight - 1; j >= 0; --j)
    {
      // Read this point.
      vnl_double_3 point;
      if(!(vcl_getline(infile, line) && sscanf(line.c_str(), "%lg %lg %lg", &point[0], &point[1], &point[2]) == 3))
        return false;

      double x = point[0], y = point[1], z = point[2];

      // Convert to spherical coordinates.
      double r = vcl_sqrt(x*x+y*y+z*z);
    
      //find bounds on range
      if (rmin > r && r > 1.0) rmin = r;
      if (rmax < r) rmax = r;
      double theta = 0.0;
      
      //Calculate phi and update bounds on phi if necessary
      if(r > 0)
      {
        double phi = vcl_acos(z/r);
        if (phi < topPhi)     
          topPhi = phi;        
        if (phi > botPhi)  
          botPhi = phi;            
      }

      //Calculate theta, and update bounds on theta if necessary
      if(x*x+y*y > 0)    
        theta = vcl_atan2(y, x) + vnl_math::pi;    
      
      double mag = point.two_norm();
      
      if (mag != 0.0)
      {
        double bin = (theta/twoPi)*n; 
        theta_histogram[(int)vcl_floor(bin)] = true;
        theta_histogram[((int)vcl_ceil(bin))%n] = true;
      }
      
      point_grid.put(j,i,PointRecord(mag != 0.0, point.normalize(), mag));
      numpts++;
    }
  }

  vcl_cout << "Read " << numpts << " range points.\n";
  int first_angle = 0;
  while (first_angle < n && !theta_histogram[first_angle])    first_angle++;

  int current_size = 0, current_index = -1, longest_size = 0, longest_index = -1;
  for (int i = (first_angle+1)%n; i != first_angle; i = (i+1)%n)
  {
    if (!theta_histogram[i])
    {
      if (!current_size)
        current_index = i;
      current_size++;       
    }
    else
    {
      if (longest_size < current_size)
      {
        longest_size = current_size;
        longest_index = current_index;
      }
      current_size = 0;
    }
  }

  if (longest_size < current_size)
  {
    longest_size = current_size;
    longest_index = current_index;
  }  

  delete [] theta_histogram;

  double beginTheta = 0.0, endTheta = twoPi;
  if (longest_index != -1)
  {
    endTheta = (longest_index/(double)n)*twoPi - vnl_math::pi;
    beginTheta = ((longest_index - n + longest_size)/(double)n)*twoPi - vnl_math::pi;
  }

  while (beginTheta < 0.0) {  beginTheta += twoPi;  endTheta += twoPi; }
  
  double originPhi, originTheta, spacingTheta, spacingPhi;

  double dTheta =  endTheta - beginTheta;
  fov = vcl_fabs(dTheta) * vcl_fabs(topPhi - botPhi);

  rightTheta = endTheta;
  leftTheta = beginTheta;

  spacingTheta = dTheta/gridWidth;
  originTheta = beginTheta;

  spacingPhi = (botPhi-topPhi)/(gridHeight-1); 
  originPhi = topPhi;  

  vcl_cout << topPhi << " " << botPhi << " " << leftTheta << " " << rightTheta << " " << vcl_endl <<
              spacingTheta << " " << spacingPhi << vcl_endl;

  // For all of the bad returns of the range data try to interpolate their values from their neighbors
  for (int i = 0; i < this->gridHeight; i++)
  {
    for (int j = 0; j < this->gridWidth; j++)
    {
      if (!point_grid(i,j).is_valid_return)
        this->interpolate_from_neighbors(i, j, point_grid, point_grid(i,j).direction );
    }
  }

  vcl_string pgname(filename);
  pgname.erase(pgname.size()-3, 3);  pgname += "pg";
    
  PreParameters params; 
  double noise = Parameters::Instance()->GetNoise();
  params.sensor_sigma = noise;
  params.initial_region_max_grid_width = 9 + (int)((noise / 0.002)*.54545);

  //estimate normals and scales
  LocationsAndNormals lan(&params);
  std_dev = lan.estimate(point_grid);
  save_point_grid(pgname.c_str(), point_grid); 
 
  vcl_cout << "completed estimation" << vcl_endl;

  points.clear();  normals.clear();  scales.clear();

  this->sbins = SphericalBins(this->gridHeight, this->gridWidth, originTheta, originPhi, spacingTheta, spacingPhi); 
   
  //Extract the normals and scales and bin the points into spherical bins
  int index = 0; 
  for (int i = 0; i < this->gridHeight; i++)
  {
    for (int j = 0; j < this->gridWidth; j++)
    {    
      const PointRecord &pr = point_grid(i,j);
      if (pr.is_valid_return && (pr.surface_type == PointRecord::PRCurve || pr.surface_type == PointRecord::PRSurface))     
      {
        vnl_double_3 point = pr.estimated_distance * pr.direction;      
        vnl_double_3 spherical_pt = Utilities::cartesianToSpherical(point);

        this->sbins.insertIndex(spherical_pt[1], spherical_pt[2], index);
   
        points.push_back(point);
        normals.push_back(pr.surface_normal);
        scales.push_back(pr.estimated_scale);
        sample_scales.push_back(pr.point_spacing_normal_to_los);
        index++;
      }
    }
  }

  //this->writePolyData("rangedata.vtp");

  return true;
}

//*****************************************************************************

void LeicaRangeData::save_point_grid(const char *filename, const vbl_array_2d<PointRecord> &point_grid)
{
  vcl_cout << "Saving point grid.\n";
  FILE *file = fopen(filename, "wb");

  int rows = (int)point_grid.rows();
  int cols = (int)point_grid.cols();

  fwrite(&rows, sizeof(int), 1, file);
  fwrite(&cols, sizeof(int), 1, file);
  fwrite(&std_dev, sizeof(double), 1, file);

  for (int r = 0; r < rows; r++)
  {
    for (int c = 0; c < cols; c++)
    {
      const PointRecord &pr = point_grid[r][c];
      fwrite(&pr.is_valid_return, sizeof(bool), 1, file);
      if (!pr.is_valid_return)
        continue;
      fwrite(&pr.measured_distance, sizeof(double), 1, file);
      fwrite(pr.direction.data_block(), sizeof(double)*3, 1, file);
      fwrite(&pr.is_initialized, sizeof(bool), 1, file);
      fwrite(&pr.is_done, sizeof(bool), 1, file);
      fwrite(&pr.point_spacing_normal_to_los, sizeof(double), 1, file);
      fwrite(&pr.priority, sizeof(double), 1, file);
      fwrite(&pr.pq_index, sizeof(int), 1, file);
      fwrite(&pr.done_index, sizeof(int), 1, file);
      fwrite(&pr.surface_type, sizeof(PointRecord::SurfaceType), 1, file);
      fwrite(&pr.estimated_scale, sizeof(double), 1, file);
      fwrite(&pr.estimated_angle_scale, sizeof(double), 1, file);
      fwrite(&pr.estimated_distance, sizeof(double), 1, file);
      fwrite(pr.surface_normal.data_block(), sizeof(double)*3, 1, file);
      for (int i = 0; i < 4; i++)
      {
        fwrite(pr.corner_directions[i].data_block(), sizeof(double)*3, 1, file);
        fwrite(&pr.corner_distances, sizeof(double), 1, file);
      }
    }
  }

  fclose(file);
}

//*****************************************************************************

void LeicaRangeData::interpolate_from_neighbors(int r, int c, vbl_array_2d<PointRecord> const& grid, 
                                           vnl_double_3 & interpolated_direction )
{            
  //  Find two closest locations with valid return:
  std::vector< int > row_offsets, col_offsets;
  std::vector< vnl_double_3 > row_directions, col_directions;
  int max_row_offset = vnl_math_max( (int)grid.rows()-1-r, r );
  int max_col_offset = vnl_math_max( (int)grid.cols()-1-c, c );
  int max_offset = vnl_math_max( max_row_offset, max_col_offset );
  
  for ( int offset=1; row_directions.size() < 2 && col_directions.size() < 2 && offset<=max_offset; ++offset )
  {
      if ( r+offset < grid.rows() && grid(r+offset,c).is_valid_return )      {
          row_directions.push_back( grid(r+offset,c).direction );
          row_offsets.push_back( offset );
      }
      if ( r-offset >= 0 && grid(r-offset,c).is_valid_return )       {
          row_directions.push_back( grid(r-offset,c).direction );
          row_offsets.push_back( -offset );
      }
      if ( c+offset < grid.cols() && grid(r,c+offset).is_valid_return )     {
          col_directions.push_back( grid(r,c+offset).direction );
          col_offsets.push_back( offset );
      }
      if ( c-offset >= 0 && grid(r,c-offset).is_valid_return )     {
          col_directions.push_back( grid(r,c-offset).direction );
          col_offsets.push_back( -offset );
      }
  }

  if ( row_directions.size() >= 2  && ( col_directions.size() < 2 || 
       dot_product(row_directions[0],row_directions[1]) > dot_product(col_directions[0],col_directions[1]) ))
      interpolated_direction = interpolate_directions( row_offsets[0], row_directions[0],
                                                       row_offsets[1], row_directions[1] );
  else if ( col_directions.size() >= 2 )
      interpolated_direction = interpolate_directions( col_offsets[0], col_directions[0],
                                                       col_offsets[1], col_directions[1] );
  else
    assert(false);
}

//*****************************************************************************

vnl_double_3 LeicaRangeData::interpolate_directions(int x1, vnl_double_3 const& d1, int x2, 
                                                     vnl_double_3 const& d2 )
{
  assert( x1!=x2 );
  vnl_double_3 result = (double(x1)*d2 - double(x2)*d1) / double(x1-x2);
  result.normalize();   
  return result; 
}

//*****************************************************************************

bool LeicaRangeData::backProject(const vnl_double_3 &image_pt, const vnl_double_3 &cam_origin,
                        vnl_double_3 &projected_pt, unsigned int *pt_index) const
{
  projected_pt = vnl_double_3(0.0, 0.0, 0.0);
  
  vnl_double_3 P1, Pinf;
  int u0, u1, v0, v1;  
  if (!this->getBinRange(image_pt, cam_origin, P1, Pinf, u0, u1, v0, v1))
    return false;  

  //find point closest to line between P1 and Pinf
  vnl_double_3 vec = Pinf - P1;
  double denom = vec.two_norm();
  double closest_dist = DBL_MAX;
  for (int i = u0; i <= u1; i++)
  {
    for (int j = v0; j <= v1; j++)
    {
      const vcl_vector<unsigned int>& bin = this->sbins.getBinAtIndex(this->gridWidth*j + i);
      for (unsigned int k = 0; k < bin.size(); k++)
      {
        const vnl_double_3 &P0 = points[bin[k]];
        double dist = vnl_cross_3d(vec, P1 - P0).two_norm() / denom;
        if (dist < closest_dist)
        {
          closest_dist = dist;
          projected_pt = P0;
          if (pt_index) *pt_index = bin[k];
        }
      }
    }
  }

  //Here we still may need to aproximate the real intersection point instead of the closest

  return closest_dist != DBL_MAX;
}

//*****************************************************************************

unsigned int LeicaRangeData::findClosestTo(const vnl_double_3 &pt) const
{
  return sbins.findClosestTo(pt, points, 0.0);
}

//*****************************************************************************

bool LeicaRangeData::isInView(LeicaCamera *camera) const
{
  double right, left, top, bottom;
  double pixelwidth = (1.0 / camera->getMetricToPixelU())*camera->getPixelSizeRatio(); 
  double pixelheight = 1.0 / camera->getMetricToPixelV();

  //get offsets to the sides of the image
  double xoffset = pixelwidth * (camera->getWidth() / 2.0);
  double yoffset = pixelheight * (camera->getHeight() / 2.0);
  
  vnl_double_2 img_center = camera->getImageCenter();

  //bounds of the image
  right = img_center[0] + xoffset;
  left = img_center[0] - xoffset;
  top = img_center[1] - yoffset;
  bottom = img_center[1] + yoffset;

  //Project the planes formed by the camera origin and each side of the square of the image
  //if it returns true then we know that we have atleast 1 range point visible in this camera
  return this->projectPlanes(left, right, top, bottom, camera, NULL);
}

//*****************************************************************************

bool LeicaRangeData::projectPlanes(double left, double right, double top, double bottom,
                              Camera *camera, vcl_vector<vnl_double_3> *pts ) const
{
  int u[8], v[8];
  vnl_double_3 TR1, TRinf, TL1, TLinf, BR1, BRinf, BL1, BLinf, cam_origin = camera->get_origin();

  //Convert these bounds into points in R3 
  vnl_double_2 pt2d;
  vnl_double_3 topRight, botRight, topLeft, botLeft;
  pt2d[0] = right;  pt2d[1] = top;
  camera->get_xyz(pt2d, topRight);
  pt2d[0] = right;  pt2d[1] = bottom;
  camera->get_xyz(pt2d, botRight);
  pt2d[0] = left;   pt2d[1] = top;
  camera->get_xyz(pt2d, topLeft);
  pt2d[0] = left;   pt2d[1] = bottom;
  camera->get_xyz(pt2d, botLeft);

  //get the bins that each point may back project to.
  this->getBinRange(topRight, cam_origin, TR1, TRinf, u[0], u[1], v[0], v[1]);
  this->getBinRange(topLeft, cam_origin, TL1, TLinf,  u[2], u[3], v[2], v[3]);
  this->getBinRange(botRight, cam_origin, BR1, BRinf, u[4], u[5], v[4], v[5]);
  this->getBinRange(botLeft, cam_origin, BL1, BLinf,  u[6], u[7], v[6], v[7]);

  //calculate plane equations for the left, top, right, and bottom planes of our frustum
  vnl_double_3 leftnorm = vnl_cross_3d(BLinf - BL1, TL1 - BL1).normalize();
  double leftd = -dot_product(leftnorm, BL1);
  vnl_double_3 rightnorm = vnl_cross_3d(TR1 - BR1, BRinf - BR1).normalize();
  double rightd = -dot_product(rightnorm, BR1);
  vnl_double_3 topnorm = vnl_cross_3d(TL1 - TR1, TRinf - TR1).normalize();
  double topd = -dot_product(topnorm, TR1);
  vnl_double_3 botnorm = vnl_cross_3d(BRinf - BR1, BL1 - BR1).normalize();
  double botd = -dot_product(botnorm, BR1);
 
  //Find maxs and mins of the search areas of the points, and we will search
  //in the bins that are between the 4 points
  int u0 = u[0], u1 = u[0], v0 = v[0], v1 = v[0];
  for (int i = 1; i < 8; i++)
  {
    if (u[i] < u0)  u0 = u[i];
    else if (u[i] > u1)  u1 = u[i];
    if (v[i] < v0)  v0 = v[i];
    else if (v[i] > v1)  v1 = v[i];
  }

  //does the search area interesect the grid at all?
  if (v1 < 0 || u1 < 0 || u0 >= gridWidth || v0 >= gridHeight)
    return false;

  //Bound the search to the grid
  v0 = v0 < 0 ? 0 : v0;
  u0 = u0 < 0 ? 0 : u0;
  v1 = v1 >= gridHeight ? gridHeight - 1 : v1;
  u1 = u1 >= gridWidth ? gridWidth - 1 : u1;

  //iterate through all of the bins and classify each point as inside or outside the (inf) frustum created
  //by the 4 bounding planes that we have projected from the camera, through the image plane.
  for (int i = u0; i <= u1; i++)
  {
    for (int j = v0; j <= v1; j++)
    {
      const vcl_vector<unsigned int>& bin = this->sbins.getBinAtIndex(this->gridWidth*j + i);
      for (unsigned int k = 0; k < bin.size(); k++)
      {
        const vnl_double_3 &P0 = points[bin[k]];
        if (Utilities::classifyPoint(leftnorm, leftd, P0) >= 0 &&
            Utilities::classifyPoint(rightnorm, rightd, P0) >= 0 &&
            Utilities::classifyPoint(topnorm, topd, P0) >= 0 &&
            Utilities::classifyPoint(botnorm, botd, P0) >= 0 ) 
        {
          if (pts)   
            pts->push_back(P0);   //keep points inside all of the planes 
          else  
            return true;
        }                
      }
    }
  }

  if (pts) 
    return !pts->empty();

  return false;
}

//*****************************************************************************

bool LeicaRangeData::getBinRange(const vnl_double_3 &image_pt, const vnl_double_3 &cam_origin,
                            vnl_double_3 &P1, vnl_double_3 &Pinf, int &u0, int &u1, int &v0, int &v1) const
{
  //vector of projection
  vnl_double_3 d = (image_pt - cam_origin).normalize();

  //With this range scan the closest possible point in the scan on the projection ray
  P1 = cam_origin + rmin * d;
  Pinf = cam_origin + (rmax*2) * d;

  double r1, rinf, theta1 = 0.0, thetainf = 0.0, phi1 = 0.0, phiinf = 0.0;
  r1 = P1.two_norm();
  rinf = Pinf.two_norm();
  if (P1[0]*P1[0] + P1[1]*P1[1]) 
    theta1 = vcl_atan2(P1[1], P1[0]);  
  if (Pinf[0]*Pinf[0] + Pinf[1]*Pinf[1]) 
    thetainf = vcl_atan2(Pinf[1], Pinf[0]); 

  if (r1 > 0.0) phi1 = vcl_acos(P1[2] / r1);
  if (rinf > 0.0) phiinf = vcl_acos(Pinf[2] / rinf);

  //Calculate index into the grid
  u0 = this->sbins.getThetaIndex(theta1); 
  v0 = this->sbins.getPhiIndex(phi1); 
  u1 = this->sbins.getThetaIndex(thetainf);
  v1 = this->sbins.getPhiIndex(phiinf);    

  //Swap the indices so that we always interate from smaller values to larger values
  if (u1 < u0)    std::swap(u0, u1);
  if (v1 < v0)    std::swap(v0, v1);

  //keypoint projects out side of range data
  if (v1 < 0 || u1 < 0 || u0 >= gridWidth || v0 >= gridHeight)
    return false;

  //Bound the search to the grid
  v0 = v0 < 0 ? 0 : v0;
  u0 = u0 < 0 ? 0 : u0;
  v1 = v1 >= gridHeight ? gridHeight - 1 : v1;
  u1 = u1 >= gridWidth ? gridWidth - 1 : u1;
  
  return true;
}

//*****************************************************************************

void LeicaRangeData::save(const char *dir)
{
  vcl_string filename = dir;
  filename += "/";
  filename += "rangedata.rd";

  double spacingTheta = sbins.getThetaSpacing();
  double spacingPhi = sbins.getPhiSpacing();  
  double originTheta = sbins.getThetaOrigin();
  double originPhi = sbins.getPhiOrigin();

  FILE *file = fopen(filename.c_str(), "wb");
  fwrite(&gridWidth, sizeof(unsigned int), 1, file);
  fwrite(&gridHeight, sizeof(unsigned int), 1, file);
  fwrite(&spacingTheta, sizeof(double), 1, file);
  fwrite(&spacingPhi, sizeof(double), 1, file);
  fwrite(&originPhi, sizeof(double), 1, file);
  fwrite(&originTheta, sizeof(double), 1, file);
  fwrite(&leftTheta, sizeof(double), 1, file);
  fwrite(&rightTheta, sizeof(double), 1, file);
  fwrite(&topPhi, sizeof(double), 1, file);
  fwrite(&botPhi, sizeof(double), 1, file);
  fwrite(&rmin, sizeof(double), 1, file);
  fwrite(&rmax, sizeof(double), 1, file);
  fwrite(&std_dev, sizeof(double), 1, file);
  fwrite(&fov, sizeof(double), 1, file);

  //Next is the data in point normal scale format
  unsigned int size = (unsigned int)points.size();
  fwrite(&size, sizeof(unsigned int), 1, file);
  for (unsigned int i = 0; i < size; i++)
  {
    fwrite(points[i].data_block(), sizeof(double)*3, 1, file); 
    fwrite(normals[i].data_block(), sizeof(double)*3, 1, file); 
    fwrite(&scales[i], sizeof(double), 1, file);
    fwrite(&sample_scales[i], sizeof(double), 1, file);
  }


  fclose(file);
}

//*****************************************************************************

bool LeicaRangeData::load(const char *dir)
{
  vcl_string filename = dir;
  filename += "/";
  filename += "rangedata.rd";

  double spacingTheta, spacingPhi, originPhi, originTheta;

  FILE *file = NULL;
  if (!(file = fopen(filename.c_str(), "rb")))
    return false;

  fread(&gridWidth, sizeof(unsigned int), 1, file);
  fread(&gridHeight, sizeof(unsigned int), 1, file);
  fread(&spacingTheta, sizeof(double), 1, file);
  fread(&spacingPhi, sizeof(double), 1, file);
  fread(&originPhi, sizeof(double), 1, file);
  fread(&originTheta, sizeof(double), 1, file);
  fread(&leftTheta, sizeof(double), 1, file);
  fread(&rightTheta, sizeof(double), 1, file);
  fread(&topPhi, sizeof(double), 1, file);
  fread(&botPhi, sizeof(double), 1, file);
  fread(&rmin, sizeof(double), 1, file);
  fread(&rmax, sizeof(double), 1, file);
  fread(&std_dev, sizeof(double), 1, file);
  fread(&fov, sizeof(double), 1, file);

  //Next is the data in point normal scale format
  unsigned int size; 
  fread(&size, sizeof(unsigned int), 1, file);
  points.resize(size); normals.resize(size); scales.resize(size); sample_scales.resize(size);
  for (unsigned int i = 0; i < size; i++)
  {
    fread(points[i].data_block(), sizeof(double)*3, 1, file); 
    fread(normals[i].data_block(), sizeof(double)*3, 1, file); 
    fread(&scales[i], sizeof(double), 1, file); 
    fread(&sample_scales[i], sizeof(double), 1, file);
  }

  fclose(file);

  this->sbins = SphericalBins(this->gridHeight, this->gridWidth, originTheta, originPhi, spacingTheta, spacingPhi);   
  for(unsigned int i = 0; i < points.size(); i++)
  {
    vnl_double_3 &point = this->points[i];
    vnl_double_3 spherical_pt = Utilities::cartesianToSpherical(point);
    this->sbins.insertIndex(spherical_pt[1], spherical_pt[2], i);
  }

  return true;
}

//*****************************************************************************

void LeicaRangeData::writePolyData(const char *filename) const
{
  vtkSmartPointer<vtkPoints> pts = vtkSmartPointer<vtkPoints>::New(); 
  vtkSmartPointer<vtkFloatArray> norms = vtkSmartPointer<vtkFloatArray>::New();

  unsigned int num_points = (unsigned int)normals.size();

  norms->SetNumberOfComponents(3);
  norms->SetNumberOfTuples(num_points);
  norms->SetName("Normals");

  for (unsigned int i = 0; i < points.size(); i++)
  {
    pts->InsertNextPoint(points[i].data_block());
    norms->SetTuple(i,normals[i].data_block());
  }

  vtkSmartPointer<vtkPolyData> pdata = vtkSmartPointer<vtkPolyData>::New();
  pdata->SetPoints(pts);
  pdata->GetPointData()->SetNormals(norms);

  vtkSmartPointer<vtkXMLPolyDataWriter> writer = vtkSmartPointer<vtkXMLPolyDataWriter>::New();
  writer->SetInput(pdata);
  writer->SetFileName(filename);
  writer->Write();  
}

//*****************************************************************************

