#ifndef RANGE_DATA_H_
#define RANGE_DATA_H_

#include <vnl/vnl_double_3.h>
#include <vcl_vector.h>
#include <vcl_map.h>

#include <vbl/vbl_array_2d.h>

#include "PointRecord.h"
#include "LeicaCamera.h"
#include "SphericalBins.h"

#include "RangeData.h"

class LeicaRangeData : public RangeData
{
public:
  LeicaRangeData();

  ~LeicaRangeData();

  ///Read scan data from a file, specified by name
  bool processData(const char *filename);

  ///back projects a point (image_pt) into the range data provided the origin of the camera that the
  ///image was taken from, and returns the projected point
  bool backProject(const vnl_double_3 &image_pt, const vnl_double_3 &cam_origin,
               vnl_double_3 &projected_pt, unsigned int *pt_index = NULL) const;

  unsigned int findClosestTo(const vnl_double_3 &pt) const;

  bool projectPlanes(double left, double right, double top, double bottom,
                     Camera *camera, vcl_vector<vnl_double_3> *pts ) const;

  ///Returns true if this range data can be seen by this camera
  bool isInView(LeicaCamera *camera) const;  

  vcl_vector<unsigned int> getIndicesInBox(const vnl_double_3 &center, const vnl_double_3 &n1, const vnl_double_3 &n2,
                                           const vnl_double_3 &n3, double s1, double s2, double s3) const;
  
  ///Save and load state of the range data (after readData has been called we can save)
  void save(const char *dir);
  bool load(const char *dir);
  
  ///writes the scan out to a file (vtp) which includes its points and its normals, but no connectivity
  void writePolyData(const char *filename) const;

  //Access functions
  double getLeftTheta() const { return leftTheta; }
  double getRightTheta() const { return rightTheta; }
  double getTopPhi() const { return topPhi; }
  double getBottomPhi() const { return botPhi; }
  double getRMax() const { return rmax; } 
  double getSpacingTheta() const { return sbins.getThetaSpacing(); }
  double getSpacingPhi() const { return sbins.getPhiSpacing(); }
  double getFOV() const { return fov; }
  SphericalBins *getSphericalBins() { return &sbins; }
  const vcl_vector< vcl_vector<unsigned int> > &get_bins() const { return sbins.getBins(); }
  const vcl_vector<double> &get_scales() const { return scales; }
  const vcl_vector<double> &get_sample_scales() const { return sample_scales; }
  int getGridWidth() const { return gridWidth; }
  int getGridHeight() const { return gridHeight; }

private:

  void save_point_grid(const char *filename, const vbl_array_2d<PointRecord> &point_grid);
   
  void interpolate_from_neighbors(int r, int c, vbl_array_2d<PointRecord> const& grid,
                                  vnl_double_3 & interpolated_direction );
  vnl_double_3 interpolate_directions(int x1, vnl_double_3 const& d1, int x2,
                                            vnl_double_3 const& d2 ); 
  bool getBinRange(const vnl_double_3 &image_pt, const vnl_double_3 &cam_origin,
                   vnl_double_3 &P1, vnl_double_3 &Pinf, int &u0, int &u1, int &v0, int &v1) const;

  int gridWidth;
  int gridHeight;  
  
  double rmin, rmax;
  double leftTheta, rightTheta, topPhi, botPhi;
  double fov;

  SphericalBins sbins;

  vcl_vector<double> scales;
  vcl_vector<double> sample_scales;
};

#endif
