#include <vcl_iostream.h>
#include <vcl_fstream.h>
#include <vcl_sstream.h>
#include <vcl_algorithm.h>

#include <vnl/vnl_math.h>
#include <vnl/vnl_matrix.h>
#include <vnl/algo/vnl_svd.h>

#include <rrel/rrel_irls.h>
#include <rrel/rrel_linear_regression.h>
#include <rrel/rrel_m_est_obj.h>
#include <rrel/rrel_muset_obj.h>
#include <rrel/rrel_ran_sam_search.h>
#include <rrel/rrel_tukey_obj.h>
#include <rrel/rrel_util.h>

#include "LocationsAndNormals.h"
#include "PreParameters.h"

#if defined(_MSC_VER)
# pragma warning (disable: 4018)
#endif

// #define trace_on
// #define trace2_on

double extreme_threshold_for_output;

LocationsAndNormals::LocationsAndNormals( PreParameters * parameters )
  : m_parameters(parameters), m_est(0)
{
  progress_callback = 0;
  progress_callback_data = 0;
}

LocationsAndNormals::~LocationsAndNormals()
{
  delete m_est;
}

double
LocationsAndNormals::estimate( vbl_array_2d<PointRecord> & point_grid)
{
  extreme_threshold_for_output = cos( 80.0 * vnl_math::pi / 180.0 );

  delete m_est;
  m_est = new rrel_tukey_obj( m_parameters->tukey_parameter );
  vcl_vector< vnl_int_2 > initial_grid_locs;
  vcl_vector< vnl_int_2 > pqueue;

  // Perform simple initialization of the status flags in the grid.
  initialize_grid( point_grid );

  int total_locations = static_cast<int>(point_grid.rows() * point_grid.cols());

  //  Generate initial estimate at each point and add to priority
  //  queue.
  int initialized_count = 0, failed_initialization = 0, not_considered = 0;
  m_min_inliers_for_surface = m_parameters->min_inliers_for_fit;
  m_min_inliers_for_curve = m_min_inliers_for_surface - 2;
  m_multiplier_for_inlier = 3.0;
  for ( unsigned int r=0; r<point_grid.rows(); ++r )  
    for ( unsigned int c=0; c<point_grid.cols(); ++c )
      {
        if ( point_grid(r,c).is_valid_return && !point_grid(r,c).is_done )
          {
            initial_estimate_at_point( point_grid, vnl_int_2(r,c), pqueue );
            if ( point_grid(r,c).is_initialized )
              {
                initialized_count ++;
                if ( initialized_count % 5000 == 0 ) 
                  vcl_cerr << "Initial estimate " << initialized_count << vcl_endl;
              }
            else
              failed_initialization ++ ;
          }
        else
          not_considered ++ ;
      }
  vcl_cerr << "Pcnt initialized = " << initialized_count / double(total_locations) * 100.0 << '\n'
            << "Pcnt failed initialization = " << failed_initialization / double(total_locations) * 100.0 << '\n'
            << "Pcnt not considered = " << not_considered / double(total_locations) * 100.0 << '\n';

  //  Generate max scale estimate
 
  double mean =  estimate_max_scale( point_grid );

  //  Temporary hack:
  if ( m_max_scale_from_initial_estimates < 1.25 * m_parameters->sensor_sigma ) 
    {
      m_max_scale_from_initial_estimates = 1.25 * m_parameters->sensor_sigma;
      vcl_cout << "max scale adjusted upward to " << m_max_scale_from_initial_estimates << '\n';
    }

  //  Prioritized generation of final estimates, one point at a time.
  m_multiplier_for_inlier = 3.0;
  m_max_scale_for_surface_inlier = m_max_scale_from_initial_estimates;
  m_max_allowed_movement_overall = m_multiplier_for_inlier * m_max_scale_from_initial_estimates;
  m_max_allowed_movement_scale_multiplier = m_multiplier_for_inlier;
  m_min_inliers_for_surface = m_parameters->min_inliers_for_fit;
  m_min_inliers_for_curve = m_min_inliers_for_surface - 2;

  int fraction_count = total_locations / 100;
  int progress_threshold = fraction_count;
  int completed_count = 0;
  while ( !pqueue.empty() )
    {
      vnl_int_2 grid_loc = pqueue[0];
      int r = grid_loc[0], c = grid_loc[1];
      remove_from_pqueue( point_grid, pqueue );
      generate_final_estimate_at_point( point_grid, grid_loc, pqueue );
      if ( point_grid(r,c).is_done ) 
        {
          point_grid(r,c).done_index = completed_count;
          completed_count ++;
          if(completed_count >= progress_threshold)
            {
              progress_threshold += fraction_count;
              if(progress_callback)
                progress_callback(progress_callback_data, float(completed_count)/total_locations);
            }
          if ( completed_count % 5000 == 0 )  vcl_cerr << "Completed " << completed_count << vcl_endl;
        }
    }

  //  Loosen the parameters a bit.
  m_multiplier_for_inlier = 3.75;
  m_max_scale_for_surface_inlier = 1.33 * m_max_scale_from_initial_estimates;
  m_max_allowed_movement_scale_multiplier = m_multiplier_for_inlier;
  m_min_inliers_for_surface = m_parameters->min_inliers_for_fit-1;
  m_min_inliers_for_curve = m_min_inliers_for_surface - 2;

  int new_surface_points = 0;
  int reconsidered_count = 0;
  for ( unsigned int r=0; r<point_grid.rows(); ++r )
    for ( unsigned int c=0; c<point_grid.cols(); ++c )
      {
        if ( point_grid(r,c).is_valid_return && !point_grid(r,c).is_done && point_grid(r,c).is_initialized )
          {
            reconsidered_count ++ ;
            generate_final_estimate_at_point( point_grid, vnl_int_2(r,c), pqueue );
            if ( point_grid(r,c).is_done ) 
              {
                new_surface_points ++ ;
                point_grid(r,c).done_index = completed_count;
                completed_count ++;
                if(completed_count >= progress_threshold)
                  {
                    progress_threshold += fraction_count;
                    if(progress_callback)
                      {
                        progress_callback(progress_callback_data, float(completed_count)/total_locations);
                      }
                  }
                if ( completed_count % 5000 == 0 )  vcl_cerr << "Completed " << completed_count << vcl_endl;
              }
          }
      }
  vcl_cerr << "After first reconsideration, reconsidered_count = " << reconsidered_count
            << ", new_surface_points = " << new_surface_points << vcl_endl;
  reconsidered_count = 0;
  new_surface_points = 0;

  //  Loosen the parameters a bit more.
  m_multiplier_for_inlier = 4.5;
  m_max_scale_for_surface_inlier = 1.67 * m_max_scale_from_initial_estimates;
  m_max_allowed_movement_scale_multiplier = m_multiplier_for_inlier;
  m_min_inliers_for_surface = m_parameters->min_inliers_for_fit-2;
  m_min_inliers_for_curve = m_min_inliers_for_surface - 2;
  bool last_try = false;

  for ( unsigned int r=0; r<point_grid.rows(); ++r )
    for ( unsigned int c=0; c<point_grid.cols(); ++c )
      if ( point_grid(r,c).is_valid_return && !point_grid(r,c).is_done )
        {
          reconsidered_count ++ ;
          reconsider_point_as_surface_or_curve( point_grid, vnl_int_2(r,c), last_try );
          if ( point_grid(r,c).is_done ) 
            {
              new_surface_points ++ ;
              point_grid(r,c).done_index = completed_count;
              completed_count ++;
              if(completed_count >= progress_threshold)
                {
                  progress_threshold += fraction_count;
                  if(progress_callback)
                    progress_callback(progress_callback_data, float(completed_count)/total_locations);
                }
                if ( completed_count % 5000 == 0 )  vcl_cerr << "Completed " << completed_count << vcl_endl;
            }
        }
  vcl_cerr << "After second reconsideration, reconsidered_count = " << reconsidered_count
            << ", new_surface_points = " << new_surface_points << vcl_endl;
  reconsidered_count = 0;
  new_surface_points = 0;

  //  Final loosening of some parameters
  m_multiplier_for_inlier = 5.0;
  m_max_scale_for_surface_inlier = 2.0 * m_max_scale_from_initial_estimates;
  m_max_allowed_movement_scale_multiplier = m_multiplier_for_inlier;

  last_try = true;
  for ( unsigned int r=0; r<point_grid.rows(); ++r )
    for ( unsigned int c=0; c<point_grid.cols(); ++c )
      if ( point_grid(r,c).is_valid_return && !point_grid(r,c).is_done )
        {
          reconsidered_count ++ ;
          reconsider_point_as_surface_or_curve( point_grid, vnl_int_2(r,c), last_try );
          if ( point_grid(r,c).is_done && point_grid(r,c).surface_type == PointRecord::PRSurface )
            new_surface_points ++;
          point_grid(r,c).done_index = completed_count;
          completed_count ++;
          if(completed_count >= progress_threshold)
            {
              progress_threshold += fraction_count;
              if(progress_callback)
                progress_callback(progress_callback_data, float(completed_count)/total_locations);
            }
          if ( completed_count % 5000 == 0 )  vcl_cerr << "Completed " << completed_count << vcl_endl;
        }
  vcl_cerr << "After final reconsideration, reconsidered_count = " << reconsidered_count
            << ", new_surface_points = " << new_surface_points << vcl_endl;

  //  compute_geodesic_scale( point_grid );

  //  Gather final statistics.
  vcl_vector<double> distances;
  int no_return_count = 0, outlier_count=0, surface_count=0, curve_count=0,
    cloud_count = 0, no_return = 0, undersampled_count = 0;
  for ( unsigned int r=0; r<point_grid.rows(); ++r )
    for ( unsigned int c=0; c<point_grid.cols(); ++c )
      {
        if ( !point_grid(r,c).is_valid_return  )
          no_return_count ++ ;
        else if ( point_grid(r,c).surface_type == PointRecord::PROutlier )
          {
            outlier_count ++;
          }
        else if ( point_grid(r,c).surface_type == PointRecord::PRSurface )
          {
            double dist = vnl_math_abs(point_grid(r,c).estimated_distance - point_grid(r,c).measured_distance);
            if ( dot_product(point_grid(r,c).surface_normal, point_grid(r,c).direction) >= 0 )
              {
                vcl_cout << "failing dot product for surface point = " << dot_product(point_grid(r,c).surface_normal, point_grid(r,c).direction)
                          << vcl_endl;
                vcl_cout << "grid loc = " << vnl_int_2(r,c) << vcl_endl;
                assert(false);
              }
            distances.push_back(dist);
            surface_count ++ ;
          }
        else if ( point_grid(r,c).surface_type == PointRecord::PRCurve )
          {
            if ( dot_product(point_grid(r,c).surface_normal, point_grid(r,c).direction) >= 0 )
              {
                vcl_cout << "failing dot product for curve point = " << dot_product(point_grid(r,c).surface_normal, point_grid(r,c).direction)
                          << vcl_endl;
                vcl_cout << "grid loc = " << vnl_int_2(r,c) << vcl_endl;
                assert(false);
              }
            curve_count ++;
          }
        else if ( point_grid(r,c).surface_type == PointRecord::PRCloud )
          cloud_count ++;
        else if ( point_grid(r,c).surface_type == PointRecord::PRUndersampled )
          undersampled_count ++;
        else
          {
            vcl_cout << "\n" << "Oops:  loc = " << vnl_int_2(r,c) << "\n";
            print( vcl_cout, point_grid(r,c) );
            vcl_cout << "\nQuitting\n" << vcl_endl;
            exit(0);
            assert(false);
          }
      }
  int total = outlier_count + surface_count + curve_count + cloud_count + no_return_count + undersampled_count;
        
  vcl_sort( distances.begin(), distances.end() );
  int i_50 = int ( 0.5 * distances.size() );
  int i_75 = int ( 0.75 * distances.size() );
  int i_90 = int ( 0.90 * distances.size() );
  int i_95 = int ( 0.95 * distances.size() );
  int i_99 = int ( 0.99 * distances.size() );
  vcl_cout << "LocationsAndNormals::estimate, at end:\n"
            << "  outliers = " << outlier_count << ", " << outlier_count * 100.0 / total << "%\n"
            << "  no return = " << no_return_count << ", " << no_return_count * 100.0 / total << "%\n"
            << "  undersampled count = " << undersampled_count << ", " << undersampled_count * 100.0 / total << "%\n"
            << "  cloud count = " << cloud_count << ", " << cloud_count * 100.0 / total << "%\n"
            << "  curve count = " << curve_count << ", " << curve_count * 100.0 / total << "%\n"
            << "  surface count = " << surface_count << ", " << surface_count * 100.0 / total << "%\n";
  if ( undersampled_count * 100.0 / total > 25 )
    vcl_cout << "WARNING:  TOO MANY UNDERSAMPLED. Perhaps PreParameters::max_sample_distance_perpendicular_to_los "
              << "is set too low." << vcl_endl;
  if ( distances.size() > 0 )
    {
      vcl_cout << "Statistics on estimation induced surface changes: \n"
                << "  50%: " << distances[i_50] << '\n'
                << "  75%: " << distances[i_75] << '\n'
                << "  90%: " << distances[i_90] << '\n'
                << "  95%: " << distances[i_95] << '\n'
                << "  99%: " << distances[i_99] << '\n'
                << "  max: " << distances.back() << vcl_endl;
    }

    return mean;
}

void
LocationsAndNormals::set_progress_callback(void* callback_data,
                                           progress_callback_type callback)
{
  progress_callback_data = callback_data;
  progress_callback = callback;
}

void
LocationsAndNormals::initialize_grid( vbl_array_2d<PointRecord> & point_grid )
{
  int upper_done_index = static_cast<int>(point_grid.rows() * point_grid.cols()) + 1;
  int undersampled_count = 0;
  for ( unsigned int r=0; r<point_grid.rows(); ++r )
    for ( unsigned int c=0; c<point_grid.cols(); ++c )
      {
        point_grid(r,c).is_initialized = false;
        point_grid(r,c).is_done = false;
        point_grid(r,c).priority = 0.0;
        point_grid(r,c).done_index = upper_done_index;
        //        if ( r == 0 && ( c == 63 || c == 64 ) )  vcl_cout << "Here: \n" ;
        if ( point_grid(r,c).is_valid_return )
          {
            int nbr_row = r + ((r < point_grid.rows()/2 ) ? 1 : -1);
            int nbr_col = c + ((c < point_grid.cols()/2 ) ? 1 : -1);
            double angle_in_col = acos( dot_product(point_grid(r,c).direction,
                                                    point_grid(nbr_row,c).direction) );
            double angle_in_row = acos( dot_product(point_grid(r,c).direction,
                                                    point_grid(r,nbr_col).direction) );
            double z0 = point_grid(r,c).measured_distance;
            if ( angle_in_row > angle_in_col )
              point_grid(r,c).point_spacing_normal_to_los = z0 * angle_in_row;
            else
              point_grid(r,c).point_spacing_normal_to_los = z0 * angle_in_col;

            if ( point_grid(r,c).point_spacing_normal_to_los 
                 > m_parameters->max_sample_distance_perpendicular_to_los )
              {
#ifdef trace_on
                vcl_cout << "PRUndersampled(1)" << vcl_endl;
#endif
#ifdef trace2_on
                vcl_cout << vnl_int_2(r,c) << ":  point_spacing_normal_to_los = " 
                          << point_grid(r,c).point_spacing_normal_to_los
                          << ".  PRUndersampled(1)" << vcl_endl;
#endif
                point_grid(r,c).is_done = true;
                point_grid(r,c).surface_type = PointRecord::PRUndersampled;
                point_grid(r,c).estimated_distance = point_grid(r,c).measured_distance;
                point_grid(r,c).surface_normal = -point_grid(r,c).direction;
                point_grid(r,c).coherence = 0.0;
                point_grid(r,c).patch_inliers = 0;
                point_grid(r,c).patch_wgted_inliers = 0;
                point_grid(r,c).patch_sigma[0] = 0;
                point_grid(r,c).patch_sigma[1] = 0;
                // point_grid(r,c).geodesic_scale = point_grid(r,c).point_spacing_normal_to_los * vnl_math::sqrt2;
                undersampled_count ++ ;
                
              }
          }
      }
  vcl_cerr << "Undersampled count = " << undersampled_count << ", undersampled pcnt = "
            << undersampled_count * 100.0 / ( point_grid.rows() * point_grid.cols() ) << vcl_endl;
}


void 
LocationsAndNormals::initial_estimate_at_point( vbl_array_2d<PointRecord> & point_grid, 
                                                vnl_int_2 grid_loc, 
                                                vcl_vector<vnl_int_2> & pqueue )
{
  /*
  vcl_cout << "\n" << "initial_estimate_at_point at point " << grid_loc << vcl_endl;

  //  Set up the estimator MUSE.  The first argument gives a maximum
  //  number of inliers to allow MUSE to precompute a table of
  //  parameters.  (In a better implementation this would be computed
  //  in a lazy fashion.)  Set up MUSE to use sk-refinement which
  //  gives a better scale estimate --- closer to the one that would
  //  be estimated if there were no outliers.
  bool use_muse_sk_refinement = true;
  rrel_muset_obj muse( vnl_math_sqr(m_parameters->initial_region_max_grid_width+1) + 1, use_muse_sk_refinement );

  //  Set a minimum inlier fraction.  Initially, I had this relatively
  //  high to decrease the number of samples taken, trusting the fits
  //  elsewhere.  I now have it significantly lower because I wanted
  //  to ensure that the initial estimates are reasonable.
  const double min_inlier_fraction( 0.33 );
  muse.set_min_inlier_fraction( min_inlier_fraction );
  */

  //  Objects to store the points after they have been rotated and
  //  centered.  The row,col indices of the points are also recorded
  //  so that we can go backwards from a point to the 2d grid.
  vcl_vector< vnl_double_3 > centered_points;
  vcl_vector< vnl_int_2 > grid_locations;

  //  The transformation to undo the centering.
  vnl_double_3x3 uncenter_rotation;
  int g_row = grid_loc[0], g_col = grid_loc[1];
  //  vcl_cout << "measured location = " << point_grid(g_row, g_col).measured_distance * point_grid(g_row,g_col).direction << vcl_endl;

  //  Gather the points and rotate them based on the LOS direction
  //  stored at the grid location.  In a small region, this points
  //  will all roughly have the same direction, so the noise will
  //  all have approximately the same direction.  By applying the
  //  computed rotation (the "centering rotation"), this direction
  //  becomes the z axis and therefore ordinary linear regression
  //  can be applied.  This function produces the rotation to undo
  //  the centering rotation.
  bool ignore_completed_points = false;
  bool use_completed_only_when_consistent = false; 
  gather_and_center_points( point_grid, grid_loc, ignore_completed_points, use_completed_only_when_consistent,
                            m_parameters->initial_region_max_grid_width-2, centered_points, grid_locations,
                            uncenter_rotation );
  //  vcl_cout << "initial_estimates:  \nuncenter_rotation = \n" << uncenter_rotation << vcl_endl;

  //   bool trace_details = (g_row >= 11 && g_row <= 14 && g_col >= 16 && g_col <= 19 );
  //   if ( trace_details)
  //   vcl_cout << "\n**************\ngrid_loc = " << grid_loc << ", num points = " << centered_points.size() << vcl_endl;
  vnl_double_3x3 centering_rotation = uncenter_rotation.transpose();


  vnl_double_3 tangent, best_tangent, best_other_tangent;
  vnl_double_3 normal;
  bool tangent_generated, normal_generated;
  generate_tangent_and_normal_from_neighbors( point_grid, grid_loc, centering_rotation, best_tangent, 
                                              best_other_tangent, normal, tangent_generated, normal_generated ); 
  if ( !tangent_generated || !normal_generated ) return;

  //   if ( trace_details )
  //    vcl_cout << " best_tangent = " << best_tangent << ", normal = " << normal << vcl_endl;

  double scale = 10 * m_parameters->sensor_sigma;   //  too large, but shouldn't be that much of a problem
  vcl_vector<double> weights( centered_points.size(), 0.0 );
  vcl_vector<double> residuals( centered_points.size(), 0.0 );

  vcl_vector<double> xy_distance_weights( centered_points.size() );
  double avg_half_width = 0.5 * m_parameters->initial_region_max_grid_width;
  double distance_sigma = avg_half_width * point_grid(g_row,g_col).point_spacing_normal_to_los;
  for ( unsigned int i=0; i<centered_points.size(); ++i )
    xy_distance_weights[i] 
      = exp( -( vnl_math_sqr( centered_points[i][0] ) + vnl_math_sqr( centered_points[i][1] ) ) /
             ( 2.0 * vnl_math_sqr( distance_sigma )) );

  /*
  if ( trace_details )
    {
      vcl_cout << "centered points:\n";
      for ( unsigned int i=0; i<centered_points.size(); ++ i )
        {
          vcl_cout << grid_locations[i] << ", " << centered_points[i] << ", xyw = " << xy_distance_weights[i] << vcl_endl;
        }
    }
  */

  vnl_double_3 params;
  double z0 = point_grid(g_row,g_col).measured_distance;
  params[0] = z0; 
  params[1] = normal[0] / -normal[2];
  params[2] = normal[1] / -normal[2];

  double angle_scale = vnl_math::pi / 6.0;
  bool use_normal_weighting = false;
  vnl_double_2 center_2d; 
  vnl_double_2x2 scatter_2d;

  int max_iterations = 2;

  /*
  if ( trace_details )
    vcl_cout << "before IRLS:  params = " << params << ", scale = " << scale << vcl_endl;
  */

  bool irls_succeeded = irls( point_grid, centered_points, grid_locations, grid_loc, xy_distance_weights, uncenter_rotation, 
                              max_iterations, use_normal_weighting, params,
                              scale, angle_scale, residuals, weights, center_2d, scatter_2d );

  /*
  if ( trace_details )
    vcl_cout << "after IRLS:  params = " << params << ", scale = " << scale << vcl_endl;
  */

  //  Find the final inlier count.  This is needed for the priority
  //  calculation for the neighboring points.
  int num_inliers = 0;
  if ( irls_succeeded )
    {
      for ( unsigned int i=0; i<centered_points.size(); ++i )
        {
          if ( vnl_math_abs( residuals[i] ) < m_multiplier_for_inlier * scale && weights[i] > 0) 
            num_inliers ++ ;
        }
    }

  if ( !irls_succeeded || num_inliers < m_min_inliers_for_surface )
    {
      best_other_tangent[0] = -best_tangent[1];
      best_other_tangent[1] = best_tangent[0];
      best_other_tangent[2] = 0.0;
      best_other_tangent.normalize();
      normal = vnl_cross_3d(best_tangent, best_other_tangent );
      double z0 = point_grid(g_row,g_col).measured_distance;
      params[0] = z0; 
      params[1] = normal[0] / -normal[2];
      params[2] = normal[1] / -normal[2];
      irls_succeeded = irls( point_grid, centered_points, grid_locations, grid_loc, xy_distance_weights, uncenter_rotation, 
                             max_iterations, use_normal_weighting, params,
                             scale, angle_scale, residuals, weights, center_2d, scatter_2d );
      num_inliers = 0;
      for ( unsigned int i=0; i<centered_points.size(); ++i )
        {
          if ( vnl_math_abs( residuals[i] ) < m_multiplier_for_inlier * scale && weights[i] > 0) 
            num_inliers ++ ;
        }
    }
  if ( !irls_succeeded ) return;

  double inlier_frac = double( num_inliers ) / centered_points.size();
  //  if ( trace_details )
  //    vcl_cout << "num_inliers = " << num_inliers << ", inlier_fraction = " << inlier_frac << vcl_endl;
  
  //  Compute the associated surface normal, converted from the z
  //  = f(x,y) form of the plane.
  normal[0] = params[1];
  normal[1] = params[2];
  normal[2] = -1;
  normal.normalize();
  vnl_double_3 uncentered_normal = uncenter_rotation * normal;
  double gradient_magnitude = sqrt(1 + vnl_math_sqr(params[1]) + vnl_math_sqr(params[2]));
  //  vcl_cout << "  uncentered_normal = " << uncentered_normal << ", gradient_magnitude = " << gradient_magnitude << vcl_endl;

  //  Compute the priority based on the scale and inlier fraction (higher priority is better). 
  double priority = 1.0/gradient_magnitude * inlier_frac * m_est->wgt( z0 - params[0], scale );
  // vcl_cout << "  resulting priority = " << priority << vcl_endl;

  // point_grid(g_row,g_col).surface_type = PointRecord::PRSurface;
  point_grid(g_row,g_col).priority = priority;
  point_grid(g_row,g_col).estimated_scale = scale;
  point_grid(g_row,g_col).estimated_angle_scale = vnl_math::pi / 6;  //  30 degrees
  point_grid(g_row,g_col).estimated_distance = params[0];
  point_grid(g_row,g_col).surface_normal = uncentered_normal;

#ifdef trace2_on
  vcl_cout << "initial_estimate_at_point at point " << grid_loc 
            << ", params = " << params << ", scale = " << scale << ", movement = " 
            << vnl_math_abs(params[0] - point_grid(g_row,g_col).measured_distance )
            << ", inliers = " << num_inliers
            << ", inlier_frac = " << inlier_frac << vcl_endl;
#endif

  if ( !point_grid(g_row,g_col).is_initialized )
    {
      //  Add it to the priority queue.
      add_to_pqueue( point_grid, pqueue, grid_loc );
      point_grid(g_row,g_col).is_initialized = true;
    }
  else
    {
      //  Improve its location in the priority queue.
      percolate_up( point_grid, pqueue, grid_loc );
    }
}


void
LocationsAndNormals::generate_tangent_and_normal_from_neighbors( vbl_array_2d<PointRecord> const& point_grid,
                                                                 vnl_int_2                        grid_loc,
                                                                 vnl_double_3x3            const& centering_rotation, 
                                                                 vnl_double_3                   & best_tangent,
                                                                 vnl_double_3                   & best_other_tangent,
                                                                 vnl_double_3                   & normal,
                                                                 bool                           & tangent_generated,
                                                                 bool                           & normal_generated )
{
  //  Among the neighbors, find the point closest in depth and
  //  fit a line to it.  Make the scale be the maximum scale.
  //  Reorient the line to the centered coordinate system.
  int g_row = grid_loc[0], g_col = grid_loc[1];
  //  vcl_cout << "generate_tangent_and_normal_from_neighbors:  g_row = " << g_row << ", g_col = " << g_col << vcl_endl;
  tangent_generated = false;
  normal_generated = false;
  bool best_initialized = false;
  int best_r, best_c;
  vnl_double_3 nbr;
  double z0 = point_grid(g_row,g_col).measured_distance;
  vnl_double_3 tangent;
  //  vcl_cout << "  z0 = " << z0 << vcl_endl;

  for ( int r=g_row-2; r<=g_row+2; ++r )
    for ( int c=g_col-2; c<=g_col+2; ++ c)
      {
        if ( r>=0 && r<point_grid.rows() && c>=0 && c<point_grid.cols()
             && ( r != g_row || c != g_col ) && point_grid(r,c).is_valid_return )
          {
            nbr = centering_rotation * point_grid(r,c).direction * point_grid(r,c).measured_distance;
            tangent[0] = nbr[0];
            tangent[1] = nbr[1];
            tangent[2] = nbr[2] - z0;
            tangent.normalize();
            // vcl_cout << "   nbr r=" << r << ", c = " << c << ", nbr = " << nbr 
            //        << ", tangent = " << tangent;
            if ( !best_initialized || vnl_math_abs(tangent[2]) < vnl_math_abs(best_tangent[2]))
              {
                best_initialized = true;
                best_tangent = tangent;
                best_r = r;
                best_c = c;
                // vcl_cout << " new best ";
              }
            // vcl_cout << vcl_endl;
          }
      }

  if ( !best_initialized ) return;
  tangent_generated = true;
  //  vcl_cout << "  tangent is now generated" << vcl_endl;

  int dr = best_r - g_row;
  int dc = best_c - g_col;
  // vcl_cout << "  dr = " << dr << ", dc = " << vcl_endl;
  bool best_other_initialized = false;
  for ( int r=g_row-1; r<=g_row+1; ++r )
    for ( int c=g_col-1; c<=g_col+1; ++ c)
      {
        if ( r>=0 && r<point_grid.rows() && c>=0 && c<point_grid.cols()
             && ( r != g_row || c != g_col ) && point_grid(r,c).is_valid_return 
             && !(r == g_row-dr && c == g_col-dc)
             && !(r == g_row-2*dr && c == g_col-2*dc)
             && !(r == g_row+dr && c == g_col+dc) 
             && !(r == g_row+2*dr && c == g_col+2*dc) )
          {
            //  vcl_cout << "r = " << r << ", c = " << c << vcl_endl;
            nbr = centering_rotation * point_grid(r,c).direction * point_grid(r,c).measured_distance;
            tangent[0] = nbr[0];
            tangent[1] = nbr[1];
            tangent[2] = nbr[2] - z0;
            tangent.normalize();
            // vcl_cout << "  tangent = " << tangent;
            if ( !best_other_initialized || vnl_math_abs(tangent[2]) < vnl_math_abs(best_other_tangent[2]))
              {
                best_other_initialized = true;
                best_other_tangent = tangent;
                // vcl_cout << ", new best";
              }
            // vcl_cout << vcl_endl;
          }
      }
  if ( !best_other_initialized ) 
    {
      //  vcl_cout << "generate_tangent_and_normal_from_neighbors: " << grid_loc
      //        << ", best_tangent = " << best_tangent 
      //        << ", no normal " << vcl_endl;
      return;
    }

  normal = vnl_cross_3d( best_tangent, best_other_tangent );
  normal_generated = true;
  //  vcl_cout << "generate_tangent_and_normal_from_neighbors: " << grid_loc
  //            << ", best_tangent = " << best_tangent 
  //            << ", best_other_tangent = " << best_other_tangent << ", best_normal = "
  //            << normal << vcl_endl;
}


double
LocationsAndNormals::estimate_max_scale( vbl_array_2d<PointRecord>& point_grid )
{
  vcl_vector<double> scales;
  for ( int r=0; r<point_grid.rows(); ++r )
    for ( int c=0; c<point_grid.cols(); ++c )
      if ( point_grid(r,c).is_initialized )
        scales.push_back( point_grid(r,c).estimated_scale );

  double mean, std_dev, inlier_frac;
  int dof = 1;
  rrel_util_intercept_adjust_stats( scales.begin(), scales.end(), mean, std_dev, inlier_frac, dof );
  m_max_scale_from_initial_estimates = mean + m_parameters->multiplier_on_std_dev_for_max_scale * std_dev;

  vcl_cout << "estimate_max_scale = " 
            << " mean = " << mean << ", std_dev = " << std_dev << ", inlier_frac = " << inlier_frac
             << ", m_max_scale_from_initial_estimates = " << m_max_scale_from_initial_estimates << vcl_endl;
  return mean;
}

void 
LocationsAndNormals::generate_final_estimate_at_point( vbl_array_2d<PointRecord> & point_grid, 
                                                       vnl_int_2 const& center_grid_loc,
                                                       vcl_vector<vnl_int_2> & pqueue )
{
  int g_row = center_grid_loc[0], g_col = center_grid_loc[1];

  //  This uses the same centering rotation computation as the initial
  //  estimates.

  //  Objects to store the points after they have been rotated.  The
  //  row,col indices of the points are also recorded so that we can
  //  go backwards from a point to the 2d grid.
  vcl_vector< vnl_double_3 > centered_points;
  vcl_vector< vnl_int_2 > grid_locations;

  //  The transformation to undo the centering.
  vnl_double_3x3 uncenter_rotation;

  //  Apply the gathering and centering (rotation) process.
  bool ignore_completed_points = false;
  bool use_completed_only_when_consistent = true;
  gather_and_center_points( point_grid, center_grid_loc, ignore_completed_points, 
                            use_completed_only_when_consistent, m_parameters->initial_region_max_grid_width,
                            centered_points, grid_locations, uncenter_rotation );

  // bool trace_details = (g_row >= 11 && g_row <= 14 && g_col >= 16 && g_col <= 19 );
  bool trace_details = false;
  //  if ( trace_details)
  //    vcl_cout << "\n**************\ngrid_loc = " << grid_loc << ", num points = " << centered_points.size() << vcl_endl;

  //  Compute the distance-based weights of the points.  This ensures
  //  that points closer to the target point have higher weights.  The
  //  distance computation is a Gaussian drop-off, but is not
  //  normalized (making the weight 1 when the distance is 0).
  vcl_vector<double> xy_distance_weights( centered_points.size() );
  double avg_half_width = 0.5 * m_parameters->initial_region_max_grid_width;
  //  double distance_sigma = avg_half_width / 2;
  double distance_sigma = avg_half_width * point_grid(g_row,g_col).point_spacing_normal_to_los;
  for ( unsigned int i=0; i<centered_points.size(); ++i )
    {
      xy_distance_weights[i] 
          = exp( -( vnl_math_sqr( centered_points[i][0] ) + vnl_math_sqr( centered_points[i][1] ) ) /
               ( 2.0 * vnl_math_sqr( distance_sigma )) );
    }

  //  Create the initial estimated normal and location in the centered coordinate system. 
  vnl_double_3 normal 
    = uncenter_rotation.transpose() * point_grid( center_grid_loc[0], center_grid_loc[1] ).surface_normal;
  vnl_double_3 location
    = point_grid( center_grid_loc[0], center_grid_loc[1] ).estimated_distance * uncenter_rotation.transpose()
    * point_grid( center_grid_loc[0], center_grid_loc[1] ).direction;

  double intercept = location[2];  // location[0] and location[1] should be 0.

  //  if (trace_details) vcl_cout << "normal = " << normal << ", location = " << location << vcl_endl;

  //  Create the linear regression parameters for the plane from the normal.
  vnl_double_3 params;
  params[0] = intercept; 
  params[1] = normal[0] / -normal[2];
  params[2] = normal[1] / -normal[2];
  double scale = point_grid( center_grid_loc[0], center_grid_loc[1] ).estimated_scale;
  double angle_scale = point_grid( center_grid_loc[0], center_grid_loc[1] ).estimated_angle_scale; 

  if ( trace_details )
    {
    // #ifdef trace_on
      double initial_residual = point_grid( center_grid_loc[0], center_grid_loc[1] ). estimated_distance 
        - point_grid( center_grid_loc[0], center_grid_loc[1] ). measured_distance;
      vcl_cout << "\n********\n" << "generate_final_estimate_at_point:  grid_loc = " << center_grid_loc
                << ", num nbrs = " << centered_points.size() << ", initial params = " << params
                << ", scale = " << scale << ", angle_scale = " << angle_scale 
                << ", point_spacing = " << point_grid(g_row, g_col).point_spacing_normal_to_los 
                << ", initial residual = " << initial_residual << vcl_endl;
      vcl_cout << "normal = " << normal << ", location = " << location << vcl_endl;
      for ( unsigned int i=0; i<centered_points.size(); ++i )
        {
          vcl_cout << "    " << centered_points[i] << '\n';
        }

      vcl_cout << '\n';
      for ( unsigned int i=0; i<centered_points.size(); ++i )
        {
          double x = centered_points[i][0];
          double y = centered_points[i][1];
          double z = centered_points[i][2];
          double r = z - ( params[0] + params[1]*x + params[2]*y );
          vcl_cout << "   " << x << ' ' << y << ' ' << r << '\n';
        }
      // #endif
    }

  //  vcl_cout << "generate_final_estimate_at_point:  initial params = " << params
  //            << ", initial scale = " << scale << '\n';

  vcl_vector<double> weights, residuals;

  vnl_double_2x2 scatter_2d;
  vnl_double_2 center_2d;
  bool use_normal_weighting = m_parameters->use_weighting_based_on_normals;


  //  Do the actual IRLS!

  if ( !irls( point_grid, centered_points, grid_locations, center_grid_loc, xy_distance_weights,
              uncenter_rotation, m_parameters->max_irls_iterations, 
              use_normal_weighting, params, scale, angle_scale, residuals, weights, 
              center_2d, scatter_2d ) )
    {
      point_grid(g_row,g_col).priority = 0.0;
      return;
    }
  intercept = params[0];
  normal[0] = -params[1];
  normal[1] = -params[2];
  normal[2] = 1;
  normal.normalize();

  if ( trace_details )
    {
      vcl_cout << "after IRLS:  params = " << params << ",normal = " << normal << '\n'
                << "  scale = " << scale << ", angle_scale = " << angle_scale << '\n';
      vcl_cout << '\n';
      for ( unsigned int i=0; i<centered_points.size(); ++i )
        {
          double x = centered_points[i][0];
          double y = centered_points[i][1];
          vcl_cout << "   " << x << ' ' << y << ' ' << residuals[i] << ' ' << weights[i] << '\n';
        }

    }

  //  Find the final inlier count.  This is needed for the priority
  //  calculation for the neighboring points.
  int num_inliers = 0;
  double wgted_inliers = 0;
  double max_weight = 0;
  for ( unsigned int i=0; i<centered_points.size(); ++i )
    {
      if ( vnl_math_abs( residuals[i] ) < 3.0 * scale && weights[i] > 0)
        {
        num_inliers ++ ;
        wgted_inliers += weights[i];
        }
      if ( weights[i] > max_weight ) max_weight = weights[i];
    }
  double inlier_frac = double( num_inliers ) / centered_points.size();

  vnl_svd<double> svd2d(scatter_2d);
  double max_2d_scale = sqrt(svd2d.W(0));
  double min_2d_scale = sqrt(svd2d.W(1));

  if ( trace_details )
    {
      vcl_cout << "  num_inliers = " << num_inliers << ", centered_points.size() = " << centered_points.size()
                << ", inlier_frac = " << inlier_frac << ", wgted_inliers = " << wgted_inliers << '\n'
                << "  max_2d_scale = " << max_2d_scale << ", min_2d_scale = " << min_2d_scale << '\n';
    }

  bool appears_to_be_curve = min_2d_scale < 0.33*point_grid(g_row,g_col).point_spacing_normal_to_los;  // must be less than half the spacing
  double measurement_movement = vnl_math_abs( params[0] - point_grid(g_row,g_col).measured_distance );
  double variance_on_surface 
    = (1 + vnl_math_sqr(params[1]))*scatter_2d(0,0) + (1+vnl_math_sqr(params[2]))*scatter_2d(1,1);
  bool appears_to_be_cloud = sqrt( variance_on_surface ) < m_parameters->min_ratio_width_to_noise_for_surface * scale;
  bool appears_to_be_outlier_max_movement =  (measurement_movement > m_max_allowed_movement_scale_multiplier * scale)
    || (measurement_movement > m_max_allowed_movement_overall);
  bool appears_to_be_outlier_max_scale = ( scale > m_max_scale_for_surface_inlier );

  //  double coherence = wgted_inliers / (4 * min_2d_scale * 4 * max_2d_scale / vnl_math_sqr(point_grid(g_row,g_col).point_spacing_normal_to_los));
  double coherence = num_inliers / (4 * min_2d_scale * 4 * max_2d_scale / vnl_math_sqr(point_grid(g_row,g_col).point_spacing_normal_to_los));
  
  //#ifdef trace_on
  if ( trace_details ) 
    vcl_cout << "coherence = " << coherence << ", appears_to_be_curve = " << appears_to_be_curve
              << ", std deviation on surface = " << sqrt(variance_on_surface)
              << ", appears_to_be_cloud = " << appears_to_be_cloud << '\n'
              << "measurement_movement = " << measurement_movement
              << ", appears_to_be_outlier_max_movement = " << appears_to_be_outlier_max_movement
              << ", appears_to_be_outlier_max_scale = " << appears_to_be_outlier_max_scale << vcl_endl;
  // #endif
#ifdef trace2_on
  vcl_cout << center_grid_loc << ":  params = " << params << ", scale = " << scale << ", angle_scale = " << 180*angle_scale/vnl_math::pi
            << ", measurement_movement = " << measurement_movement
            << ", inliers = " << num_inliers << ".  ";
#endif

  if ( appears_to_be_cloud && !appears_to_be_curve && !appears_to_be_outlier_max_movement )
    {
      // #ifdef trace_on
      if ( trace_details )
        vcl_cout << "unfinished but looks to be a cloud" << vcl_endl;
      // #endif
#ifdef trace2_on
      vcl_cout << "unfinished but looks to be a cloud" << vcl_endl;
#endif
      point_grid(g_row,g_col).is_done = false;
      return;
    }
  else if ( appears_to_be_outlier_max_movement || appears_to_be_outlier_max_scale )
    {
#ifdef trace_on
      if ( appears_to_be_outlier_max_movement )
        vcl_cout << "unfinished but looks to be outlier --- too few inliers " << vcl_endl;
      else
        vcl_cout << "unfinished but looks to be outlier --- scale too large" << vcl_endl;
#endif
#ifdef trace2_on
      if ( appears_to_be_outlier_max_movement )
        vcl_cout << "unfinished but looks to be outlier --- movement too large " << vcl_endl;
      else
        vcl_cout << "unfinished but looks to be outlier --- scale too large" << vcl_endl;
#endif
      //  This appears to be an outlier.  Don't allow it to continue, and reset
      //  to allow a different initialization to yield a good result.
      point_grid(g_row,g_col).is_done = false;
      return;
    }
  else if ( appears_to_be_curve )
    {
#ifdef trace_on
      vcl_cout << "unfinished but looks to be a curve" << vcl_endl;
#endif
#ifdef trace2_on
      vcl_cout << "unfinished but looks to be a curve" << vcl_endl;
#endif
      point_grid(g_row,g_col).is_done = false;
      return;
    }

  assert( !appears_to_be_cloud && !appears_to_be_curve && !appears_to_be_outlier_max_movement && !appears_to_be_outlier_max_scale );
  point_grid(g_row,g_col).surface_type = PointRecord::PRSurface;

  //  Need the normal of the plane in the uncentered coodinate system.
  //  We  don't  need to  do  anything  about  the estimated  distance
  //  because it is the distance along the line!
  vnl_double_3 uncentered_normal = uncenter_rotation * normal;

  point_grid(g_row, g_col).is_done = true;
  point_grid(g_row, g_col).estimated_scale = scale;
  point_grid(g_row, g_col).estimated_angle_scale = angle_scale;
  point_grid(g_row, g_col).estimated_distance = intercept;
  point_grid(g_row, g_col).surface_normal = uncentered_normal;
  if ( dot_product( point_grid(g_row, g_col).surface_normal, point_grid(g_row, g_col).direction ) > 0 )
    point_grid(g_row, g_col).surface_normal *= -1;
  point_grid(g_row,g_col).coherence = coherence;
  point_grid(g_row,g_col).patch_inliers = num_inliers;
  point_grid(g_row,g_col).patch_wgted_inliers = wgted_inliers;
  point_grid(g_row,g_col).patch_sigma[0] = min_2d_scale;
  point_grid(g_row,g_col).patch_sigma[1] = max_2d_scale;

  //  Compute the covariance matrix for 2d points.  
  vnl_double_2x2 inverse_2d_covar = svd2d.inverse();

#if defined(trace_on)
  vcl_cout << "PRSurface" << vcl_endl;
#endif
#ifdef trace2_on
  {
  double dot = dot_product( point_grid(g_row, g_col).surface_normal, -point_grid(g_row, g_col).direction );
  if ( dot > 1 ) dot = 1;
  double angle = acos(dot)*180/vnl_math::pi;
  vcl_cout << ", normal angle (degrees) = " << angle
            << (angle > 80 ? " - extreme:  " : ":  ") << "PRSurface" << vcl_endl;
  }
#endif
  /*
  double dot = dot_product( point_grid(g_row, g_col).surface_normal, -point_grid(g_row, g_col).direction );
  if ( dot > 1 ) dot = 1;
  if ( dot < extreme_threshold_for_output )
    {
      double angle = acos(dot)*180/vnl_math::pi;
      vcl_cout << center_grid_loc << ":  params = " << params << ", scale = " << scale << ", angle_scale = " << 180*angle_scale/vnl_math::pi
                << ", measurement_movement = " << measurement_movement << ", angle wrt normal = " << angle
                << ", inliers = " << num_inliers << ".  ordinary PRSurface\n";
    }
  */

  if ( trace_details ) vcl_cout << " this is a surface point\n";

  //   Cache the point on the estimated plane and the dot product
  //   between the normal and the location.  This is needed as part of
  //   finding the intersection of a line (through the origin) and the
  //   plane in the for loop that follows.
  vnl_double_3 uncentered_location 
    = point_grid(g_row, g_col).estimated_distance * point_grid(g_row, g_col).direction;
  double location_dot_normal = dot_product( uncentered_location, uncentered_normal );
  double gradient_magnitude = sqrt( 1 + vnl_math_sqr(params[1]) + vnl_math_sqr(params[2]));

  vnl_double_2 point_2d;

  int added_or_changed = 0;

  //  Go through the uncompleted points, computing their priority
  //  based on this estimate.  For those having an improved priority,
  //  their initial estimates are updated and so are their positions
  //  in the priority queue.  In addition, an initial estimate of the
  //  locations and surface normals of these points is generated by
  //  intersecting the backprojection lines of the points with the
  //  plane just estimated.
  for ( unsigned int i=0; i<centered_points.size(); ++i )
    {
      int r = grid_locations[i][0], c = grid_locations[i][1];
      /*
      if ( pqueue_trace )
        vcl_cout << "r = " << r << "c = " << c << ", is_done = "
                  << point_grid(r,c).is_done
                  << ", residual = " << residuals[i] 
                  << ", weights[i] = " << weights[i] << vcl_endl;
      */

      if ( point_grid(r,c).is_done || vnl_math_abs(residuals[i]) > 3.0*scale )
        continue;

      point_2d[0] = centered_points[i][0]; //  - center_2d[0];
      point_2d[1] = centered_points[i][1]; //  - center_2d[1];
      double xy_distance = point_2d.magnitude();
      double mahalanobis_distance = sqrt( dot_product( point_2d, inverse_2d_covar * point_2d ) );

      /*
      if ( pqueue_trace )
        vcl_cout << "r = " << r << "c = " << c << ", point_2d = " << point_2d
                  << ", xy_distance = " << xy_distance 
                  << ", mahalanobis_distance = " << mahalanobis_distance << vcl_endl;
      */

      //  if ( xy_distance > vnl_math::sqrt2 * point_grid(g_row,g_col).point_spacing_normal_to_los
      //       && mahalanobis_distance > 1.0 ) 
      if ( xy_distance > 1.6 * point_grid(g_row,g_col).point_spacing_normal_to_los )
        {
          // if ( pqueue_trace ) vcl_cout << "skipping this point " << vcl_endl;
          continue;   //
        }

      //  if ( pqueue_trace ) vcl_cout << "adding this point " << vcl_endl;

      //  Compute the priority and if it is higher than previous,
      //  update the point's initial estimate and revise its priority
      //  queue position.
      
      double new_priority = 1.0/gradient_magnitude * inlier_frac * weights[i];
      if ( trace_details )
        vcl_cout << "pqueue ops:  loc = (" << r << "," << c 
                  << "),  old priority = " << point_grid(r,c).priority
                  << ", new_priority = " << new_priority << '\n';


      if ( new_priority > point_grid(r,c).priority )
        {
          added_or_changed ++ ;
          //  Assign revised estimate. Start by intersecting the line
          //  through the point with the plane to find the initial
          //  estimated location.
          vnl_double_3 & point_dir = point_grid(r,c).direction;
          double direction_dot_normal = dot_product( point_dir, uncentered_normal );
          if ( vnl_math_abs( direction_dot_normal ) < 1.0e-6 ) continue;   // no interesection, so skip this point

          point_grid(r,c).estimated_distance = point_grid(r,c).measured_distance - residuals[i];  // because of the computation
          point_grid(r,c).surface_normal = uncentered_normal;
          point_grid(r,c).priority = new_priority;
          point_grid(r,c).estimated_scale = scale;
          point_grid(r,c).estimated_angle_scale = angle_scale;

          if ( trace_details )
            vcl_cout << "    estimated_distance = " << point_grid(r,c).estimated_distance
                      << ", surface_normal = " << uncentered_normal << ", estimated_scale = " << scale
                      << ", estimated_angle_scale = " << angle_scale << vcl_endl;

          //  If it is not initialized it is not yet in the priority queue
          if ( !point_grid(r,c).is_initialized )
            {
              point_grid(r,c).is_initialized = true;
              add_to_pqueue( point_grid, pqueue, grid_locations[i] );
            }
          else
            {
              percolate_up( point_grid, pqueue, grid_locations[i] );
            }
        }
    }
#ifdef trace_on
  vcl_cout << "at end:  added_or_changed in pqueue = " << added_or_changed << vcl_endl;
#endif
}


bool 
LocationsAndNormals::irls( vbl_array_2d< PointRecord >  const& point_grid,
                           vcl_vector< vnl_double_3 >  const& centered_points,
                           vcl_vector< vnl_int_2 >     const& grid_locations,
                           vnl_int_2                           center_grid_loc,
                           vcl_vector< double >        const& xy_distance_weights,
                           vnl_double_3x3               const& uncenter_rotation,
                           int                                 max_iterations,
                           bool                                use_normal_weighting,
                           vnl_double_3                      & params,
                           double                            & scale,
                           double                            & angle_scale,
                           vcl_vector< double >             & residuals,
                           vcl_vector< double >             & weights,
                           vnl_double_2                      & center_2d,
                           vnl_double_2x2                    & scatter_2d )
                           
{
  int iterations=0;
  int inlier_count = 0;
  int g_row = center_grid_loc[0], g_col = center_grid_loc[1];
  assert( centered_points.size() == xy_distance_weights.size() );
  residuals.resize( centered_points.size() );
  weights.resize( centered_points.size() );
  vnl_double_3x3 XtX;
  vnl_double_3 Xtz;
  double dot_product_threshold = cos( m_parameters->min_angle_delta_for_convergence );
  bool converged = false;
  vnl_double_3 new_params;
  vnl_double_3 normal, uncentered_normal;
  normal[0] = params[1];
  normal[1] = params[2];
  normal[2] = -1;
  normal.normalize();
  double intercept = params[0];

  bool trace_details = (g_row >= 11 && g_row <= 14 && g_col >= 16 && g_col <= 19 );
  trace_details = false;
  if ( trace_details ) vcl_cout << "\nInside IRLS for (" << g_row << "," << g_col << "), scale = " << scale << "\n";

  // IRLS
  do
    {
      uncentered_normal = uncenter_rotation.transpose() * normal;

      iterations ++ ;
      if ( trace_details ) vcl_cout << "irls iteration " << iterations << '\n';
      // vcl_cout << "generate_final_estimate_at_point:  iterations = " << iterations << vcl_endl;

      //  Compute residuals and weights based on the previous estimate.
      for ( unsigned int i=0; i<centered_points.size(); ++i )
        {
          residuals[i] = centered_points[i][2] - ( params[0] + params[1]*centered_points[i][0] +
                                                   params[2]*centered_points[i][1] );
          weights[i] = xy_distance_weights[i] * m_est->wgt( residuals[i], scale );
          if ( trace_details )
            vcl_cout << "   " << centered_points[i] << ' ' << residuals[i] << ' ' << weights[i] << '\n';
        }

      if ( trace_details )
        {
          vcl_cout << "A:  first 10\n";
          for ( unsigned int i=0; i<10; ++i )
            vcl_cout << "   " << centered_points[i] << ' ' << residuals[i] << ' ' << weights[i] << '\n';
        }

      //  Compute scale early in the computation.
      if ( iterations <= m_parameters->max_irls_scale_iterations && inlier_count > 3 )
        {
          const int num_points_for_fit = 3;
          double * dummy_return_type = 0;
          scale = rrel_util_weighted_scale( residuals.begin(), residuals.end(), weights.begin(),
                                            num_points_for_fit, dummy_return_type );

          if ( scale < m_parameters->sensor_sigma ) scale = m_parameters->sensor_sigma;

          //  Trace 
#ifdef trace_on
          vcl_cout << "re-estimated scale = " << scale << vcl_endl;
#endif
        }

      //  If normals are to be used in weighting, do the weighting and recompute scale
      if ( use_normal_weighting )
        {
          if ( trace_details ) vcl_cout << "foo\n";
          const double min_angle_scale = 15 * vnl_math::pi / 180; // Should be a parameter
          //  Compute the angle errors
          const double min_done_count_for_scale = vnl_math_max(20, vnl_math_rnd(0.25 * vnl_math_sqr(m_parameters->initial_region_max_grid_width)) );
          int done_and_consistent_count = 0;
          double sum_wgted_sq_angle_errors = 0.0;
          double sum_wgts = 0.0;
          for ( unsigned int i=0; i<centered_points.size(); ++i )
            {
              int r = grid_locations[i][0], c = grid_locations[i][1];
              if ( point_grid(r,c).is_done && point_grid(r,c).surface_type == PointRecord::PRSurface )
                {
                  double dot = vnl_math_abs( dot_product( point_grid(r,c).surface_normal, uncentered_normal ) );
                  if ( dot > 1 ) dot = 1;
                  double angle_error = acos(dot);
                  weights[i] *=  vnl_math_sqr(min_angle_scale / angle_scale) * m_est->wgt( angle_error / angle_scale );
                  sum_wgted_sq_angle_errors += weights[i] * vnl_math_sqr(angle_error);
                  sum_wgts += weights[i];
                  if ( weights[i] > 0 ) done_and_consistent_count ++ ;
                }
              else
                weights[i] *= 0.5;   // ??????  Downweight points that have no normal yet??
            }

          // #ifdef trace_on
          // vcl_cout << "done_and_consistent_count = " << done_and_consistent_count << vcl_endl;
          //  #endif
          if ( iterations <= m_parameters->max_irls_scale_iterations 
               && done_and_consistent_count > min_done_count_for_scale )
            {
              // vcl_cout << "generate_final_estimate_at_point:  sum angle wgts = " << sum_wgts << vcl_endl;
              angle_scale = sqrt( sum_wgted_sq_angle_errors / ( sum_wgts * (inlier_count-3.0) / inlier_count) );
              if ( angle_scale < min_angle_scale ) angle_scale = min_angle_scale;  // ensure a minimum
              //  Tracing 
              // #ifdef trace_on
              if ( trace_details )
                vcl_cout << " new angle_scale = " << angle_scale * 180 / vnl_math::pi << '\n';
              // #endif
            }
        }

      //  Compute number of inliers and the sum of the combined weights
      inlier_count = 0;
      for ( unsigned int i=0; i<centered_points.size(); ++i )
        if ( weights[i] > 0.0 && vnl_math_abs(residuals[i]) < 3.0 * scale ) inlier_count ++ ;

      // #ifdef trace_on
      if ( trace_details )
        vcl_cout << "  inlier count = " << inlier_count << vcl_endl;
      // #endif

      if ( trace_details )
        {
          vcl_cout << "B:  first 10\n";
          for ( unsigned int i=0; i<10; ++i )
            vcl_cout << "   " << centered_points[i] << ' ' << residuals[i] << ' ' << weights[i] << '\n';
        }
      
      //  If there are too few inliers then stop.  This allows one chance for resampling.
      if ( inlier_count < m_parameters->min_inliers_for_fit )
        {
          // Tracing 
#ifdef trace_on
          vcl_cout << "irls --- too few inliers;  inlier_count = " << inlier_count  << vcl_endl;
#endif
#ifdef trace2_on
          vcl_cout << center_grid_loc << ":  irls --- too few inliers;  inlier_count = " << inlier_count 
                    << vcl_endl;
#endif
          return false;
        }

      //  Form the ordinary regression least-squares scatter matrix. 
      XtX.fill(0.0);
      Xtz.fill(0.0);
      center_2d.fill(0.0);
      double sum_weights = 0.0;
      for ( unsigned int i=0; i<centered_points.size(); ++i )
        {
          double w = weights[i];
          sum_weights += w;
          double x = centered_points[i][0], y = centered_points[i][1], z = centered_points[i][2];
          XtX(0,0) += w;   XtX(0,1) += w*x;  XtX(0,2) += w*y;
          XtX(1,1) += w*x*x;  XtX(1,2) += w*x*y;
          XtX(2,2) += w*y*y;
          Xtz(0) += w*z;  Xtz(1) += w*x*z;  Xtz(2) += w*y*z; 
          center_2d[0] += w*x;
          center_2d[1] += w*y;
        }
      XtX(1,0) = XtX(0,1);    XtX(2,0) = XtX(0,2);   XtX(2,1) = XtX(1,2);
      center_2d /= sum_weights;

      //  Now gather the information needed to analyze the
      //  distribution of constraints in the "image plane"
      scatter_2d.fill(0.0);
      for ( unsigned int i=0; i<centered_points.size(); ++i )
        {
          double w = weights[i];
          double x = centered_points[i][0] - center_2d[0], y = centered_points[i][1]-center_2d[1];
          scatter_2d(0,0) += w*x*x;
          scatter_2d(0,1) += w*x*y;
          scatter_2d(1,1) += w*y*y;
        }
      scatter_2d(1,0) = scatter_2d(0,1);
      scatter_2d /= sum_weights;

      vnl_svd<double> svd3d(XtX);
      vnl_svd<double> svd2d(scatter_2d);
      double max_2d_scale = sqrt(svd2d.W(0));
      double min_2d_scale = sqrt(svd2d.W(1));

      int XtX_rank = 3; //  will only be lowered within the 

#ifdef trace_on
      vcl_cout << "2d center of mass = " << center_2d 
                << ", max_2d_scale = " << max_2d_scale << ", min_2d_scale = " << min_2d_scale << vcl_endl;
#endif

      //  Test for what would be a curve.  These bounds are based on sampling
      if ( min_2d_scale < 0.33*point_grid(g_row,g_col).point_spacing_normal_to_los )  // if the point is halfway inbetween the scale would be 0.5*point_spacing_normal_to_los
        {
          double minor_axis_distance = vnl_math_abs( dot_product(center_2d,svd2d.V().get_column(1)) );
          double major_axis_distance = vnl_math_abs( dot_product(center_2d,svd2d.V().get_column(0)) );
          // vcl_cout << "minor_axis_distance = " << minor_axis_distance << ", major_axis_distance = " << major_axis_distance << vcl_endl;

          if ( minor_axis_distance > point_grid(g_row,g_col).point_spacing_normal_to_los 
               || major_axis_distance > 2.0*max_2d_scale ) 
            {
#ifdef trace_on
              vcl_cout << "curve point, but too far from inlier set: "
                        << " min_2d_scale = " << min_2d_scale
                        << ", center_2d.magnitude = " << center_2d.magnitude() << vcl_endl;
#endif
#ifdef trace2_on
          vcl_cout << center_grid_loc << ":  curve point, but too far from inlier set: "
                    << " min_2d_scale = " << min_2d_scale << ", center_2d.magnitude = "
                    << center_2d.magnitude() << "inlier count = " << inlier_count << vcl_endl;
#endif
              return false;
            }
#ifdef trace_on
          vcl_cout << "Point appears to be curve" << vcl_endl;
#endif
          XtX_rank = 2;
        }
      
      //  Compute the parameter estimate using the pseudo-inverse --- 
      vnl_double_3 new_params = svd3d.pinverse(XtX_rank) * Xtz;
      
      //  Convert to a normal vector:
      vnl_double_3 new_normal;
      new_normal[0] = new_params[1];  new_normal[1] = new_params[2]; new_normal[2] = -1;
      new_normal.normalize();

#ifdef  trace_on
      vcl_cout << "new_params = " << new_params << vcl_endl;
#endif

      // vcl_cout << "generate_final_estimate_at_point:  new_normal = " << new_normal << vcl_endl;
      
      //  Two separate convergence tests: the first is the change in
      //  the normal, and the second is the change in the position of
      //  the intercept value.
      converged = 
        dot_product( new_normal, normal ) > dot_product_threshold 
        && vnl_math_abs( intercept - new_params[0] ) < m_parameters->intercept_mult_for_convergence * scale;

      // vcl_cout << "generate_final_estimate_at_point:  dot_product = " << dot_product( new_normal, normal )
      //          << ", change in intercept = " << vnl_math_abs( intercept - new_params[0] )
      //          << ", converged = " << converged << '\n';
      
      //  Save the new estimate, both 
      normal = new_normal;
      params = new_params;
      intercept = new_params[0];

      if ( trace_details )
        {
          vcl_cout << "C:  first 10\n";
          for ( unsigned int i=0; i<10; ++i )
            vcl_cout << "   " << centered_points[i] << ' ' << residuals[i] << ' ' << weights[i] << '\n';
        }

    }
  while ( iterations < m_parameters->max_irls_iterations && !converged );

  if ( trace_details )
    {
      vcl_cout << "D:  first 10\n";
      for ( unsigned int i=0; i<10; ++i )
        vcl_cout << "   " << centered_points[i] << ' ' << residuals[i] << ' ' << weights[i] << '\n';
    }

  return true;
}

void
LocationsAndNormals::reconsider_point_as_surface_or_curve( vbl_array_2d<PointRecord> & point_grid, 
                                                           vnl_int_2 grid_loc,
                                                           bool last_try )
{
  // Gather points
  int g_row = grid_loc[0], g_col = grid_loc[1];

  //  This uses the same centering rotation computation as the initial
  //  estimates.

  //  Objects to store the points after they have been rotated.  The
  //  row,col indices of the points are also recorded so that we can
  //  go backwards from a point to the 2d grid.
  vcl_vector< vnl_double_3 > centered_points;
  vcl_vector< vnl_int_2 > grid_locations;

  //  The transformation to undo the centering.
  vnl_double_3x3 uncenter_rotation;

  //  Apply the gathering and centering (rotation) process.
  int region_width = (last_try ? m_parameters->initial_region_max_grid_width+2 : m_parameters->initial_region_max_grid_width );
  bool ignore_completed_points = false;
  bool use_completed_only_when_consistent = true;
  gather_and_center_points( point_grid, grid_loc, ignore_completed_points, use_completed_only_when_consistent,
                            region_width, centered_points, grid_locations, uncenter_rotation );
  vnl_double_3x3 centering_rotation = uncenter_rotation.transpose();


  //  Compute the scatter for the coherence measure...

  vnl_double_2 center_2d(0.0,0.0);
  for ( unsigned int i=0; i<centered_points.size(); ++i )
    center_2d += vnl_double_2(centered_points[i][0], centered_points[i][1]);
  center_2d /= (double)centered_points.size();
      
  vnl_double_2x2 scatter_2d;
  scatter_2d.fill(0.0);
  for ( unsigned int i=0; i<centered_points.size(); ++i )
    {
      double x = centered_points[i][0] - center_2d[0], y = centered_points[i][1]-center_2d[1];
      scatter_2d(0,0) += x*x;
      scatter_2d(0,1) += x*y;
      scatter_2d(1,1) += y*y;
    }
  scatter_2d(1,0) = scatter_2d(0,1);
  scatter_2d /= double(centered_points.size());
  vnl_svd<double> svd2d(scatter_2d);
  double max_2d_scale = sqrt(svd2d.W(0));
  double min_2d_scale = sqrt(svd2d.W(1));
  double final_area_covered =  4*min_2d_scale * 4*max_2d_scale / vnl_math_sqr(point_grid(g_row,g_col).point_spacing_normal_to_los);
  
  int surface_count = 0;
  for ( unsigned int i=0; i<centered_points.size(); ++i )
    {
      if ( point_grid( grid_locations[i][0], grid_locations[i][1] ).is_done && 
           point_grid( grid_locations[i][0], grid_locations[i][1] ).surface_type == PointRecord::PRSurface )
        surface_count ++ ;
    }
#ifdef trace2_on
  vcl_cout << "\n" << "reconsider_point_as_surface_or_curve:  grid_point " << grid_loc
            << ", number of neighbors = " << centered_points.size() << ", surface count = " << surface_count << vcl_endl;
#endif

  //  vcl_cout << "  number gathered = " << centered_points.size() << vcl_endl;

  if ( centered_points.size() < m_min_inliers_for_curve ) 
    {
      // point_grid(g_row,g_col).is_done = true;
      if ( last_try ) 
        {
          point_grid(g_row,g_col).surface_type = PointRecord::PROutlier;
          point_grid(g_row,g_col).estimated_distance = point_grid(g_row,g_col).measured_distance;
          point_grid(g_row,g_col).surface_normal = -point_grid(g_row,g_col).direction;
          point_grid(g_row,g_col).estimated_scale = 1.0;
          point_grid(g_row,g_col).coherence = 0.0;
          point_grid(g_row,g_col).patch_inliers = 0;
          point_grid(g_row,g_col).patch_wgted_inliers = 0;
          point_grid(g_row,g_col).patch_sigma[0] = 0;
          point_grid(g_row,g_col).patch_sigma[1] = 0;
#ifdef trace2_on
          vcl_cout << "reconsider: " << grid_loc << ", PROutlier (isolated)" << vcl_endl;
#endif
        }
      return;
    }

  //  Compute the distance-based weights of the points.  This ensures
  //  that points closer to the target point have higher weights.  The
  //  distance computation is a Gaussian drop-off, but is not
  //  normalized (making the weight 1 when the distance is 0).
  vcl_vector<double> xy_distance_weights( centered_points.size() );
  double avg_half_width = 0.5 * region_width;
  //  double distance_sigma = avg_half_width / 2;
  double distance_sigma = avg_half_width * point_grid(g_row,g_col).point_spacing_normal_to_los;

  vcl_vector<double> distances, weights;
  vcl_vector< vnl_double_3 > normals;
  vcl_vector< double > angle_scales;
  double sum_weighted_los_distances = 0.0;
  double inlier_count = 0;
  double wgted_inliers = 0;
  double sum_weights = 0.0;
  vnl_double_3 centered_normal;

  //  Consider each centered point.  Start by computing the usual xy
  //  distance weight and then start the real work of this loop, which
  //  is gathering information from the neighboring points that are
  //  marked as surfaces to see what they predict about this surface.
  for ( unsigned int i=0; i<centered_points.size(); ++i )
  {
      xy_distance_weights[i] 
          = exp( -( vnl_math_sqr( centered_points[i][0] ) + vnl_math_sqr( centered_points[i][1] ) ) /
               ( 2.0 * vnl_math_sqr( distance_sigma )) );

      int r = grid_locations[i][0], c = grid_locations[i][1];
      if ( point_grid(r,c).is_done && point_grid(r,c).surface_type == PointRecord::PRSurface )
      {
          centered_normal = centering_rotation * point_grid(r,c).surface_normal;
          if ( vnl_math_abs(centered_normal[2]) >= 1.0e-4)    //  safety check for a situation that should occur at most rarely
          {
              double distance_to_plane_along_los = dot_product(centered_normal,centered_points[i]) / centered_normal[2];
              distances.push_back( distance_to_plane_along_los );
              double error_distance = distance_to_plane_along_los - point_grid(g_row,g_col).measured_distance;              
              double weight = m_est->wgt( error_distance, point_grid(r,c).estimated_scale );
              weights.push_back( weight );
              normals.push_back( centered_normal );
              sum_weighted_los_distances += weight * distance_to_plane_along_los;
              sum_weights += weight;
              inlier_count ++;
              wgted_inliers += weight;
              angle_scales.push_back( point_grid(r,c).estimated_angle_scale );
          }
      }
  }

  //  vcl_cout << "  inlier_count = " << inlier_count << vcl_endl;

  //  If enough neighboring points consider this point an inlier,
  //  further tests are needed to see if we want to correct this point
  //  to a surface.  Note that this correction produces a fit that is
  //  different from what we might get with a least-squares fit.  For
  //  example, for a point on a sphere, it will be above to the sphere
  //  rather than below.  Not sure yet how much this matters.
  if ( inlier_count >= m_min_inliers_for_surface )
  {
      //  Compute the actual distance (this is relative to the measured distance)
      double actual_distance = sum_weighted_los_distances / sum_weights;
      double measurement_movement = vnl_math_abs(actual_distance-point_grid(g_row,g_col).measured_distance);

      /*
      vcl_cout << "  actual_distance = " << actual_distance
                << ", measured_distance = " << point_grid(g_row,g_col).measured_distance
                << ", movement = " << measurement_movement << vcl_endl;
      */

      //  Compute the scale from the weight square residuals
      double sum_sq_residuals = 0.0;
      for ( unsigned int i=0; i<distances.size(); ++i )
        sum_sq_residuals += weights[i] * vnl_math_sqr(distances[i] - actual_distance);
      double scale = sqrt( sum_sq_residuals / sum_weights );

      if ( scale < m_parameters->sensor_sigma ) scale = m_parameters->sensor_sigma;
      
      //  Recount the inliers based on the estimated scale.  While
      //  doing so, compute the sum_weighted_normals, in preparation
      //  for estimating the normal and computing a final test.
      inlier_count = 0;
      wgted_inliers = 0;
      vnl_double_3 sum_weighted_normals(0.0, 0.0, 0.0);
      vcl_vector<bool> neighbor_inliers(distances.size(), false );
      for ( unsigned int i=0; i<distances.size(); ++i )
        {
          if ( vnl_math_abs(distances[i]-actual_distance) < m_max_allowed_movement_scale_multiplier*scale &&
               vnl_math_abs(distances[i]-actual_distance) < m_max_allowed_movement_overall )
            {
              inlier_count ++ ;
              wgted_inliers += weights[i];
              neighbor_inliers[i] = true;
              sum_weighted_normals += weights[i] * normals[i];
            }
        }
      // vcl_cout << "  scale = " << scale << "  inlier_count based on position = " << inlier_count << vcl_endl;

      bool scale_too_big = ( scale > m_max_scale_for_surface_inlier );
      bool error_too_large =  (measurement_movement >=  m_max_allowed_movement_scale_multiplier * scale)
        || (measurement_movement > m_max_allowed_movement_overall);
      // vcl_cout << "  scale_too_big = " << scale_too_big << ", error_too_large = " << error_too_large << vcl_endl;

      //  If there still enough inliers, compute the normal and then
      //  see if this estimated normal is consistent with enough
      //  others.
      if ( !scale_too_big && !error_too_large && inlier_count >= m_min_inliers_for_surface )
        {
          vnl_double_3 normal = sum_weighted_normals;
          normal.normalize();
          double estimated_angle_scale = 0.0;
          double sum_weighted_sq_angle_error = 0.0;
          double sum_weights = 0.0;

          // vcl_cout << "  normal = " << normal << vcl_endl;
          
          //  Check the estimated normal against the normals of the
          //  neighboring points that are currently considered to be
          //  inliers.  Those for which the normal is inconsistent are
          //  considered outliers.
          inlier_count = 0;
          wgted_inliers = 0;

          for ( unsigned int i=0; i<distances.size(); ++ i )
            {
              if ( neighbor_inliers[i] )
                {
                  double dot = dot_product( normal, normals[i] );
                  if ( dot > 1.0 ) dot = 1.0;
                  double angle_error = acos(dot);
                  if ( angle_error < m_max_allowed_movement_scale_multiplier*angle_scales[i] )
                    {
                      inlier_count ++;
                      double wgt = m_est->wgt( angle_error, angle_scales[i] );
                      wgted_inliers += wgt;
                      sum_weights += wgt;
                      sum_weighted_sq_angle_error += wgt * vnl_math_sqr(angle_error);
                    }
                }
            }


          //  vcl_cout << "  inlier_count based on normals = " << inlier_count << vcl_endl;

          //  Finally, if there are STILL enough inliers, it is really a surface point!
          if ( inlier_count >= m_min_inliers_for_surface )
            {
              double angle_scale = sqrt( sum_weighted_sq_angle_error / sum_weights );
              double min_angle_scale =4 * vnl_math::pi / 180; // Should be a parameter
              if ( angle_scale <  min_angle_scale ) angle_scale = min_angle_scale;

              //  REALLY NEED TO CALL IRLS HERE TO MAKE THINGS CONSISTENT
              point_grid(g_row,g_col).surface_type = PointRecord::PRSurface;
              point_grid(g_row,g_col).is_done = true;
              point_grid(g_row,g_col).estimated_scale = scale;
              point_grid(g_row,g_col).estimated_angle_scale = angle_scale; 
              point_grid(g_row,g_col).estimated_distance = actual_distance;
              point_grid(g_row,g_col).surface_normal = uncenter_rotation * normal;
              if ( dot_product(point_grid(g_row,g_col).surface_normal, point_grid(g_row,g_col).direction) > 0 )
                point_grid(g_row,g_col).surface_normal *= -1;
              //              point_grid(g_row,g_col).coherence = wgted_inliers / final_area_covered;
              point_grid(g_row,g_col).coherence = inlier_count / final_area_covered;
              point_grid(g_row,g_col).patch_inliers = static_cast<int>(inlier_count);
              point_grid(g_row,g_col).patch_wgted_inliers = wgted_inliers;
              point_grid(g_row,g_col).patch_sigma[0] = min_2d_scale;
              point_grid(g_row,g_col).patch_sigma[1] = max_2d_scale;
#ifdef trace2_on 
              {
              double dot = dot_product( point_grid(g_row, g_col).surface_normal, -point_grid(g_row, g_col).direction );
              if ( dot > 1 ) dot = 1;
              double angle = acos(dot)*180/vnl_math::pi;
              vcl_cout << "reconsider: " << grid_loc << ":  params " << actual_distance << ' '
                        << normal[0]/-normal[2] << ' ' << normal[1]/-normal[2] 
                        << ", scale = " << scale << ", angle_scale = " << 180*angle_scale/vnl_math::pi << '\n'
                        << "             measurement_movement = " << vnl_math_abs(point_grid(g_row,g_col).measured_distance - actual_distance)
                        << ", inliers = " << inlier_count 
                        << ", normal angle (degrees) = " << angle
                        << (angle > 80 ? " - extreme." : ".") << "  PRSurface (from neighbors)" << vcl_endl;
              }
#endif //trace2_on 
              /*
              double dot = dot_product( point_grid(g_row, g_col).surface_normal, -point_grid(g_row, g_col).direction );
              if ( dot > 1 ) dot = 1;
              if ( dot < extreme_threshold_for_output )
                {
                  double angle = acos(dot)*180/vnl_math::pi;
                  vcl_cout << grid_loc << ":  params = " << actual_distance << ' '
                            << normal[0]/-normal[2] << ' ' << normal[1]/-normal[2] 
                            << ", scale = " << scale << ", angle_scale = " << 180*angle_scale/vnl_math::pi 
                            << ", movement = " << vnl_math_abs(point_grid(g_row,g_col).measured_distance - actual_distance)
                            << ", inliers = " << inlier_count 
                            << ", angle wrt normal = " << angle << ".  PRSurface (from neighbors)\n" ;
                }
              */
              return;
            }
        }
    }

  //  return;

  ////////////////////////////////////////////////////////////////////////
  //  If we've reached this point then the surface refitting has failed
  //  and we need to consider this as a curve.
  ////////////////////////////////////////////////////////////////////////

  //  vcl_cout << "proceeding to consider re-estimation as a curve or surface" << vcl_endl;
  
  //  Among the neighbors, find the point closest in depth and
  //  fit a line to it.  Make the scale be the maximum scale.
  //  Reorient the line to the centered coordinate system.
  double initial_depth = point_grid(g_row, g_col).measured_distance;
  vnl_double_3 best_tangent, best_other_tangent, normal;
  bool tangent_generated, normal_generated;
  generate_tangent_and_normal_from_neighbors( point_grid, grid_loc, centering_rotation, best_tangent, best_other_tangent, normal,
                                              tangent_generated, normal_generated ); 

  if ( !tangent_generated )
    {
      if ( last_try )
        {
          point_grid(g_row, g_col).surface_type = PointRecord::PROutlier;
          point_grid(g_row, g_col).estimated_distance = point_grid(g_row, g_col).measured_distance;
          point_grid(g_row, g_col).surface_normal = -point_grid(g_row, g_col).direction;
          point_grid(g_row, g_col).estimated_scale = 1.0;
          point_grid(g_row, g_col).coherence = 0.0;
          point_grid(g_row,g_col).patch_inliers = 0;
          point_grid(g_row,g_col).patch_wgted_inliers = 0;
          point_grid(g_row,g_col).patch_sigma[0] = 0;
          point_grid(g_row,g_col).patch_sigma[1] = 0;
#ifdef trace2_on 
          vcl_cout << "reconsider: " << grid_loc << ":  PROutlier (isolated 2)" << vcl_endl;
#endif
        }
      return;
    }

  double z0 = point_grid(g_row, g_col).measured_distance;
  double surface_scale = m_max_scale_for_surface_inlier;
  weights.resize( centered_points.size(), 0.0 );
  vcl_vector<double> residuals( centered_points.size(), 0.0 );
  vnl_double_3 surface_params;
  bool attempted_surface = false;
  
  if ( normal_generated )
    {
      attempted_surface = true;
      surface_params[0] = z0; 
      surface_params[1] = normal[0] / -normal[2];
      surface_params[2] = normal[1] / -normal[2];
      //      vcl_cout << "testing for surface: with initial surface_params = " << surface_params << vcl_endl;
      /*
      if ( (vnl_math_abs(surface_params[1]) > 2 || vnl_math_abs(surface_params[2]) > 2) && g_row >= 1 && g_row < point_grid.rows()-1
           && g_col >= 1 && g_col < point_grid.cols()-1 )
        {
          vcl_cout << "big slope: " << grid_loc << ", z0 = " << z0 << '\n';
          for ( int r=g_row-1; r<=g_row+1; ++r )
            for ( int c=g_col-1; c<=g_col+1; ++c )
              {
                if ( point_grid(r,c).is_valid_return )
                  vcl_cout << "  " << vnl_int_2(r,c) 
                            << ", centered_point = " << centering_rotation * point_grid(r,c).direction * point_grid(r,c).measured_distance
                            << '\n';
              }
        }
      */

      double angle_scale = vnl_math::pi / 6.0;
      bool use_normal_weighting = true;
      vnl_double_2 center_2d; 
      vnl_double_2x2 scatter_2d;

      int max_iterations = m_parameters->max_irls_iterations;
      if ( irls( point_grid, centered_points, grid_locations, grid_loc, xy_distance_weights, uncenter_rotation, 
                 max_iterations, use_normal_weighting, surface_params,
                 surface_scale, angle_scale, residuals, weights, center_2d, scatter_2d ) )
        {
          // vcl_cout << "returned from IRLS with surface_params = " << surface_params << ", surface_scale = " << surface_scale
          //           << vcl_endl;
          
          //  Find the final inlier count.  This is needed for the priority
          //  calculation for the neighboring points.
          int num_inliers = 0;
          double wgted_inliers = 0;
          double max_weight = 0;
          for ( unsigned int i=0; i<centered_points.size(); ++i )
            {
              if ( vnl_math_abs( residuals[i] ) < 3.0 * surface_scale && weights[i] > 0)
                {
                num_inliers ++ ;
                wgted_inliers += weights[i];
                }
              if ( weights[i] > max_weight ) max_weight = weights[i];
            }
          double inlier_frac = double( num_inliers ) / centered_points.size();
          //  vcl_cout << "num_inliers = " << num_inliers << vcl_endl;

          vnl_svd<double> svd2d(scatter_2d);
          double max_2d_scale = sqrt(svd2d.W(0));
          double min_2d_scale = sqrt(svd2d.W(1));
          //  vcl_cout << "min_2d_scale = " << min_2d_scale << ", point_spacing_normal_to_los = " 
          //        << point_grid(g_row,g_col).point_spacing_normal_to_los << vcl_endl;

          bool appears_to_be_curve = min_2d_scale < 0.4*point_grid(g_row,g_col).point_spacing_normal_to_los;  // must be less than half the spacing
          double measurement_movement = vnl_math_abs( surface_params[0] - point_grid(g_row,g_col).measured_distance );
          double variance_on_surface 
            = (1 + vnl_math_sqr(surface_params[1]))*scatter_2d(0,0) + (1+vnl_math_sqr(surface_params[2]))*scatter_2d(1,1);
          bool appears_to_be_cloud = sqrt( variance_on_surface ) < m_parameters->min_ratio_width_to_noise_for_surface * surface_scale;
          bool appears_to_be_outlier_max_movement 
            =  (measurement_movement > m_max_allowed_movement_overall
                || measurement_movement > m_max_allowed_movement_scale_multiplier * surface_scale);
          bool appears_to_be_outlier_max_scale = ( surface_scale > m_max_scale_for_surface_inlier );
          
#ifdef trace2_on
          vcl_cout << "reconsider:  " << grid_loc << " as surface:  appears_to_be_curve = " << appears_to_be_curve
                    << ", measurement_movement = " << measurement_movement
                    << ", appears_to_be_outlier_max_movement = " << appears_to_be_outlier_max_movement
                    << ", appears_to_be_outlier_max_scale = " << appears_to_be_outlier_max_scale << vcl_endl;
#endif

          if ( num_inliers >= m_min_inliers_for_surface && !appears_to_be_curve && !appears_to_be_outlier_max_movement 
               && ( !appears_to_be_outlier_max_scale || last_try ) )
            {
              normal[0] = surface_params[1];
              normal[1] = surface_params[2];
              normal[2] = -1;
              normal.normalize();
              vnl_double_3 uncentered_normal = uncenter_rotation * normal;
              int g_row = grid_loc[0], g_col = grid_loc[1];
              point_grid(g_row, g_col).estimated_scale = surface_scale;
              point_grid(g_row, g_col).estimated_angle_scale = angle_scale;
              point_grid(g_row, g_col).estimated_distance = surface_params[0];
              point_grid(g_row, g_col).surface_normal = uncentered_normal;
              if ( dot_product( point_grid(g_row, g_col).surface_normal, point_grid(g_row, g_col).direction ) > 0 )
                point_grid(g_row, g_col).surface_normal *= -1;
              point_grid(g_row,g_col).surface_type = (appears_to_be_outlier_max_scale ? PointRecord::PRCloud : PointRecord::PRSurface);
              point_grid(g_row, g_col).is_done = (point_grid(g_row,g_col).surface_type == PointRecord::PRSurface);
              //              point_grid(g_row,g_col).coherence = wgted_inliers / final_area_covered;
              point_grid(g_row,g_col).coherence = num_inliers / final_area_covered;
              point_grid(g_row,g_col).patch_inliers = num_inliers;
              point_grid(g_row,g_col).patch_wgted_inliers = wgted_inliers;
              point_grid(g_row,g_col).patch_sigma[0] = min_2d_scale;
              point_grid(g_row,g_col).patch_sigma[1] = max_2d_scale;

#ifdef trace2_on 
              {
              double dot = dot_product( point_grid(g_row, g_col).surface_normal, -point_grid(g_row, g_col).direction );
              if ( dot > 1 ) dot = 1;
              double angle = acos(dot)*180/vnl_math::pi;
              vcl_cout << "reconsider: " << grid_loc << ":  surface_params " << surface_params
                        << ", surface_scale = " << surface_scale << ", angle_scale = " << 180*angle_scale/vnl_math::pi << '\n'
                        << "            measurement_movement = " << measurement_movement << ", inliers = " << num_inliers
                        << ", normal angle (degrees) = " << angle
                        << (angle > 80 ? " - extreme:  " : ":  ");
              vcl_cout << (appears_to_be_outlier_max_scale ? "PRCloud (from re-estimate)" : "PRSurface (from re-estimate)") << vcl_endl;
              }
#endif //trace2_on 
              /*
              double dot = dot_product( point_grid(g_row, g_col).surface_normal, -point_grid(g_row, g_col).direction );
              if ( dot > 1 ) dot = 1;
              if ( dot < extreme_threshold_for_output )
                {
                  double angle = acos(dot)*180/vnl_math::pi;
                  vcl_cout << grid_loc << ":  params = " << surface_params
                            << ", scale = " << surface_scale << ", angle_scale = " << 180*angle_scale/vnl_math::pi 
                            << ", movement = " << measurement_movement
                            << ", inliers = " << num_inliers
                            << ", angle wrt normal = " << angle 
                            << (appears_to_be_outlier_max_scale ? ".  PRCloud (from re-estimate)" : ". PRSurface (from re-estimate)") << vcl_endl;
                }
              */

              return;
            }
        }
    }


  if ( !last_try ) return; 


  //  Gather points in a bigger area and adjust the parameters...

  /*
  ignore_completed_points = false;
  use_completed_only_when_consistent = true;
  centered_points.resize(0);  grid_locations.resize(0);
  gather_and_center_points( point_grid, grid_loc, ignore_completed_points, use_completed_only_when_consistent,
                            m_parameters->initial_region_max_grid_width+2, centered_points, grid_locations, uncenter_rotation );
  centering_rotation = uncenter_rotation.transpose();
  int temp_min_inliers_for_surface = (m_parameters->initial_region_max_grid_width+2)/2 + 4;
  int temp_min_inliers_for_curve = (m_parameters->initial_region_max_grid_width+2)/2 + 1;
  if ( grid_loc[0] == 237 )
    vcl_cout << "\nlast_try for " << grid_loc << ".  point gathering width = " << m_parameters->initial_region_max_grid_width+2
              << ", min_inliers_for_surface = " << temp_min_inliers_for_surface << ", min_inliers_for_curve = " << temp_min_inliers_for_curve
              << vcl_endl;
  */
  int temp_min_inliers_for_surface = region_width/2 + 3;
  int temp_min_inliers_for_curve = region_width/2 + 1;
  //  vcl_cout << "\nlast_try for " << grid_loc << ".  point gathering width = " << region_width
  //          << ", min_inliers_for_surface = " << temp_min_inliers_for_surface << ", min_inliers_for_curve = " << temp_min_inliers_for_curve
  //         << vcl_endl;


  //   vcl_cout << "Since surface failed, proceeding to test curve." << vcl_endl;

  double line_scale = 2*m_max_scale_for_surface_inlier;  //  hack
  double tx = best_tangent[0], ty = best_tangent[1];
  double normalize_2d = sqrt( vnl_math_sqr(tx) + vnl_math_sqr(ty) );
  double slope_z = best_tangent[2] / normalize_2d;
  tx /= normalize_2d; ty /= normalize_2d;
  double nx = -ty, ny = tx;
  vnl_double_2 line_params( z0, slope_z );
  double cx = 0, cy = 0;
  double distance_bound_2d = 1.5 * point_grid(g_row,g_col).point_spacing_normal_to_los;

  //  if ( grid_loc[0] == 237 )
  //    vcl_cout << "before loop:  tx = " << tx << ", ty = " << ty << ", nx = " << nx << ", ny = " << ny
  //           << ", tangent = " << best_tangent << ", distance_bound_2d = " << distance_bound_2d
  //           << ", line_scale = " << line_scale
  //          << ", line_params = " << line_params << vcl_endl;

  vcl_vector<double> tangent_distances( centered_points.size(), 0.0 );

  int num_inliers;
  int iteration = 0;
  bool failed = false;
  bool converged = true;

  //  Iterate: Gather all points whose 2d distance is within the
  //  appropriate sample spacing and whose z distance is within the
  //  appropriate scale.
  do
    {
      iteration ++ ;

      //      if ( grid_loc[0] == 237 )
      //  vcl_cout << "start of iteration " << iteration << vcl_endl;

      //  Compute residuals and weights
      double sum_weights = 0.0, sum_weighted_x = 0.0, sum_weighted_y = 0.0;
      double sum_sq_weighted_residuals = 0.0;
      num_inliers = 0;
      wgted_inliers = 0;
      for ( unsigned int i=0; i<centered_points.size(); ++i )
        {
          double x = centered_points[i][0], y = centered_points[i][1];
          double normal_distance_2d = (x-cx)*nx + (y-cy)*ny;
          tangent_distances[i] = (x-cx)*tx + (y-cy)*ty;   // signed
          residuals[i] = centered_points[i][2] - (line_params[0] + tangent_distances[i] * line_params[1]);
          weights[i] = xy_distance_weights[i] * m_est->wgt( residuals[i] / line_scale );
            
          if ( vnl_math_abs(normal_distance_2d) <= distance_bound_2d && weights[i] > 0 )
            {
              num_inliers ++ ;
              wgted_inliers += weights[i];
              sum_weights += weights[i];
              sum_weighted_x += weights[i] * x;
              sum_weighted_y += weights[i] * y;
              sum_sq_weighted_residuals += weights[i] * vnl_math_sqr(residuals[i]);
            }
          else
            weights[i] = 0.0;
          /*
          if ( grid_loc[0] == 52 )
            vcl_cout << "  " << i << ":  " << centered_points[i] << ", normal_distance_2d = " << normal_distance_2d
                  << ", tangent_distances[i] = " << tangent_distances[i] << ", residuals[i] = " << residuals[i]
                   << ", weights[i] = " << weights[i] << vcl_endl;
          */
        }
      
      //      if ( grid_loc[0] == 237 )
      //        vcl_cout << "  num_inliers = " << num_inliers << vcl_endl;
      if ( sum_weights < 1.0e-4 || num_inliers <= 2 ) 
        {
          // vcl_cout << "failed because too few inliers." << vcl_endl;
          failed = true;
          break;
        }

      if ( iteration <= m_parameters->max_irls_scale_iterations )
        {
          line_scale = sqrt( sum_sq_weighted_residuals / ( sum_weights * double(num_inliers-2) / double(num_inliers) ) );
          if ( line_scale < m_parameters->sensor_sigma )  line_scale = m_parameters->sensor_sigma;
          //          if ( grid_loc[0] == 237 )
          //  vcl_cout << "  new line_scale = " << line_scale << vcl_endl;
        }

      //  Two-d center  --- changed this to leave cx and cy at 0.
      // cx = sum_weighted_x / sum_weights;  cy = sum_weighted_y / sum_weights;
      //      if ( grid_loc[0] == 237 )
      //  vcl_cout << "  cx = " << cx << ", cy = " << cy << vcl_endl;

      //  Two-d normal
      vnl_double_2x2 scatter2d(0.0);
      for ( unsigned int i=0; i<centered_points.size(); ++i )
        {
          if ( weights[i] == 0.0 ) continue;
          double x = centered_points[i][0]-cx, y = centered_points[i][1]-cy;
          scatter2d(0,0) += weights[i] * x * x;
          scatter2d(0,1) += weights[i] * x * y;
          scatter2d(1,1) += weights[i] * y * y;
        }
      scatter2d(1,0) = scatter2d(0,1);
      vnl_svd<double> scatter_svd(scatter2d);
      double old_tx = tx, old_ty = ty;
      nx = scatter_svd.V(0,1);  ny = scatter_svd.V(1,1);
      tx = scatter_svd.V(0,0);  ty = scatter_svd.V(1,0);
      // tx = scatter_svd.V(0,1);  ty = scatter_svd.V(1,1);
      // nx = scatter_svd.V(0,0);  ny = scatter_svd.V(1,0);
      if ( old_tx*tx + old_ty*ty < 0 )  { tx *= -1; ty *= -1; }
      /*
      if ( grid_loc[0] == 52 )
        vcl_cout << "  nx = " << nx << ", ny = " << ny 
                  << ", tx = " << tx << ", ty = " << ty << ", W = " << scatter_svd.W() 
                  << '\n' << "V = " << scatter_svd.V() << vcl_endl;
      */

      //  Line in z
      vnl_double_2x2 scatter_line(0.0);
      vnl_double_2 b(0.0, 0.0);
      for ( unsigned int i=0; i<centered_points.size(); ++i )
        {
          if ( weights[i] == 0.0 ) continue;
          scatter_line(0,0) += weights[i];
          scatter_line(0,1) += weights[i] * tangent_distances[i];
          scatter_line(1,1) += weights[i] * vnl_math_sqr(tangent_distances[i]);
          b(0) += weights[i] * centered_points[i][2];
          b(1) += weights[i] * centered_points[i][2] * tangent_distances[i];
        }
      scatter_line(1,0) = scatter_line(0,1);
      vnl_svd<double> line_svd(scatter_line);
      vnl_double_2 new_line_params = line_svd.inverse() * b;
      
      converged = iteration > 1 
        && vnl_math_abs(new_line_params[0]-line_params[0]) < m_parameters->intercept_mult_for_convergence * line_scale;
      line_params = new_line_params;
      //      if ( grid_loc[0] == 237 )
      //  vcl_cout << "  line_params = " << line_params << vcl_endl;
          //      ", slope_z = " << slope_z << ", converged = " << converged << vcl_endl;
    }
  while ( iteration < m_parameters->max_irls_iterations && !converged );

  double tangent_distance = vnl_math_abs( cx*tx + cy*ty );
  double normal_distance =  vnl_math_abs( cx*nx + cy*ny );
  double new_z = line_params[0];
  double change_in_z = vnl_math_abs( new_z - initial_depth );
  //  if ( grid_loc[0] == 237 )
  //    vcl_cout << "center point tangent_distance = " << tangent_distance << ", normal_distance = " << normal_distance
  //              << vcl_endl;

  bool not_enough_inliers = failed || (num_inliers < temp_min_inliers_for_curve);
  bool scale_too_large = (line_scale > 1.5*m_max_scale_for_surface_inlier );
  bool inlier_relative_to_scale = (change_in_z <= m_max_allowed_movement_overall 
                                   && change_in_z < m_max_allowed_movement_scale_multiplier * line_scale);
  bool outlier_2d = normal_distance > distance_bound_2d;
  //  if ( grid_loc[0] == 237 )
  //    vcl_cout << "not_enough_inliers = " << not_enough_inliers << ", scale_too_large = " << scale_too_large
  //           << ", inlier_relative_to_scale = " << inlier_relative_to_scale << ", outlier_2d = " << outlier_2d 
  //              << vcl_endl;

  vnl_double_3 tangent_in_3d;
  tangent_in_3d(0) = tx; tangent_in_3d(1) = ty;  tangent_in_3d(2) = line_params[1];
  tangent_in_3d.normalize();

#ifdef trace2_on 
  vcl_cout << "reconsider: " << grid_loc << ":  tried curve, and obtained "
            << "tangent direction = " << tangent_in_3d 
            << ", line_scale = " << line_scale << ", change_in_z = " << change_in_z
            << ", num_inliers = " << num_inliers << ".  ";
#endif //trace2_on 

  //  if ( grid_loc[0] == 237 )
  //    vcl_cout << "reconsider: " << grid_loc << ":  tried curve, and obtained "
  //           << " tangent direction (centered) = " << tangent_in_3d 
  //              << ", line_scale = " << line_scale << ", change_in_z = " << change_in_z
  //              << ", num_inliers = " << num_inliers << ".\n";
 

  if ( !scale_too_large && !not_enough_inliers && !outlier_2d && inlier_relative_to_scale )
    {
      //  point_grid(g_row,g_col).is_done = true;
      point_grid(g_row,g_col).surface_type = PointRecord::PRCurve;
      point_grid(g_row,g_col).estimated_distance = new_z;
      point_grid(g_row,g_col).estimated_scale = line_scale;
      tangent_in_3d = uncenter_rotation * tangent_in_3d;
      vnl_double_3 uncentered_normal = -point_grid(g_row,g_col).direction;
      uncentered_normal = uncentered_normal - dot_product(uncentered_normal,tangent_in_3d)*tangent_in_3d;
      uncentered_normal.normalize();
      if ( dot_product(uncentered_normal, point_grid(g_row,g_col).direction) >= 0 )
        uncentered_normal *= -1;
      //      point_grid(g_row,g_col).coherence = wgted_inliers / final_area_covered;
      point_grid(g_row,g_col).coherence = num_inliers / final_area_covered;
      point_grid(g_row,g_col).patch_inliers = num_inliers;
      point_grid(g_row,g_col).patch_wgted_inliers = wgted_inliers;
      point_grid(g_row,g_col).patch_sigma[0] = min_2d_scale;
      point_grid(g_row,g_col).patch_sigma[1] = max_2d_scale;

#ifdef trace2_on
      vcl_cout << "PRCurve with tangent = " << tangent_in_3d << vcl_endl;
#endif
      //      if ( grid_loc[0] == 237 )
      //        vcl_cout << "PRCurve with tangent = " << tangent_in_3d << vcl_endl;
        
      // point_grid(g_row,g_col).surface_x_dir = uncenter_rotation * tangent_in_3d;
      point_grid(g_row,g_col).surface_normal = uncentered_normal;
      if ( last_try ) 
        {
          //          point_grid(g_row,g_col).is_done = true;
          return;
        }
    }

  //  if ( !last_try || point_grid(g_row,g_col).is_done ) return;

  //  if ( grid_loc[0] == 237 )
  //    vcl_cout << "Making a last, least-squares, try on only the non-surface points." << vcl_endl;
  vnl_double_3x3 xtx(0.0);
  vnl_double_3 xty(0.0, 0.0, 0.0);

  int unfinished_count = 0;
  for ( unsigned int i=0; i<centered_points.size(); ++i )
    {
      int r = grid_locations[i][0],  c = grid_locations[i][1];
      if (point_grid(r,c).is_done ) continue;
      unfinished_count ++ ;
      double x = centered_points[i][0];
      double y = centered_points[i][1];
      double z = centered_points[i][2];
      xtx(0,0) += 1;
      xtx(0,1) += x;
      xtx(0,2) += y;
      xtx(1,1) += x*x;
      xtx(1,2) += x*y;
      xtx(2,2) += y*y;
      xty(0) += z;
      xty(1) += x*z;
      xty(2) += y*z;
    }
  xtx(1,0) = xtx(0,1);
  xtx(2,0) = xtx(0,2);
  xtx(2,1) = xtx(1,2);
  //  vcl_cout << "unfinished count = " << unfinished_count << vcl_endl;
  if ( unfinished_count <= temp_min_inliers_for_surface )
    {
      if ( last_try )
        {
          point_grid(g_row, g_col).surface_type = PointRecord::PROutlier;
          point_grid(g_row, g_col).estimated_distance = point_grid(g_row, g_col).measured_distance;
          point_grid(g_row, g_col).surface_normal = -point_grid(g_row, g_col).direction;
          point_grid(g_row, g_col).estimated_scale = 1.0;
          point_grid(g_row, g_col).coherence = 0.0;
          point_grid(g_row,g_col).patch_inliers = 0;
          point_grid(g_row,g_col).patch_wgted_inliers = 0;
          point_grid(g_row,g_col).patch_sigma[0] = 0;
          point_grid(g_row,g_col).patch_sigma[1] = 0;

#ifdef trace2_on
          vcl_cout << grid_loc << ":  PROutlier (isolated 3)" << vcl_endl;
#endif
          //          if ( grid_loc[0] == 237 )  
          // vcl_cout << grid_loc << ":  PROutlier (isolated 3)" << vcl_endl;
        }

      return;
    }

  vnl_svd<double> svd_last( xtx );
  //  vcl_cout << "singular values = " << svd_last.W() << vcl_endl;
  vnl_double_3 params = svd_last.pinverse() * xty;
  // vcl_cout << "params = " << params << vcl_endl;
  double sum_sq_errors = 0.0;
  for ( unsigned int i=0; i<centered_points.size(); ++i )
    {
      int r = grid_locations[i][0],  c = grid_locations[i][1];
      if (point_grid(r,c).is_done ) continue;
      double x = centered_points[i][0];
      double y = centered_points[i][1];
      double z = centered_points[i][2];
      double residual = z - params[0] - x*params[1] - y*params[2];
      //      vcl_cout << "  " << grid_locations[i] << ", pt = " << centered_points[i] << ",  residual = " << vnl_math_abs(residual) << vcl_endl;
      sum_sq_errors += vnl_math_sqr(residual);
    }
  double final_scale = sqrt( sum_sq_errors / (unfinished_count-3));
  //   vcl_cout << "final scale = " << final_scale << vcl_endl;

  //  Could run IRLS FROM HERE!!!

  bool distribution_is_1d = ( min_2d_scale < 0.45 * point_grid(g_row,g_col).point_spacing_normal_to_los );
  //  if ( grid_loc[0] == 237 )
  //    vcl_cout << "before least-squares fit to all points, center_2d " << center_2d
  //              << ", smaller std dev = " << svd_last_2d.W(1)
  //              << ", point spacing = " << point_grid(g_row,g_col).point_spacing_normal_to_los
  //              << ", distribution_is_1d = " << distribution_is_1d << vcl_endl;

  /*
  vcl_cout << "surrounding depths:\n";
  for ( int r = vnl_math_max(0, g_row-3); r<=vnl_math_min(g_row+3, point_grid.rows()-1); ++r )
    {
      for ( int c = vnl_math_max(0, g_col-3); c<=vnl_math_min(g_col+3, point_grid.cols()-1); ++c )
        {
          if ( point_grid(r,c).is_valid_return )
            vcl_cout << "  " << point_grid(r,c).measured_distance;
          else
            vcl_cout << "      " << -1;
        }
      vcl_cout << '\n';
    }
  */

  if ( final_scale < 20 * m_max_scale_from_initial_estimates && !distribution_is_1d )  //  This number, 20, is pretty random.  Need a better method for deciding.
    {
      normal[0] = params[1];
      normal[1] = params[2];
      normal[2] = -1;
      normal.normalize();
      vnl_double_3 uncentered_normal = uncenter_rotation * normal;
      //      point_grid(g_row,g_col).is_done = true;
      point_grid(g_row, g_col).estimated_scale = final_scale;
      point_grid(g_row, g_col).estimated_angle_scale = 30 * vnl_math::pi / 180;
      point_grid(g_row, g_col).estimated_distance = point_grid(g_row, g_col).measured_distance;
      point_grid(g_row, g_col).surface_normal = uncentered_normal;
      if ( dot_product( point_grid(g_row, g_col).surface_normal, point_grid(g_row, g_col).direction ) > 0 )
        point_grid(g_row, g_col).surface_normal *= -1;
      point_grid(g_row,g_col).surface_type = PointRecord::PRCloud;
      point_grid(g_row,g_col).coherence = unfinished_count / final_area_covered;
      point_grid(g_row,g_col).patch_inliers = unfinished_count;
      point_grid(g_row,g_col).patch_wgted_inliers = 0;
      point_grid(g_row,g_col).patch_sigma[0] = min_2d_scale;
      point_grid(g_row,g_col).patch_sigma[1] = max_2d_scale;
#ifdef trace2_on 
      {
      double dot = dot_product( point_grid(g_row, g_col).surface_normal, -point_grid(g_row, g_col).direction );
      if ( dot > 1 ) dot = 1;
      double angle = acos(dot)*180/vnl_math::pi;
      vcl_cout << "reconsider: " << grid_loc << ":  params " << params
                << ", scale = " << final_scale 
                << ", no movement allowed "
                << ", points used = " << unfinished_count 
                << ", normal angle (degrees) = " << acos(dot)*180/vnl_math::pi
                << ( angle > 80 ? " - extreme: " : ": ") << "PRCloud (at end)" << vcl_endl;
      }
#endif //trace2_on 
 
      /*
      double dot = dot_product( point_grid(g_row, g_col).surface_normal, -point_grid(g_row, g_col).direction );
      if ( dot > 1 ) dot = 1;
      if ( dot < extreme_threshold_for_output )
        {
          double angle = acos(dot)*180/vnl_math::pi;
          vcl_cout << grid_loc << ":  params = "  << params << ", pixel size = " 
                    <<  point_grid(g_row,g_col).point_spacing_normal_to_los 
                    << ", scale = " << final_scale 
                    << ", points_used = " << unfinished_count 
                    << ", neighbors = " << centered_points.size() 
                    << ", angle wrt normal = " << angle 
                    << ".  PRCloud (at end)\n";
          vcl_cout << "  as a curve the following was obtained "
                    << "tangent direction = " << tangent_in_3d 
                    << ", line_scale = " << line_scale << ", change_in_z = " << change_in_z
                    << ", num_inliers = " << num_inliers << ".\n";
        }
      */
     }
  else
    {
      //      point_grid(g_row,g_col).is_done = true;
      point_grid(g_row, g_col).estimated_scale = final_scale;
      point_grid(g_row,g_col).surface_type = PointRecord::PROutlier;
      point_grid(g_row,g_col).estimated_distance = point_grid(g_row,g_col).measured_distance;
      point_grid(g_row,g_col).surface_normal = -point_grid(g_row,g_col).direction;
      point_grid(g_row,g_col).coherence = 0.0;
      point_grid(g_row,g_col).patch_inliers = 0;
      point_grid(g_row,g_col).patch_wgted_inliers = 0;
      point_grid(g_row,g_col).patch_sigma[0] = 0;
      point_grid(g_row,g_col).patch_sigma[1] = 0;
      
#ifdef trace2_on 
      vcl_cout << "reconsider: " << grid_loc << ":  params " << params
                << ", scale = " << final_scale 
                << ", no movement allowed "
                << ", points used = " << unfinished_count 
                << ":  PROutlier (scale too large)" << vcl_endl;
#endif //trace2_on 
    }
}


void
LocationsAndNormals::gather_and_center_points( vbl_array_2d<PointRecord> & point_grid,
                                               vnl_int_2 const& center_grid_loc,
                                               bool ignore_completed_points, 
                                               bool use_completed_only_when_consistent,
                                               int max_grid_width,
                                               vcl_vector< vnl_double_3 > & centered_points,
                                               vcl_vector< vnl_int_2 > & grid_locations,
                                               vnl_double_3x3 & uncenter_rotation )
{
  int g_row = center_grid_loc[0], g_col = center_grid_loc[1];
  vnl_double_3 & new_z_axis = point_grid( g_row, g_col ).direction;

  //  vcl_cout << "gather_and_center_points:  center_grid_loc = " << center_grid_loc << vcl_endl;

  //  Form a right-handed coordinate system based on the old x and y
  //  axes by using Gram-Schmidt followed by a cross product.  The issue
  //  to be careful of is that the local new z axis could conceivably
  //  be along the x or the y axis of the global coordinate system.
  //  Therefore, we start by choosing between the current x and y axes
  //  based on which is further from the new z axis.
  vnl_double_3 old_x_axis(0.0, 0.0, 0.0), old_y_axis(0.0, 0.0, 0.0);
  old_x_axis[0] = 1.0;  old_y_axis[1] = 1.0;
  vnl_double_3 new_x_axis(0.0, 0.0, 0.0), new_y_axis(0.0, 0.0, 0.0);
  if ( vnl_math_abs( dot_product(old_x_axis, new_z_axis) ) 
       < vnl_math_abs( dot_product(old_y_axis, new_z_axis) ) )
    {
      new_x_axis = old_x_axis - dot_product(old_x_axis, new_z_axis) * new_z_axis;
      new_x_axis.normalize();
      new_y_axis = vnl_cross_3d(new_z_axis, new_x_axis);
    }
  else 
    {
      new_y_axis = old_y_axis - dot_product(old_y_axis, new_z_axis) * new_z_axis;
      new_y_axis.normalize();
      new_x_axis = vnl_cross_3d(new_y_axis, new_z_axis);
    }

  //  Compute the rotation matrix needed to center and uncenter the
  //  points.  See my notes from 2/17/2006.
  vnl_double_3x3 centering_rotation;
  centering_rotation.set_row(0, new_x_axis);
  centering_rotation.set_row(1, new_y_axis);
  centering_rotation.set_row(2, new_z_axis);
  uncenter_rotation = centering_rotation.transpose();

  //  Compute the grid size adaptively based on the angular sampling.
  int nbr_row = (g_row < point_grid.rows()/2 ? g_row+1 : g_row-1);
  int nbr_col = (g_col < point_grid.cols()/2 ? g_col+1 : g_col-1);
  double horizontal_angle = acos( dot_product( point_grid(g_row,g_col).direction, point_grid(g_row,nbr_col).direction) );
  double vertical_angle = acos( dot_product( point_grid(g_row,g_col).direction, point_grid(nbr_row,g_col).direction) );

  /*
  vcl_cout << "gather_and_center_points:  loc = " << center_grid_loc
            << ", nbr_row = " << nbr_row << ", nbr_col = " << nbr_col 
            << ", horizontal_angle = " << 180 / vnl_math::pi * horizontal_angle
            << ", vertical_angle = " << 180 / vnl_math::pi * vertical_angle
            << vcl_endl;
   */
   
  int half_num_cols, half_num_rows;
  if ( horizontal_angle < vertical_angle )
    {
      half_num_rows = max_grid_width / 2;
      half_num_cols = vnl_math_rnd( vertical_angle / horizontal_angle * half_num_rows );
      //  vcl_cout << "  horizontal less:  half_num_rows = " << half_num_rows
      //            << ", half_num_cols = " << half_num_cols;
      if ( half_num_cols > max_grid_width / 2 )
        {
          half_num_cols = max_grid_width / 2;
          // vcl_cout << ", adjusted to " << half_num_cols;
        }
      // vcl_cout << vcl_endl;
    }
  else
    {
      half_num_cols = max_grid_width / 2;
      half_num_rows = vnl_math_rnd( horizontal_angle / vertical_angle * half_num_cols );
      // vcl_cout << "  vertical less:  half_num_cols = " << half_num_cols
      //          << ", half_num_rows = " << half_num_rows;
      if ( half_num_rows > max_grid_width / 2 )
        {
          half_num_rows = max_grid_width / 2;
          // vcl_cout << ", adjusted to " << half_num_rows;
        }
      // vcl_cout << vcl_endl;
    }


  // Form the bounds on the grids over which we are gathering points to
  // fit the transformation estimate.
  int min_row = g_row - half_num_rows;
  int max_row = g_row +  half_num_rows;
  if ( min_row < 0 ) min_row = 0;
  if ( max_row >= point_grid.rows() ) max_row = (int)point_grid.rows()-1;
  int min_col = g_col - half_num_cols;
  int max_col = g_col +  half_num_cols;
  if ( min_col < 0 ) min_col = 0;
  if ( max_col >= point_grid.cols() ) max_col = (int)point_grid.cols()-1;

  //   vcl_cout << "gather_and_center_points:  min_row = " << min_row << ", max_row = " << max_row
  //            << ", min_col = " << min_col << ", max_col = " << max_col << vcl_endl;
  
  //  Now actually gather and center the points.
  vnl_double_3 centered_point;
  vnl_double_3 centered_normal;
  double center_point_measured_distance = point_grid(g_row,g_col).measured_distance;
  centered_points.clear();
  grid_locations.clear();
  double sqr_distance_limit = vnl_math_sqr( 2 * max_grid_width * point_grid(g_row,g_col).point_spacing_normal_to_los );
//  vcl_cout << "max_grid_width = " << max_grid_width 
//           << ", point_spacing_normal_to_los = " << point_grid(g_row, g_col).point_spacing_normal_to_los
//            << ", distance_limit = " << sqrt(sqr_distance_limit) << vcl_endl;
  for ( unsigned int row=min_row; row<=max_row; ++row )
    for ( unsigned int col=min_col; col<=max_col; ++col )
      {
        bool use_point;
        centered_point = point_grid(row,col).measured_distance * centering_rotation * point_grid(row,col).direction;
        double sqr_point_distance = vnl_math_sqr(centered_point[0]) + vnl_math_sqr(centered_point[1]) +
          vnl_math_sqr(centered_point[2]-center_point_measured_distance);
        if ( !point_grid(row,col).is_valid_return )
          use_point = false;
        else if ( //vnl_math_abs(row-g_row) > 1 && vnl_math_abs(col-g_col) > 1 &&
                  sqr_point_distance > sqr_distance_limit )
          use_point = false;
        else if ( !point_grid(row,col).is_done )
          use_point = true;
        else if ( ignore_completed_points )
          use_point = false;
        else if ( use_completed_only_when_consistent && point_grid(row,col).is_done
                  && point_grid(row,col).surface_type == PointRecord::PRSurface )
          {
            centered_normal = centering_rotation * point_grid(row,col).surface_normal;
            if ( vnl_math_abs(centered_normal[2]) < 1.0e-4)    //  safety check for a situation that should occur at most rarely
              { use_point = false; vcl_cout << "e: "; }
            else
              {
                double distance_to_plane_along_los = dot_product(centered_normal,centered_point) / centered_normal[2];
                use_point = ( vnl_math_abs(distance_to_plane_along_los - center_point_measured_distance) 
                              < (m_max_allowed_movement_scale_multiplier+1) * point_grid(row,col).estimated_scale );
              }
          }
        else if ( point_grid(row,col).is_done && 
                  point_grid(row,col).surface_type == PointRecord::PROutlier )
          use_point = false;
        else
          use_point = true;

        //        vcl_cout << "  " << vnl_int_2( row - g_row, col - g_col ) << ":  " << centered_point << ", use = " << use_point << vcl_endl;
        if ( use_point )
          {
            centered_points.push_back( centered_point );
            grid_locations.push_back( vnl_int_2(row,col) );
            // vcl_cout << "gather_and_center_points:  centered_point " << centered_point << vcl_endl;
            // assert( vnl_vector_ssd( point_grid(row,col).measured_distance * point_grid(row,col).direction, 
            //                      uncenter_rotation * centered_point ) < 1.0e-10 );
          }
      }
}
 

//  Check the status of the priority queue.  This is used only for debugging.
void 
LocationsAndNormals::check_pqueue( vbl_array_2d<PointRecord> const& point_grid,
                                   vcl_vector< vnl_int_2 > const& pqueue )
{
  if ( pqueue.size() == 0 ) return;

  for ( unsigned int i=(unsigned int)pqueue.size()-1; i>0; i-- )
    {
      int par = (i-1)/2;
      if ( point_grid( pqueue[par][0], pqueue[par][1] ).priority <
           point_grid( pqueue[i][0], pqueue[i][1] ).priority )
        {
          vcl_cout << "PQ:  Failed in comparison against parent at location " << i << vcl_endl;
          assert(false);
        }
    }
}


//  Add a grid location that is ready for estimation to the priority
//  queue.  The priority queue is just a vector of these locations.
//  All information needed for priority queue operations is stored in
//  the grid.
void
LocationsAndNormals::add_to_pqueue( vbl_array_2d<PointRecord> & point_grid,
                                    vcl_vector< vnl_int_2 > & pqueue,
                                    vnl_int_2 & grid_loc )
{
  pqueue.push_back( grid_loc );
  point_grid( grid_loc[0], grid_loc[1] ).pq_index = (int)pqueue.size()-1;
  percolate_up( point_grid, pqueue, grid_loc );
}


//  Remove the top grid location from the priority queue.  Swap in
//  the last element of the priority queue and run the standard
//  percolate down operation.

void
LocationsAndNormals::remove_from_pqueue( vbl_array_2d<PointRecord> & point_grid,
                                         vcl_vector< vnl_int_2 > & pqueue )
{
  int pq_index = 0; 
  vnl_int_2 last_grid_loc = pqueue.back();
  double insert_priority = point_grid( last_grid_loc[0], last_grid_loc[1] ).priority;
  pqueue.pop_back();
  bool done = false;

  if( pqueue.empty() ) 
    return;

  while ( pq_index <= int(pqueue.size())/2-1 && !done )
    {
      int lchild = 2*pq_index+1;
      int rchild = 2*pq_index+2;
      int child;

      //  If the right child does not exist or if it does and the left
      //  child has the higher priority, use the left child.
      if ( rchild >= pqueue.size() ||
           point_grid( pqueue[lchild][0], pqueue[lchild][1] ).priority >
           point_grid( pqueue[rchild][0], pqueue[rchild][1] ).priority )
        child = lchild;
      else
        child = rchild;

      //  If the child has lower priority, we're done.  Otherwise,
      //  move the child up and consider the child index as the next
      //  possible location.
      vnl_int_2 child_grid_loc = pqueue[child];
      if ( insert_priority >= point_grid( child_grid_loc[0], child_grid_loc[1] ).priority )
        done = true;
      else
        {
          point_grid( child_grid_loc[0], child_grid_loc[1] ).pq_index = pq_index;
          pqueue[pq_index] = child_grid_loc;
          pq_index = child;
        }
    }

  //  Drop the grid location into the right index in the priority queue.
  pqueue[pq_index] = last_grid_loc;
  point_grid( last_grid_loc[0], last_grid_loc[1] ).pq_index = pq_index;

  //  Debug call
  // check_pqueue( point_grid, pqueue );
}


// Run the classic percolate up operation on the priority queue of
// grid locations.
void
LocationsAndNormals::percolate_up( vbl_array_2d<PointRecord> & point_grid,
                                   vcl_vector< vnl_int_2 > & pqueue,
                                   vnl_int_2 & grid_loc )
{
  //  Get the priority queue index from the grid entry.
  int pq_index = point_grid( grid_loc[0], grid_loc[1] ).pq_index;

  //  Quit if it is already at the top of the priority queue.
  if ( pq_index == 0 ) return;

  //  Find the parent pqueue index and grid location.
  int parent_pq_index = (pq_index-1)/2;
  vnl_int_2 parent_grid_loc = pqueue[parent_pq_index];

  //  Quit when we've reached the top or the child has lower priority
  //  than the parent.
  while ( pq_index > 0 && 
          point_grid( grid_loc[0], grid_loc[1] ).priority 
          > point_grid( parent_grid_loc[0], parent_grid_loc[1] ). priority )
    {
      point_grid( parent_grid_loc[0], parent_grid_loc[1]) . pq_index = pq_index;
      pqueue[ pq_index ] = parent_grid_loc;
      pq_index = parent_pq_index;
      parent_pq_index = (pq_index-1)/2;
      parent_grid_loc = pqueue[ parent_pq_index ];
    }

  //  Drop the grid location into the right index in the priority queue.
  point_grid( grid_loc[0], grid_loc[1] ).pq_index = pq_index;
  pqueue[pq_index] = grid_loc;

  //  Debug call
  // check_pqueue( point_grid, pqueue );
}

