#include "MatchPSK.h"

#include "vtkSmartPointer.h"
#include "vtkPolyData.h"
#include "vtkPointData.h"
#include "vtkXMLPolyDataWriter.h"
#include "vtkCellArray.h"
#include "vtkIntArray.h"
#include "vtkLine.h"

#include <rrel/rrel_cauchy_obj.h>
#include <rsdl/rsdl_point.h>
#include <rsdl/rsdl_kd_tree.h>

#include <vcl_set.h>
#include <vcl_algorithm.h>
#include <vcl_map.h>

#define RATIO_CUT 0.8

//*****************************************************************************

bool matches_less(const pskmatch &l, const pskmatch &r)
{
  return l.sortby < r.sortby;
}

//*****************************************************************************

void ComputeXFormsPSK(Scan *moving_scan, Scan *fixed_scan, vcl_vector<SimpleRigid> &xforms, vcl_vector<pskmatch> &matches)
{
  const KeyVec &movingkpts = moving_scan->GetPSKs();
  const KeyVec &fixedkpts = fixed_scan->GetPSKs();

  if (matches.empty()) {
    MatchPSK(fixedkpts, movingkpts, matches);
    RangeMatchPSK2(fixed_scan, moving_scan, matches);  

    //Use learned sorting parameters
    const int D = 3;
    vnl_vector<double> a(D);
    a[0] = 0.00212957;
    a[1] = -0.0066905;
    a[2] = 0.00294268;

    for (unsigned int i = 0; i < matches.size(); i++)
    {
      const pskmatch &m = matches[i];
      if (m.ratio > RATIO_CUT)
        matches[i].sortby = 1e6 + m.ratio;
      else {
        vnl_vector<double> v(D);
        v[0] = 1.0;
        v[1] = m.ratio;
        v[2] = m.range_error; 

        matches[i].sortby = -dot_product(a,v);
      }
    }
  
    vcl_sort(matches.begin(), matches.end(), matches_less);
  }

  for (unsigned int i = 0; i < matches.size(); i++)
  {
    SimpleRigid xform;
    ComputePSKXForm(matches[i].f, matches[i].m, xform);
    xforms.push_back(xform);
  }
}

//*****************************************************************************

void MatchPSK(const KeyVec &fixed, const KeyVec &moving, vcl_vector<pskmatch> &matches)
{
  vcl_cout << "Matching keypoints...\n";
  for (unsigned int s = 0; s < fixed.size(); s++)
  {
    if (fixed[s].empty() || moving[s].empty())
      continue;

    vcl_vector<rsdl_point> kdpts;

    for(unsigned int i = 0; i < fixed[s].size(); i++)
    {
      rsdl_point point(fixed[s][i]->feature_vector);
      kdpts.push_back(point);
    }

    rsdl_kd_tree kdtree(kdpts);

    for (unsigned int i = 0; i < moving[s].size(); i++)
    {
      vcl_vector<rsdl_point> pts;
      vcl_vector<int> inds;
      rsdl_point pt(moving[s][i]->feature_vector);
      kdtree.n_nearest(pt, 2, pts, inds, true, 200);
      if (inds.size() == 2)
      {
        // Compute the ratio of distance to closest and second closest matches.
        double ratio = 1.0;

        PSKpt *first = fixed[s][inds[0]];
        PSKpt *second = fixed[s][inds[1]];
        double dist_to_first = (first->feature_vector - moving[s][i]->feature_vector).two_norm();
        double dist_to_second = (second->feature_vector - moving[s][i]->feature_vector).two_norm();

        ratio = dist_to_first/dist_to_second;
        
        pskmatch m;
        m.m = moving[s][i];
        m.f = first;
        m.b = second;
        m.ratio = ratio;
        m.index = i;

        matches.push_back(m);
      }
    }
  }  
}

//*****************************************************************************

void RangeMatchPSK2(const Scan *scan1, const Scan *scan2, vcl_vector<pskmatch> &matches)
{
  vcl_cout << "Range match2!\n";

  const vcl_vector<vnl_double_3> &fnormals = scan1->GetRangeData()->get_normals();
  const vcl_vector<vnl_double_3> &mnormals = scan2->GetRangeData()->get_normals();
  
  const vcl_vector<vnl_double_3> &fpts = scan1->GetRangeData()->get_points();
  const vcl_vector<vnl_double_3> &mpts = scan2->GetRangeData()->get_points();

  size_t end = fpts.size();  
  
  vcl_vector<rsdl_point> kdpts;
  kdpts.resize(mpts.size());
  for(unsigned int i = 0; i < mpts.size(); i++)
    kdpts[i] = rsdl_point(mpts[i]);
  
  rsdl_kd_tree *kdtree = new rsdl_kd_tree(kdpts);

  double std1 = scan1->GetRangeData()->getStdDev();
  double std2 = scan2->GetRangeData()->getStdDev();
  double std = sqrt(std1*std1 + std2*std2);

  double radius = 0.75;

  vcl_cout << "C = " << 3.0*std << "\n";
  rrel_cauchy_obj cauchy(3.0*std); 
  for (unsigned int i = 0; i < matches.size(); i++)
  {
    pskmatch &m = matches[i];
    if (m.ratio > 0.8)  {
      m.range_error = 0.0;
    }

    //xform: moving -> fixed
    //ixform: fixed -> moving
    SimpleRigid xform, ixform;
    ComputePSKXForm(m.f, m.m, xform);
    ixform = xform.inverse();

    vcl_vector<rsdl_point> pts;
    vcl_vector<int> indi;

    kdtree->points_in_radius(rsdl_point(m.m->location), radius, pts, indi);

    vcl_vector<int> moving_set;

    vnl_double_3 moving_center(0.0,0.0,0.0);
    vnl_double_3 fixed_center(0.0,0.0,0.0);

    vnl_double_3 mapped_loc = m.m->location;

    double error = 0.0;
    vcl_set<int> fixed_set;
    double sgauss = 0.0;
    for (unsigned int j = 0; j < indi.size(); j++)
    {
      const vnl_double_3 &mpt = mpts[indi[j]];

      vnl_double_3 mapped_pt = xform.map_point(mpt);

      if (dot_product((ixform.m_t - mpt).normalize(), mnormals[indi[j]]) < 0.1) {
        continue;
      }

      unsigned int index = scan1->GetRangeData()->findClosestTo(mapped_pt);
      if (index == end || (fpts[index] - m.f->location).two_norm() > radius)
        continue;

      if (dot_product((xform.m_t - fpts[index]).normalize(), fnormals[index]) < 0.1)  {
        continue;
      }

      if (fixed_set.insert(index).second)
        fixed_center += fpts[index];

      moving_set.push_back(indi[j]);
      moving_center += mpt;

      double d = fabs(dot_product(mapped_pt - fpts[index], fnormals[index]));
      double w = cauchy.wgt(d);
      error += w;
      sgauss += 1.0;
    }

    if (fixed_set.empty())  {
      m.range_error = 0.0;
      continue;
    }

    moving_center /= (double)moving_set.size();
    fixed_center /= (double)fixed_set.size();

    double mflat = 0.0;
    for (unsigned int j = 0; j < moving_set.size(); j++)  {
      double d = dot_product(mpts[moving_set[j]] - moving_center, m.m->normal);
      mflat += d*d;
    }
    mflat /= (double)moving_set.size();

    double fflat = 0.0;
    for (vcl_set<int>::const_iterator itr = fixed_set.begin(); itr != fixed_set.end(); itr++)  {
      double d = dot_product(fpts[*itr] - fixed_center, m.f->normal);
      fflat += d*d;
    }
    fflat /= (double)fixed_set.size();

    matches[i].range_error = error / sgauss;
  }

  delete kdtree; 
}

//*****************************************************************************

void ComputePSKXForm(const PSKpt *kFixed, const PSKpt *kMoving, SimpleRigid& xform)
{
  // Construct fixed coordinate system.
  vnl_vector_fixed<double, 3> Of = kFixed->location;
  vnl_matrix_fixed<double, 3, 3> Rf;
  Rf.set_row(0, kFixed->direction);
  Rf.set_row(1, vnl_cross_3d(kFixed->normal, kFixed->direction));
  Rf.set_row(2, kFixed->normal);

  // Construct moving coordinate system.
  vnl_vector_fixed<double, 3> Om = kMoving->location;
  vnl_matrix_fixed<double, 3, 3> Rm;
  Rm.set_row(0, kMoving->direction);
  Rm.set_row(1, vnl_cross_3d(kMoving->normal, kMoving->direction));
  Rm.set_row(2, kMoving->normal);

  // Compute the transformation to align coordinate systems.
  vnl_matrix_fixed<double, 3, 3> R = Rf.transpose()*Rm;
  vnl_vector_fixed<double, 3> t = Of - R*Om;

  xform = SimpleRigid(R, t);
}

//*****************************************************************************

//Requires both fixed and moving kpts to have been written
void WritePSKMatches(const vcl_vector<pskmatch> &matches, const KeyVec &fixedkpts, const char *filename)
{
  FILE *file = fopen(filename, "wb");
  unsigned int size = (unsigned int)matches.size();
  vcl_map<double, int> scale_map;
  for (int s = 0; s < fixedkpts.size(); s++)
  {
    double scale;
    if (fixedkpts[s].empty())
      scale = -1.0;
    else
      scale = fixedkpts[s][0]->scale;
    scale_map[scale] = s;
  }

  fwrite(&size, sizeof(unsigned int), 1, file);
  for (unsigned int i = 0; i < size; i++)
  {
    int sindex = scale_map[matches[i].f->scale];
    fwrite(&sindex, sizeof(int), 1, file);
    fwrite(&matches[i].f->id, sizeof(int), 1, file);
    fwrite(&matches[i].m->id, sizeof(int), 1, file);
    fwrite(&matches[i].b->id, sizeof(int), 1, file);
    fwrite(&matches[i].ratio, sizeof(double), 1, file);
    fwrite(&matches[i].range_error, sizeof(double), 1, file);
    fwrite(&matches[i].index, sizeof(int), 1, file);
    fwrite(&matches[i].sortby, sizeof(double), 1, file);
  }

  fclose(file);
  vcl_cout << "Wrote " << size << " matches to " << filename << vcl_endl;
}

//*****************************************************************************

bool ReadPSKMatches(vcl_vector<pskmatch> &matches, const KeyVec &fixedkpts, const KeyVec &movingkpts, const char *filename)
{
  FILE *file = fopen(filename, "rb");
  if (!file)
    return false;

  unsigned int size;
  fread(&size, sizeof(unsigned int), 1, file);
  matches.resize(size);
  for (unsigned int i = 0; i < size; i++)
  {
    int sindex;
    fread(&sindex, sizeof(int), 1, file);
    int fid, mid, bid;
    fread(&fid, sizeof(int), 1, file);
    fread(&mid, sizeof(int), 1, file);
    fread(&bid, sizeof(int), 1, file);
    matches[i].f = fixedkpts[sindex][fid];
    matches[i].m = movingkpts[sindex][mid];
    matches[i].b = fixedkpts[sindex][bid];

    fread(&matches[i].ratio, sizeof(double), 1, file);
    fread(&matches[i].range_error, sizeof(double), 1, file);
    fread(&matches[i].index, sizeof(int), 1, file);
    fread(&matches[i].sortby, sizeof(double), 1, file);
  }

  fclose(file);
  vcl_cout << "Read " << size << " matches from " << filename << vcl_endl;
  return true;
}

//*****************************************************************************