#include "PSKgraph.h"

#include "Scan.h"

#include "vtkSmartPointer.h"
#include "vtkPolyData.h"
#include "vtkPointData.h"
#include "vtkXMLPolyDataWriter.h"
#include "vtkCellArray.h"
#include "vtkIntArray.h"
#include "vtkLine.h"
#include "vtkFloatArray.h"
#include "vtkDelaunay2D.h"

#include "vtkAppendPolyData.h"
#include "vtkSampleFunction.h" 
#include "vtkContourFilter.h"
#include "vtkImageData.h"
#include "vtkTransformPolyDataFilter.h"
#include "vtkTransform.h"

#include <vul/vul_file.h>
#include <vnl/vnl_random.h>

#include <vbl/vbl_array_2d.h>

#include <vcl_iostream.h>
#include <vcl_algorithm.h>
#include <vcl_list.h>
#include <vcl_map.h>

#include <vnl/algo/vnl_svd.h>
#include <vnl/vnl_det.h>
#include <vnl/vnl_trace.h>
#include <vnl/vnl_inverse.h>
#include <vnl/algo/vnl_symmetric_eigensystem.h>

#include <rsdl/rsdl_bins.h>
#include "utilities.h"

#define VERBOSE

#define MINPEAKHEIGHT 0.03
#define PEAKHEIGHTFRACTION 0.1
#define DESCRIPTOR_COMPONENT_MAXIMUM 0.2

const double _double2fixmagic = 68719476736.0*1.5;     //2^36 * 1.5,  (52-_shiftamt=36) uses limited precisicion to floor
const int _shiftamt        = 16;                       //16.16 fixed point representation,

#if BigEndian_
  #define iexp_     0
  #define iman_     1
#else
  #define iexp_     1
  #define iman_     0
#endif //BigEndian_

const double two_pi = 2.0*vnl_math::pi;
const double bilateral = 0.4; 

#define SAVE_SCALES 0

//*****************************************************************************

inline int Real2Int(double val)
{
  val    = val + _double2fixmagic;
  return ((int*)&val)[iman_] >> _shiftamt; 
}

//*****************************************************************************

//Build PSKGraph from image base
PSKGraph::PSKGraph(Scan &scan, int image_num, const vcl_vector<double> &scales, int left, int right, int top, int bottom, SimpleRigid *xform_, bool load_)
{
  load = load_;
  xform = xform_;
  graph_file = NULL;
  kdtree = NULL;
  controlkd = NULL;
  name = vul_file::strip_directory(scan.GetFullScanName());
  const vil_image_view<double> &img = scan.GetImages()[image_num];

  bool loaded = false;
  if (load) {
    graph_file = fopen((scan.GetFullScanName() + "imesh.dat").c_str(), "rb");
    if (graph_file) {
      LoadPSKGraph();
      loaded = true;
    }
  } 
  if (!loaded) {  
    if (left == -1 && right == -1 && top == -1 && bottom == -1)
    {
      left = 0;    top = 0;
      right = (int)img.ni();
      bottom = (int)img.nj();
    }

    int width = right - left;
    int height = bottom - top;

    vbl_array_2d< int > grid(width, height);

    int numpts = 0;
    int a = 0;
    for (int i = left; i < right; i++, a++)
    {
      int b = 0;
      for (int j = top; j < bottom; j++, b++)
      {
        vnl_double_2 pt2d(i+0.5,j+0.5);
        vnl_double_3 pt3d, worldpt, camorigin = scan.GetCameras()[image_num]->get_origin();
        scan.GetCameras()[image_num]->get_xyz(pt2d, pt3d);
        unsigned int ptid;
        if (!scan.GetRangeData()->backProject(pt3d, camorigin, worldpt, &ptid)) {
                  grid.put(a, b, -1);
          continue;
        }

        PSKNode *n = new PSKNode;
        n->normal = scan.GetRangeData()->get_normals()[ptid];
        vnl_double_3 d = (pt3d - camorigin).normalize();
        double dot = dot_product(d, n->normal);
        double dot_center_norm = dot_product((camorigin - worldpt), n->normal);
        n->point = (camorigin - (dot_center_norm/dot)*d);
        n->index = numpts++;
        n->imgindex1 = i; 
        n->imgindex2 = j;
        n->intensity = ((double)img(i,j))/255.0;
        nodes.push_back(n);
        grid.put(a, b, n->index);
      }
    }

    if (numpts < 50) {
      for (unsigned int i = 0; i < nodes.size(); i++)
        delete nodes[i];
      nodes.clear();
      vcl_cout << "Too few points!\n";
      return;
    }

  #ifdef VERBOSE
    vcl_cout << "Connecting Triangles\n";
  #endif

#ifdef NO_DELAUNAY
    int tricount = 0;
    for (int i = 0; i < width-1; i++)
    {
      for (int j = 0; j < height-1; j++)
      {
        int index = grid(i,j);
        if (index == -1) continue;     
        int right = grid(i+1,j), down = grid(i,j+1), diag = grid(i+1,j+1);
        if (diag == -1 && down != -1 && right != -1) {
          PSKTriangle t(nodes[index], nodes[down], nodes[right]);
          ConnectTriangle(t, tricount++);
          triangles.push_back(t);
        }
        else 
        {
          if (right != -1 && diag != -1) {
            PSKTriangle t(nodes[index], nodes[diag], nodes[right]);
            ConnectTriangle(t, tricount++);
            triangles.push_back(t);
          }
          if (down != -1 && diag != -1) {
            PSKTriangle t(nodes[index], nodes[down], nodes[diag]);
            ConnectTriangle(t, tricount++);
            triangles.push_back(t);
          }
        }
      }
    }
#else
    Delaunay(0.25);
#endif

    #ifdef VERBOSE
      vcl_cout << name << ": Successfully projected " << nodes.size() << " points of " << height*width << vcl_endl;
    #endif

#if SAVE_SCALES
    graph_file = fopen((scan.GetFullScanName() + "imesh.dat").c_str(), "wb");
    SavePSKGraph();
#endif
  }

  IsManifold();
  if (xform)
    this->ApplyRigidXForm(*xform);

  vcl_cout << "Num pts: " << nodes.size() << " Num triangles: " << triangles.size() << "\n";

  ComputeMinScale(scales);
  scale = minscale;

  vcl_vector<rsdl_point> kdpts(nodes.size());
  for(unsigned int i = 0; i < nodes.size(); i++)  
    kdpts[i] = rsdl_point(nodes[i]->point);  
  kdtree = new rsdl_kd_tree(kdpts); 
  
  if (load) {
    LoadNextScale();
  }
  else {
    control.resize(nodes.size());
    for (unsigned int i = 0; i < control.size(); i++) control[i] = i;
    vcl_vector<rsdl_point> ctrlkdpts(control.size());
    for(unsigned int i = 0; i < control.size(); i++)
      ctrlkdpts[i] = rsdl_point(nodes[control[i]]->point);
    controlkd = new rsdl_kd_tree(ctrlkdpts);

    Smooth(scale);
#if SAVE_SCALES
    SaveScale(); 
#endif
    DownSampleControl(scale/2.0);
  }
}

//*****************************************************************************

PSKGraph::~PSKGraph()
{
  for (unsigned int i = 0; i < nodes.size(); i++)
    delete nodes[i];
  if (kdtree)     delete kdtree;
  if (controlkd)  delete controlkd;
  if (graph_file) fclose(graph_file);
}

//*****************************************************************************

double PSKGraph::ComputeMinScale(const vcl_vector<double> &scales)
{
  vcl_vector<double> edgelens;
  for (unsigned int i = 0; i < triangles.size(); i++)
  {
    const PSKTriangle &t = triangles[i];
    if (t.a->index > t.b->index)
      edgelens.push_back((t.a->point - t.b->point).two_norm());
    if (t.a->index > t.c->index)
      edgelens.push_back((t.a->point - t.c->point).two_norm());
    if (t.b->index > t.c->index)
      edgelens.push_back((t.b->point - t.c->point).two_norm());
  }

  double min = edgelens[0];
  double max = min;
  for (unsigned int i = 1; i < edgelens.size(); i++)
  {
    if (edgelens[i] < min)
      min = edgelens[i];
    if (edgelens[i] > max)
      max = edgelens[i];
  }

  // Build a histogram of the responses.
  int nbins = 1000;
  vnl_vector<int> histogram(nbins, 0);
  double z = double(nbins)/(max-min);
  for (unsigned int i = 1; i < edgelens.size(); i++)
  {
    double v = edgelens[i];
    int bin = int((v-min)*z);
    if(bin > nbins-1) { bin = nbins-1; }
    ++histogram(bin);
  }

  // Convert to a cumulative histogram.  Also compute the thresholds
  // for the bottom and top fraction ("percentile") of the data.
  int edges = (int)edgelens.size();
  double total = 0.0f;
  double lower = max;
  double upper = min;
  for (int i = 0; i < nbins; i++)
  {
    if(lower == max && total/edges >= .5)
      lower = i/z + min;      
    total += histogram(i);
    histogram(i) = total;
  }

  minscale = scales[0];
  minscaleindex = 0;
  double closestd = fabs(scales[0] - lower);
  for (unsigned int i = 0; i < scales.size(); i++)
    if (fabs(scales[i] - lower) < closestd) {
      minscale = scales[i];
      minscaleindex = i;
      closestd = fabs(scales[i] - lower);
    }

   vcl_cout << "median edge length: " << lower << "\nmin scale: " << minscale << " at index " << minscaleindex << "\n";

  return minscale;
}

//*****************************************************************************

void PSKGraph::Delaunay(double scale)
{
#ifdef VERBOSE
  vcl_cout << "Starting delaunay...\n";
#endif
  vtkSmartPointer<vtkPoints> pts = vtkSmartPointer<vtkPoints>::New(); 
  for (unsigned int i = 0; i < nodes.size(); i++)
  {
    pts->InsertNextPoint(nodes[i]->imgindex1, nodes[i]->imgindex2, 0.0);
  }

  vtkSmartPointer<vtkPolyData> polydata = vtkSmartPointer<vtkPolyData>::New();
  polydata->SetPoints(pts);
  vtkSmartPointer<vtkDelaunay2D> delaunay = vtkSmartPointer<vtkDelaunay2D>::New();
  delaunay->SetInput(polydata);
  delaunay->Update();
  vtkPolyData *output = delaunay->GetOutput();

#ifdef VERBOSE
  vcl_cout << "Finished Delaunay\n";
#endif
  double thresh = 3.0*scale;
  int count = 0;
  for (int i = 0; i < output->GetNumberOfCells(); i++)
  {
    PSKTriangle t;
    t.a = nodes[output->GetCell(i)->GetPointId(0)];
    t.b = nodes[output->GetCell(i)->GetPointId(1)];
    t.c = nodes[output->GetCell(i)->GetPointId(2)];

    double ablen = (t.a->point - t.b->point).two_norm();
    double aclen = (t.a->point - t.c->point).two_norm();
    double bclen = (t.b->point - t.c->point).two_norm();
    
    if (ablen > thresh || aclen > thresh || bclen > thresh)
      continue;

    double s = (ablen + aclen + bclen)/2.0;
    t.area = sqrt(s*(s-ablen)*(s-aclen)*(s-bclen));

    t.a->triangles.push_back(count);
    t.b->triangles.push_back(count);
    t.c->triangles.push_back(count);
    
    t.a->adjacent.push_back(t.b);
    t.a->adjacent.push_back(t.c);

    t.b->adjacent.push_back(t.a);
    t.b->adjacent.push_back(t.c);

    t.c->adjacent.push_back(t.a);
    t.c->adjacent.push_back(t.b);

    triangles.push_back(t);
    count++;
  }
#ifdef VERBOSE
  vcl_cout << count << " " << triangles.size() << "\n";
#endif
}


vcl_string PSKGraph::MakeName(double scale)
{
  char buf[50];
  sprintf(buf, "%s-%.3f.scale", name.c_str(), scale);
  return vcl_string(buf);
}

//*****************************************************************************

//Decimate the pskgraph to the new scale and the control pts
bool PSKGraph::DecimateToScale(double newscale)
{
  if (newscale <= scale)  {
    vcl_cout << "skipping scale " << newscale << "\n";
    return false;
  }
  
#ifdef VERBOSE
    vcl_cout << "\nDecimating to scale: " << newscale << "\n";
#endif
  double diff = sqrt(newscale*newscale - scale*scale);
  if (load) {
    LoadNextScale();
  }
  else {
    Smooth(scale);  
    scale = newscale;
#if SAVE_SCALES
    SaveScale();  
#endif
    DownSampleControl(scale/2.0); 
    if (control.size() < 10)
      return false;
  }
  
  return true;
}

//*****************************************************************************

//Smooth the intensities and normals using bilateral filter
void PSKGraph::Smooth(double s) 
{
  double dist = s*2.0;
  double gaussian_exp_denom = 2.0f*s*s;
  double bilateral_exp_denom = 2.0f*bilateral*bilateral;

#ifdef VERBOSE
  vcl_cout << "Support: " << dist << "\n";
#endif

  vcl_vector<double> vals(nodes.size());
  vcl_vector<vnl_double_3> normals(nodes.size());
  
  unsigned int avgpts = 0;
  for (unsigned int i = 0; i < nodes.size(); i++)
  {
#ifdef VERBOSE
    if (!(i%20000))  printf("%d%% ",(int)(100*(i/(double)nodes.size())));
#endif
    PSKNode *n = nodes[i];
    vcl_vector<PSKNode *> locs;
    GetControlPointsInRange(n->point, dist, locs);
    avgpts += (unsigned int)locs.size();
    double newcolor = 0.0f, sgauss = 0.0;
    vnl_double_3 newpoint(0.0,0.0,0.0), newnormal(0.0,0.0,0.0);
    for (unsigned int m = 0; m < locs.size(); m++)
    {      
      double ssd = (locs[m]->point - n->point).squared_magnitude();
      double nd = 1.0 - dot_product(locs[m]->normal, n->normal); 
      double wgauss = exp(-ssd/gaussian_exp_denom)* exp(-nd*nd/bilateral_exp_denom);
      sgauss += wgauss;
      newcolor += wgauss * locs[m]->intensity;
      newnormal += wgauss * locs[m]->normal;
    }  

    vals[i] = newcolor/sgauss;    
    normals[i] = newnormal.normalize();
  }  
 
  for (unsigned int i = 0; i < nodes.size(); i++)
  {
    nodes[i]->intensity = vals[i];
    nodes[i]->normal = normals[i];
  }

#ifdef VERBOSE
  vcl_cout << "\nFinished Smoothing\n";
  vcl_cout << "Average neighborhood size: " << (double)avgpts/(double)nodes.size() << "\n";
#endif
}

//*****************************************************************************

void PSKGraph::DownSampleControl(double thresh)
{
  if (control.size() < 100)
    return;

#ifdef VERBOSE
  vcl_cout << "Downsampling to " << thresh << " meters.\n";
#endif
  vnl_vector_fixed<double,3> min_point(1e6);
  vnl_vector_fixed<double,3> max_point(-1e6);
  for (unsigned int i = 0; i < control.size(); i++)
  {
    vnl_double_3 const& loc = nodes[control[i]]->point;
    for (int k = 0; k < 3; k++ )
    {
      if (loc[k] < min_point[k] ) 
        min_point[k] = loc[k];
      else if ( loc[k] > max_point[k] )
        max_point[k] = loc[k];
    }
  }

  //  Assign the bin sizes to be isotropic.  
  double bsize = 0;
  const int max_bins_per_dim = 50;
  for ( unsigned int i = 0; i < 3; i++ )
  {
    double lower_bound_on_size = (max_point[i] - min_point[i]) / max_bins_per_dim;
    if ( bsize < lower_bound_on_size ) 
      bsize = lower_bound_on_size;
  }

  #ifdef VERBOSE
  vcl_cout << "Creating bins...\n";
  #endif

  vnl_vector_fixed<double,3> bin_sizes( bsize );
  rsdl_bins<3, double, int > bins(min_point, max_point, bin_sizes);

  vcl_vector<int> sampled_nodes;
  int count = 0;
  for (unsigned int i = 0; i < control.size(); i++)
  {
    PSKNode *n = nodes[control[i]];
    vnl_vector_fixed<double,3> loc;
    loc[0] = n->point[0];  loc[1] = n->point[1];  loc[2] = n->point[2];
    if (!bins.is_any_point_within_radius(loc, thresh))
    {
      sampled_nodes.push_back(control[i]);
      bins.add_point(loc, i);
    }
  }

#ifdef VERBOSE
  vcl_cout << "Downsampling culled " << control.size() - sampled_nodes.size() << " of " << control.size() << " control points.\n";
#endif
  control = sampled_nodes;

  //Not the most efficient...
  delete controlkd;
  vcl_vector<rsdl_point> ctrlkdpts(control.size());
  for(unsigned int i = 0; i < control.size(); i++)
    ctrlkdpts[i] = rsdl_point(nodes[control[i]]->point);
  controlkd = new rsdl_kd_tree(ctrlkdpts);
}

//*****************************************************************************

void PSKGraph::ComputeGradients(bool project)
{
  //Compute PSKTriangle gradients
  for (unsigned int i = 0; i < triangles.size(); i++)
  {
    const vnl_double_3 &a = triangles[i].a->point;
    const vnl_double_3 &b = triangles[i].b->point;
    const vnl_double_3 &c = triangles[i].c->point;

    triangles[i].gradient = (1.0/(4.0*triangles[i].area * triangles[i].area)) *
      ((double)triangles[i].a->intensity*(dot_product(a - b, b - c)*(c - a) + dot_product(a - c, c - b)*(b - a))
     + (double)triangles[i].b->intensity*(dot_product(b - a, a - c)*(c - b) + dot_product(b - c, c - a)*(a - b))
     + (double)triangles[i].c->intensity*(dot_product(c - b, b - a)*(a - c) + dot_product(c - a, a - b)*(b - c)));
  }

  //Average triangle gradients
  for (unsigned int i = 0; i < nodes.size(); i++)
  {
    PSKNode *n = nodes[i];
    double areasum = 0.0;
    vnl_double_3 gradient(0.0,0.0,0.0);
    for (unsigned int j = 0; j < n->triangles.size(); j++)
    {
      gradient += triangles[n->triangles[j]].area * triangles[n->triangles[j]].gradient;
      areasum += triangles[n->triangles[j]].area;
    }
    
    if (n->triangles.empty())
      n->gradient = gradient;
    else
      n->gradient = gradient/areasum;
  }

  //project gradient into the normal plane, this is done after computing LBO but can be forced here
  if (project) {
    for (unsigned int i = 0; i < nodes.size(); i++)
    {
      PSKNode *n = nodes[i];
      n->gradient = n->gradient - dot_product(n->gradient, n->normal)*n->normal;
    }
  }
}

//*****************************************************************************

void PSKGraph::ComputeLaplaceBeltrami()
{
  for (unsigned int i = 0; i < nodes.size(); i++)
  {
    PSKNode *n = nodes[i];
    if (!n->manifold) {
      n->laplacian = 0.0;
      continue;
    }

    double areasum = 0.0, laplacian = 0.0;
    for (unsigned int j = 0; j < n->triangles.size(); j++)
    {
      //rotate the lables such that a is the current node
      PSKTriangle t = triangles[n->triangles[j]];
      if (t.b == n) {vcl_swap(t.a, t.b); vcl_swap(t.b, t.c);}
      if (t.c == n) {vcl_swap(t.a, t.c); vcl_swap(t.b, t.c);}
      const vnl_double_3 &a = t.a->point;
      const vnl_double_3 &b = t.b->point;
      const vnl_double_3 &c = t.c->point;

      vnl_double_3 normal = -(dot_product(a-b,b-c)*(c-a) + dot_product(a-c,c-b)*(b-a))/(2.0*t.area);
      laplacian += dot_product(normal, t.b->gradient + t.c->gradient);
      areasum += t.area;
    }

    if (n->triangles.empty())
      n->laplacian = 0.0;
    else
      n->laplacian = scale * scale * laplacian/(2.0*areasum);
  }

  for (unsigned int i = 0; i < nodes.size(); i++)
  {
    PSKNode *n = nodes[i];
    n->gradient = n->gradient - dot_product(n->gradient, n->normal)*n->normal;
  }
}

//*****************************************************************************

void PSKGraph::IsManifold()
{
  //To be "manifold" a vertex must have a full ring around it, this can be tested
  //by checking to make sure 2 of each edge exist around the vertex
  for (unsigned int i = 0; i < nodes.size(); i++)
  {
    PSKNode *n = nodes[i];
    vcl_map<PSKNode *, int> count;
    for (unsigned int j = 0; j < n->triangles.size(); j++)
    {
      const PSKTriangle &t = triangles[n->triangles[j]];
      count[t.a] = count[t.a]+1;
      count[t.b] = count[t.b]+1;
      count[t.c] = count[t.c]+1;
    }

    count.erase(n);
    n->manifold = true;
    for (vcl_map<PSKNode *, int>::const_iterator itr = count.begin(); itr != count.end(); itr++)
    {
      if (itr->second != 2) {
        n->manifold = false;
        break;
      }
    }
  }
}

//*****************************************************************************

void PSKGraph::DetectExtrema(vcl_vector<PSKpt *> &keypoints)
{
  ComputeGradients();
  ComputeLaplaceBeltrami();

  vcl_vector<PSKpt *> extrema;
#ifdef VERBOSE
  vcl_cout << "computing kpts for scale " << scale;
#endif
  double upper, lower;
  ComputeLaplacianBounds(lower, upper);

  for (unsigned int i = 0; i < nodes.size(); i++)
  { 
    PSKNode *n = nodes[i];
    if (!n->manifold)
      continue;

    //Make sure peak lies on the outter limits of the LBO responses
    if(n->laplacian >= lower && n->laplacian <= upper)
        continue; 
    
    vcl_vector<PSKNode *> good;
    for (unsigned int j = 0; j < n->adjacent.size(); j++)  
      if (n->adjacent[j]->manifold)
        good.push_back(n->adjacent[j]);

    if (good.size() < 5)
      continue;
      
    //test for minima
    bool isMinima = true, isMaxima = false;
    for (unsigned int j = 0; j < good.size(); j++)  {
      if (good[j]->laplacian > n->laplacian) {
        isMinima = false;     break;
      }      
    }

    //test for maxima
    if (!isMinima) 
    {
      isMaxima = true;
      for (unsigned int j = 0; j < good.size(); j++)  {
        if (good[j]->laplacian < n->laplacian) {
          isMaxima = false;     break;
        }
      }
    }

    if (isMaxima || isMinima)
    {
      PSKpt *k = RefinePeak(n);
      if (k) {        
        extrema.push_back(k);
      }
    }   
  }
  

#ifdef VERBOSE
  vcl_cout << " found " << extrema.size() << "\n";
#endif
  vcl_vector<PSKpt *>::size_type numkpts = extrema.size();
  double filterwidth = 3.0*scale;
  vcl_vector<PSKpt *> filtered;
  if (extrema.size() > 1)
    FilterPSKeypoints(extrema.begin(), extrema.end(), filterwidth, filtered);
#ifdef VERBOSE
  vcl_cout << "filtered " << numkpts - filtered.size() << "\n";
#endif
  vcl_cout << "Computed " << filtered.size() << " keypoints for scale " << scale << "\n";

  for (unsigned int i = 0; i < filtered.size(); i++)  {
    PSKpt *k = filtered[i];
    AssignOrientation(k);
    if (ComputeDescriptor(k))
      keypoints.push_back(k);
    else
      delete k;
  }
}

//*****************************************************************************

PSKpt *PSKGraph::RefinePeak(PSKNode *n)
{ 
  double height = fabs(n->laplacian);
  PSKpt *k = new PSKpt;
  k->n = n;
  k->scale = scale;
  k->location = n->point;
  k->strength = height;
  k->normal = n->normal;
  return k;
}

//*****************************************************************************

//Helper struct for the filter function
struct keypoint_at_index_is_greater {
  bool operator()( vcl_vector< PSKpt *>::const_iterator i, vcl_vector< PSKpt *>::const_iterator j )
  {  return (*i)->strength > (*j)->strength;  }
};

void FilterPSKeypoints(vcl_vector<PSKpt *>::const_iterator begin_itr, vcl_vector<PSKpt *>::const_iterator end_itr,
                     double min_distance_between_points, vcl_vector<PSKpt *> &keypoints3d)
{
  if (begin_itr == end_itr) 
    return;
 
  //  Create a vector of subscripts into the original vector of
  //  keypoints, sorted according to increasing strength of the
  //  associated keypoint.

  vcl_vector<vcl_vector<PSKpt *>::const_iterator> sorted_subscripts;
  for (vcl_vector<PSKpt *>::const_iterator itr = begin_itr; itr != end_itr; itr++) sorted_subscripts.push_back(itr);
  keypoint_at_index_is_greater keypoint_compare;
  vcl_stable_sort(sorted_subscripts.begin(), sorted_subscripts.end(), keypoint_compare);

  // find the minimum and maximum coordinates in the keypoints

  vnl_vector_fixed<double,3> min_point;
  min_point[0] = (*begin_itr)->location[0];
  min_point[1] = (*begin_itr)->location[1]; 
  min_point[2] = (*begin_itr)->location[2];
  vnl_vector_fixed<double,3> max_point = min_point;
  for (vcl_vector<PSKpt *>::const_iterator itr = begin_itr+1; itr != end_itr; itr++ )
  {
    vnl_double_3 const& loc = (*itr)->location;
    for (int k = 0; k < 3; k++ )
    {
      if (loc[k] < min_point[k] ) 
        min_point[k] = loc[k];
      else if ( loc[k] > max_point[k] )
        max_point[k] = loc[k];
    }
  }

  //  Assign the bin sizes to be isotropic.  The default size is 2
  //  times the min point distance.  The subsequent for loop ensures
  //  that a reasonable size is set when the min_distance is too small.

  double bsize = 0;
  const int max_bins_per_dim = 50;
  for ( unsigned int i = 0; i < 3; i++ )
  {
    double lower_bound_on_size = (max_point[i] - min_point[i]) / max_bins_per_dim;
    if ( bsize < lower_bound_on_size ) 
      bsize = lower_bound_on_size;
  }

  vnl_vector_fixed<double,3> bin_sizes( bsize );
  rsdl_bins<3, double, int > subscript_bins(min_point, max_point, bin_sizes );

  //  For each remaining point (in order of strength) only put it into
  //  the bins and the filtered result if there isn't already a point
  //  within the min distance of its physical location.
  unsigned int subscript = 0;
  while ( subscript < sorted_subscripts.size() )
  {
    PSKpt *keypoint = *sorted_subscripts[subscript];
    vnl_vector_fixed<double,3> loc;
    loc[0] = keypoint->location[0];  loc[1] = keypoint->location[1];  loc[2] = keypoint->location[2];

    vcl_vector<int> pts_in_rad;
    subscript_bins.points_within_radius(loc, min_distance_between_points, pts_in_rad);
    bool passed = true;
    for (unsigned int i = 0; i < pts_in_rad.size(); i++)
    {
      PSKpt *close = *sorted_subscripts[pts_in_rad[i]];
      passed = false;
      break;
    }

    if (passed)
    {
      subscript_bins.add_point(loc, subscript );
      keypoints3d.push_back( keypoint );
    }
    else
      delete keypoint;

    subscript++;
  }
}

//*****************************************************************************

void PSKGraph::AssignOrientation(PSKpt *kp)
{
  vcl_vector<PSKNode *> locs;
  GetPointsInRange(kp->location, 3.0*kp->scale, locs);

  vnl_matrix_fixed<double, 3, 3> covar(0.0);
  for (unsigned int i = 0; i < locs.size(); i++)
  {
    PSKNode *node = locs[i];
    vnl_double_3 vec = node->point - kp->location;
    vec = vec - dot_product(vec, kp->normal)*kp->normal;
    covar += outer_product(vec, vec);
  }

  // Get the eigenvector/eigenvalue pairs for the covariance matrix.
  // Random coordinate system does not seem to work
  vnl_symmetric_eigensystem<double> eig(covar);

  // Get a normal estimate.  Flip the sign to agree with view
  // direction.
  vnl_double_3 u = eig.get_eigenvector(2).normalize();
  vnl_double_3 v = eig.get_eigenvector(1).normalize();

  double gauss_exp_denom = 2.0*kp->scale*kp->scale;
  double bilateral_exp_denom = 2.0*bilateral*bilateral; 
  vnl_double_3 dir(0.0,0.0,0.0);
  const unsigned int bins = 36;
  vnl_vector<double> histo(bins, 0.0);
  for (unsigned int l = 0; l < locs.size(); l++)
  {
    PSKNode *n = locs[l];

    //compute gradient contribution and direction
    vnl_double_3 vec = n->point - kp->location;
    double ssd = (kp->location-n->point).squared_magnitude();
    double nd = 1.0 - dot_product(kp->normal, n->normal);
    double weight = exp(-ssd/gauss_exp_denom) * exp(-nd*nd/bilateral_exp_denom);
    vnl_double_2 grad2(dot_product(u, n->gradient), dot_product(v, n->gradient));
    double gmag = weight * /*n->gradient.two_norm();//*/grad2.two_norm();
    
    //compute angle of vector in the plane
    double angle = atan2(grad2[1], grad2[0]);
    if (angle < 0.0)
      angle += two_pi;

    //enter into histogram
    double bin = (angle/two_pi)*bins; 
    double fpart = bin - vcl_floor(bin);
    histo[(int)vcl_floor(bin)] += (1.0-fpart)*gmag;
    histo[((int)vcl_ceil(bin))%bins] += fpart*gmag;
  }

  //find histogram's max
  int maxindex = 0; double max = histo[0];
  for (unsigned int i = 1; i < bins; i++)
  {
    if (histo[i] > max) {
      max = histo[i];
      maxindex = i;
    }
  }

  //Parabola fit  
  int left = maxindex == 0 ? bins-1 : maxindex - 1;
  int right = (maxindex+1)%bins;

  double a = 0.5 * (histo[left] + histo[right] - 2.0*max);
  double b = -0.5 * (histo[left] - histo[right]);
  double interpbin = (-b / (2.0 * a)) + maxindex;
  double interpangle = (interpbin/bins)*two_pi;

  kp->direction = vcl_cos(interpangle)*u + vcl_sin(interpangle)*v;
  kp->direction.normalize();
}

//*****************************************************************************

bool PSKGraph::ComputeDescriptor(PSKpt *kp)
{
  const int n = 8;               //number of orientation histogram bins
  const double ndub = (double)n;
  const int m = 4;               //dimensions of spacial bins (ie m x m)

  double dh = 4.0 * kp->scale;   //bin width
  double radius = dh*m/2.0;                 //radius of descriptor

  double gaussian_exp_denom = 2.0*radius*radius;
  double bilateral_exp_denom = 2.0*bilateral*bilateral; 

  kp->feature_vector.fill(0.0);

  double bin_counts[16];
  memset(bin_counts, 0, 16*sizeof(double));

  vnl_double_3 yaxis = vnl_cross_3d(kp->normal, kp->direction);

  vcl_vector<PSKNode *> locs;
  this->GetPointsInRange(kp->location, radius * sqrt(2.0), locs);

  for (unsigned int i = 0; i < locs.size(); i++)
  {   
    PSKNode *node = locs[i];

    //Locate sample's projection onto keypoint plane
    vnl_double_3 vec = node->point - kp->location;

    if (dot_product(kp->normal, vec) > dh)
      continue;

    vnl_double_2 pt2(dot_product(kp->direction, vec), dot_product(yaxis, vec));

    //Project gradient into keypoint plane
    vnl_double_2 gproj2(dot_product(node->gradient, kp->direction), dot_product(node->gradient, yaxis));

    double grad_mag = gproj2.magnitude();   //projected magnitude ???
    double grad_ori = atan2(gproj2[1], gproj2[0]);

    if(grad_ori < 0) grad_ori += two_pi; 
    if(grad_ori >= two_pi) grad_ori -= two_pi; 

    // The index of the spatial bin in which the point falls.
    int indi = Real2Int(vcl_floor((pt2[0]/dh)+0.5));
    int indj = Real2Int(vcl_floor((pt2[1]/dh)+0.5));

    // The index of the orientation bin in which the point falls.
    double t = ndub*(grad_ori/two_pi);
    if (t >= ndub) t -= ndub;
    int oribin = Real2Int(t);
    double dt = t - vcl_floor(t);
    double dx = (pt2[0]/dh)-((double)indi-0.5);
    double dy = (pt2[1]/dh)-((double)indj-0.5);

    // Get the gradient magnitude and gaussian weight.
    double ndp = dot_product(kp->normal, node->normal);
    double nd = 1.0 - ndp;
    double wgauss = (double)(exp(-vec.squared_magnitude()/gaussian_exp_denom)*exp(-nd*nd/bilateral_exp_denom));

    // Contribute weight to each histogram center around this bin in
    // both spatial position and orientation.
    for(int idx = 0; idx < 8; ++idx)
    {
      // Weight with partial volume interpolation.
      double wx = (idx & 1)? dx : (1.0-dx);
      double wy = (idx >> 1 & 1)? dy : (1.0-dy);
      double wt = (idx >> 2 & 1)? dt : (1.0-dt);
      double w = wgauss*grad_mag*wx*wy*wt;
      assert(w >= 0);

      // Contribute only to bins in range.
      int ii = indi+1 + (idx & 1);
      int ij = indj+1 + (idx>>1 & 1);
      int ik = (oribin + (idx>>2 & 1))%n;
      if (ii >= 0 && ii <= 3 && ij >= 0 && ij <= 3)     
      {
        kp->feature_vector(n*4*ij + n*ii + ik) += w;    
        bin_counts[4*ij + ii]+=wx*wy*wt;    
      }
    } 
  }

  // Normalize the descriptor vector.
  double two_norm = kp->feature_vector.two_norm();
  assert(two_norm >= 0);
  if(two_norm > 0)
    kp->feature_vector /= two_norm;

  // Threshold components that are too dominant and normalize again.
  for(unsigned int i = 0; i < kp->feature_vector.size(); ++i)
  {
    if(kp->feature_vector(i) > DESCRIPTOR_COMPONENT_MAXIMUM)
      kp->feature_vector(i) = DESCRIPTOR_COMPONENT_MAXIMUM;
  }
  two_norm = kp->feature_vector.two_norm();
  if(two_norm > 0)  
    kp->feature_vector /= two_norm;

  // Convert the descriptor to a 9-bit fixed-point representation.
  bool nan = false;
  for(unsigned int i=0; i < kp->feature_vector.size(); ++i) {
    kp->feature_vector(i) = vcl_floor(kp->feature_vector(i)*512+0.5);
    if (kp->feature_vector(i) != kp->feature_vector(i))
      nan = true;
  } 

  return !locs.empty() && !nan;
}

//*****************************************************************************

void PSKGraph::ComputeLaplacianBounds(double &lower, double &upper)
{
  vcl_vector<double> laplacians;
  for (unsigned int i = 0; i < nodes.size(); i++)
    if (nodes[i]->manifold)
      laplacians.push_back(nodes[i]->laplacian);

  double min = laplacians[0];
  double max = min;
  for (unsigned int i = 1; i < laplacians.size(); i++)
  {
    if (laplacians[i] < min)
      min = laplacians[i];
    if (laplacians[i] > max)
      max = laplacians[i];
  }

  // Build a histogram of the responses.
  int nbins = 1000;
  vnl_vector<int> histogram(nbins, 0);
  double z = double(nbins)/(max-min);
  for (unsigned int i = 1; i < laplacians.size(); i++)
  {
    double v = laplacians[i];
    int bin = int((v-min)*z);
    if(bin > nbins-1) { bin = nbins-1; }
    ++histogram(bin);
  }

  // Convert to a cumulative histogram.  Also compute the thresholds
  // for the bottom and top fraction ("percentile") of the data.
  int pixels = (int)laplacians.size();
  double total = 0.0f;
  lower = max;
  upper = min;
  for (int i = 0; i < nbins; i++)
  {
    if(lower == max && total/pixels >= PEAKHEIGHTFRACTION)    
      lower = i/z + min;      
    if(upper == min && total/pixels >= (1.0-PEAKHEIGHTFRACTION))      
      upper = i/z + min;      
    total += histogram(i);
    histogram(i) = total;
  }
}

//*****************************************************************************

void PSKGraph::ApplyRigidXForm(const SimpleRigid &sr)
{
  for (unsigned int i = 0; i < nodes.size(); i++)
  {
    nodes[i]->normal = sr.map_direction(nodes[i]->normal);
    nodes[i]->point = sr.map_point(nodes[i]->point);
  }
}

//*****************************************************************************

void PSKGraph::WriteVTP(const char *filename) const
{
  vtkSmartPointer<vtkPoints> pts = vtkSmartPointer<vtkPoints>::New(); 
  vtkSmartPointer<vtkFloatArray> norms = vtkSmartPointer<vtkFloatArray>::New();
  vtkSmartPointer<vtkCellArray> tris = vtkSmartPointer<vtkCellArray>::New();
  vtkSmartPointer<vtkFloatArray> gradmag = vtkSmartPointer<vtkFloatArray>::New();
  vtkSmartPointer<vtkFloatArray> laplacian = vtkSmartPointer<vtkFloatArray>::New();
  vtkSmartPointer<vtkFloatArray> colors = vtkSmartPointer<vtkFloatArray>::New();
  vtkSmartPointer<vtkIntArray> manifold = vtkSmartPointer<vtkIntArray>::New();
  manifold->SetName("manifold");
  gradmag->SetName("gradmag");
  laplacian->SetName("laplacian");
  colors->SetName("color");
  norms->SetNumberOfComponents(3);

  for (unsigned int i = 0; i < nodes.size(); i++)  
  {
    PSKNode *n = nodes[i];
    pts->InsertNextPoint(n->point.data_block());  
    if (n->manifold) gradmag->InsertNextValue(n->gradient.magnitude());
    else gradmag->InsertNextValue(0.0);
    laplacian->InsertNextValue(n->laplacian);
    colors->InsertNextValue(n->intensity);
    manifold->InsertNextValue(n->manifold);
    norms->InsertNextTuple3(n->normal[0], n->normal[1], n->normal[2]);
  }

  for (unsigned int i = 0; i < triangles.size(); i++)
  {
    vtkIdType triangle[3];
    triangle[0] = triangles[i].a->index;
    triangle[1] = triangles[i].b->index;
    triangle[2] = triangles[i].c->index;
    tris->InsertNextCell(3, triangle);
  }  

  vtkSmartPointer<vtkPolyData> polydata = vtkSmartPointer<vtkPolyData>::New();
  polydata->SetPoints(pts);
  polydata->SetPolys(tris);
  polydata->GetPointData()->SetNormals(norms);
  polydata->GetPointData()->AddArray(gradmag);
  polydata->GetPointData()->AddArray(laplacian);
  polydata->GetPointData()->AddArray(colors);
  polydata->GetPointData()->AddArray(manifold);

  vtkSmartPointer<vtkXMLPolyDataWriter> writer = vtkSmartPointer<vtkXMLPolyDataWriter>::New();
  writer->SetInput(polydata);
  writer->SetFileName(filename);
  writer->Write(); 
}

//*****************************************************************************

void PSKGraph::SavePSKGraph() const
{
  unsigned int size = (unsigned int)nodes.size();
  fwrite(&size, sizeof(unsigned int), 1, graph_file);
  for (unsigned int i = 0; i < size; i++)
  {
    fwrite(nodes[i]->point.data_block(), 3*sizeof(double), 1, graph_file);
    fwrite(nodes[i]->normal.data_block(), 3*sizeof(double), 1, graph_file);
    fwrite(&nodes[i]->intensity, sizeof(double), 1, graph_file);
  }   

  unsigned int numtriangles = (unsigned int)triangles.size();
  fwrite(&numtriangles, sizeof(unsigned int), 1, graph_file);
  for (unsigned int i = 0; i < numtriangles; i++)
  {
    fwrite(&triangles[i].a->index, sizeof(int), 1, graph_file);
    fwrite(&triangles[i].b->index, sizeof(int), 1, graph_file);
    fwrite(&triangles[i].c->index, sizeof(int), 1, graph_file);
  }

  fwrite(&scale, sizeof(double), 1, graph_file);
}

//*****************************************************************************

void PSKGraph::SaveScale() const
{
  fwrite(&scale, sizeof(double), 1, graph_file);
 
  for (unsigned int i = 0; i < nodes.size(); i++)
  {
    fwrite(&nodes[i]->intensity, sizeof(double), 1, graph_file);
    
    if (xform)  {
      vnl_double_3 normal = xform->inverse().map_direction(nodes[i]->normal);
      fwrite(normal.data_block(), 3*sizeof(double), 1, graph_file);    
    }
    else 
      fwrite(nodes[i]->normal.data_block(), 3*sizeof(double), 1, graph_file);  
  }   
}

//*****************************************************************************

void PSKGraph::LoadPSKGraph()
{
  unsigned int size;
  fread(&size, sizeof(unsigned int), 1, graph_file);
  for (unsigned int i = 0; i < size; i++)
  {
    PSKNode *n = new PSKNode;
    fread(n->point.data_block(), 3*sizeof(double), 1, graph_file);
    fread(n->normal.data_block(), 3*sizeof(double), 1, graph_file);  
    fread(&n->intensity, sizeof(double), 1, graph_file);
    n->index = i;
    nodes.push_back(n);
  }

  unsigned int numtriangles;
  fread(&numtriangles, sizeof(unsigned int), 1, graph_file);
  triangles.resize(numtriangles);
  for (unsigned int i = 0; i < numtriangles; i++)
  {
    int a, b, c;
    fread(&a, sizeof(int), 1, graph_file);
    fread(&b, sizeof(int), 1, graph_file);
    fread(&c, sizeof(int), 1, graph_file);
    triangles[i].a = nodes[a];
    triangles[i].b = nodes[b];
    triangles[i].c = nodes[c];
    triangles[i].gradient = vnl_double_3(0.0,0.0,0.0);
    ConnectTriangle(triangles[i], i);
  }

  fread(&scale, sizeof(double), 1, graph_file);
}

//*****************************************************************************

void PSKGraph::LoadNextScale()
{
  fread(&scale, sizeof(double), 1, graph_file);
  
  for (unsigned int i = 0; i < nodes.size(); i++)
  {
    fread(&nodes[i]->intensity, sizeof(double), 1, graph_file);
    
    if (xform)  {
      vnl_double_3 normal;
      fread(normal.data_block(), 3*sizeof(double), 1, graph_file);    
      nodes[i]->normal = xform->map_direction(normal);
    }
    else 
      fread(nodes[i]->normal.data_block(), 3*sizeof(double), 1, graph_file);  
  }   
}

//*****************************************************************************

void PSKGraph::GetPointsInRange(const vnl_double_3 &point, double dist, vcl_vector<PSKNode *> &locs) const
{
  vcl_vector<rsdl_point> pts;
  vcl_vector<int> inds;
  rsdl_point pt(point);
  kdtree->points_in_radius(pt, dist, pts, inds);
  for (unsigned int i = 0; i < inds.size(); i++)
    locs.push_back(nodes[inds[i]]);
}

//*****************************************************************************

void PSKGraph::GetControlPointsInRange(const vnl_double_3 &point, double dist, vcl_vector<PSKNode *> &locs) const
{
  vcl_vector<rsdl_point> pts;
  vcl_vector<int> inds;
  rsdl_point pt(point);
  controlkd->points_in_radius(pt, dist, pts, inds);
  for (unsigned int i = 0; i < inds.size(); i++)
    locs.push_back(nodes[control[inds[i]]]);
}

//*****************************************************************************
