
#ifndef PointRecord_h_
#define PointRecord_h_

// Input assumptions:
//   All grid points have an associated direction.  This is a unit
//   vector in the scanner's coordinate system.   If there is a
//   measured depth then "is_valid_return" will be marked true and there
//   will be a postive value to "measured_distance".  If there is no
//   return along a particular line of sight, then "valid_return"
//   will be marked false and "measured_distance" is given a value of
//   0. 
//
// Output conditions:
//   Grid points marked with no valid return will be unchanged.  All
//   other grid points will be marked with a surface_type.
//
//   1. PRUndersampled is reserved for regions where the surfaces are
//   sufficiently far from the scanner that the distance between
//   points on the surface, even normal to the line of sight, is too
//   great.  These points have normals that point back to the
//   scanner, and their estimated distances are the same as their measured
//   distances.  No other values within the PointRecord class should
//   be trusted.
//
//   2. PROutlier is a point that the algorithm considers does not
//   belong to any surface.  It should be ignored in subsequent
//   processing.  Its normal and estimated distance are assigned in
//   the same manner as PRUndersampled points.
//  
//   All other point types PRSurface, PRCurve and PRCloud can be
//   treated as valid points with estimated scales, distances and
//   surface normals.  PRSurface points have estimated angle scales.
//   In new developments the points are really now to be treated as a
//   rectangular or elliptical (planar) surface patch.  These patches,
//   under either interpretation, should cover the surface.  Information
//   about the patch is recorded in the two additional vectors and two
//   additional scalars in the PointRecord:
//
//     . surface_x_dir is the direction perpendicular to the surface
//       normal along which the planar patch is most elongated.  If
//       the surface is viewed as an ellipse, surface_x_dir gives the
//       direction of the major axis.
//
//     . surface_x_scale is the half-length of the patch along this
//       direction.  The resulting rectangle length would therefore
//       be treated as twice this value.  
//
//     . surface_y_dir is the drection perpendicular to the surface
//       normal (and surface_x_dir) along which the surface patch is
//       least  elongated.  When viewing the patch as an ellipse,
//       this is the minor axis.
//  
//     . surface_y_scale is the half-width of the patch along the
//       surface_y_dir.   The resulting rectangle width would
//       therefore be twice the value.
//
//   There is really not much functional difference between a PRCloud
//   and a PRSurface labeling, although the PRCloud points have an
//   estimated distance that is the same and the measured distance.
//   Now a PRCurve is also no different:  the surface_x_dir is the
//   direction along the curve tangent, and the surface_y_dir is normal
//   to this.  The surface_normal is the direction in tangent plane
//   that is closest to the negative line of sight direction.
//

#include <iostream>
#include <fstream>
// #include <vnl/vnl_vector.h>
#include <vnl/vnl_double_3.h>

class PointRecord {
 public:
  PointRecord() {}

  PointRecord( bool valid,
               vnl_double_3 const& dir,
               double distance )
    : is_valid_return(valid), direction(dir), measured_distance(distance)
  {}

  //  Input stuff
  bool is_valid_return;
  vnl_double_3 direction;   // unit vector
  double measured_distance;

  //  Intermediate information and status stuff.  
  bool is_initialized;
  bool is_done;
  double point_spacing_normal_to_los;
  double priority;
  int pq_index;
  int done_index;

  //  Results
  enum SurfaceType { PRSurface, PRCurve, PRCloud, PRUndersampled, PROutlier };
  SurfaceType surface_type;
  double estimated_scale;
  double estimated_angle_scale;
  double estimated_distance;
  double coherence;
  double patch_sigma[2];
  int patch_inliers;
  double patch_wgted_inliers;
  vnl_double_3 surface_normal;

  //  The corner directions start at the upper left with respect to the
  //  grid and proceed clockwise.
  vnl_double_3 corner_directions[4];
  double corner_distances[4];
};

void print( std::ostream & str, PointRecord const& p );

#endif
