#ifndef SIMPLERIGID_H_
#define SIMPLERIGID_H_

#include <vcl_fstream.h>
#include <vcl_string.h>
#include <vcl_sstream.h>
#include <vnl/vnl_double_3.h>
#include <vnl/vnl_matrix_fixed.h>
#include <vnl/algo/vnl_svd.h>

class SimpleRigid
{
public:

  SimpleRigid() : m_t(vnl_double_3(0.0,0.0,0.0)) { m_R.set_identity(); }
  SimpleRigid(const vnl_matrix_fixed<double, 3, 3> &R, const vnl_double_3 &t) : m_R(R), m_t(t) {}
  SimpleRigid(FILE *file)
  {
    fread(m_R.data_block(), sizeof(double), 9, file);
    fread(m_t.data_block(), sizeof(double), 3, file);
  }  
  SimpleRigid(const char *xformfile) { read_xform_file(xformfile); }

  void OrthogonalizeR()
  {
    vnl_svd<double> svd(m_R);    
    m_R = svd.U() * svd.V().transpose();
  }

  SimpleRigid inverse() const
  {
    SimpleRigid sr;
    sr.m_t = - m_R.transpose() * m_t;
    sr.m_R = m_R.transpose();
    return sr;
  }

  SimpleRigid operator*(const SimpleRigid &right) const
  {
    SimpleRigid sr;
    sr.m_t = m_R * right.m_t + m_t;
    sr.m_R = m_R * right.m_R;
    return sr;
  }

  vnl_double_3 map_point(const vnl_double_3 &pt) const
  {
    return m_R * pt + m_t;
  }

  vnl_double_3 map_direction(const vnl_double_3 &dir) const
  {
    return m_R * dir;
  }

  void write_to_file(FILE *file) const
  {
    fwrite(m_R.data_block(), sizeof(double), 9, file);
    fwrite(m_t.data_block(), sizeof(double), 3, file);
  }

  void write_xform_file(const char *filename, const vcl_string &fixed, const vcl_string &moving) const
  {
    vcl_ofstream outfile(filename);
    outfile << fixed << "\n" << moving << "\n1\n";
    outfile << m_t << " " << m_R.get_row(0) << " " << m_R.get_row(1)
            << " " << m_R.get_row(2) << "\n";
    outfile.close();
  }

  void write_to_xform_file(vcl_ofstream &xformfile)
  {
    xformfile << m_t << " " << m_R.get_row(0) << " " << m_R.get_row(1)
              << " " << m_R.get_row(2) << "\n";
  }

  //reads in only first xform in the file
  void read_xform_file(const char *filename)
  {
    vcl_ifstream infile(filename);
    vcl_string x; vnl_double_3 r1, r2, r3;
    infile >> x >> x >> x >> m_t >> r1 >> r2 >> r3;
    m_R.set_row(0, r1);    m_R.set_row(1, r2);    m_R.set_row(2, r3);  
    infile.close();
  }

  void read_from_xform_file(vcl_ifstream &xformfile)
  {
    vnl_double_3 r1, r2, r3;
    vcl_string line;
    vcl_getline(xformfile, line);
    vcl_stringstream linestream(line);
    linestream >> m_t >> r1 >> r2 >> r3;
    m_R.set_row(0, r1);    m_R.set_row(1, r2);    m_R.set_row(2, r3);  
  }

  void print() const
  {
    vcl_cout << "Translation:\t" << m_t << "\n" << "Rotation:\t" << m_R.get_row(0) << "\n\t\t" << m_R.get_row(1)
             << "\n\t\t" << m_R.get_row(2) << "\n\n";
  }

  vnl_matrix_fixed<double, 3, 3> m_R;
  vnl_double_3 m_t;
};

#endif