#include "SphericalBins.h"
#include <vcl_cmath.h>
#include <vnl/vnl_math.h>
#include <vcl_cassert.h>
#include <float.h>

const double twoPi = 2*vnl_math::pi;

//*****************************************************************************

const vcl_vector<unsigned int>& SphericalBins::getBinAtIndex(unsigned int index) const
{
  return m_bins[index];
}

//*****************************************************************************

int SphericalBins::getBinAtThetaPhi(double theta, double phi, vcl_vector<unsigned int>& bin)
{
  while (theta < m_theta_origin) theta += twoPi;
  int u = static_cast<int>(vcl_floor((theta - m_theta_origin)/m_theta_spacing));
  int v = static_cast<int>(vcl_floor((phi - m_phi_origin)/m_phi_spacing));

  int index = m_bins_width*v + u;

  if (u < 0 || v < 0 || u >= m_bins_width || v >= m_bins_height)
    return -1;

  bin = m_bins[index];
  return index;
}

//*****************************************************************************

int SphericalBins::getThetaIndex(double theta) const
{
  while (theta < m_theta_origin) theta += twoPi;
  return static_cast<int>(vcl_floor((theta - m_theta_origin)/m_theta_spacing));
}

//*****************************************************************************

int SphericalBins::getPhiIndex(double phi) const
{
  return static_cast<int>(vcl_floor((phi - m_phi_origin)/m_phi_spacing));
}

//*****************************************************************************

bool SphericalBins::insertIndex(double theta, double phi, unsigned int ind)
{
  while (theta < m_theta_origin) theta += twoPi;
  int u = static_cast<int>(vcl_floor((theta - m_theta_origin)/m_theta_spacing));
  int v = static_cast<int>(vcl_floor((phi - m_phi_origin)/m_phi_spacing));

  int index = m_bins_width*v + u;
  if (u >= 0 && v >= 0 && u < m_bins_width && v < m_bins_height)
  {
    for (unsigned int i = 0; i < m_bins[index].size(); i++)
      if (m_bins[index][i] == ind)
        return false;

    m_bins[index].push_back(ind);
    return true;
  }

  return false;
}

//*****************************************************************************

bool SphericalBins::insertIndex(unsigned int index, unsigned int ind)
{
  if (index < m_bins.size())
  {
    for (unsigned int i = 0; i < m_bins[index].size(); i++)
      if (m_bins[index][i] == ind)
        return false;

    m_bins[index].push_back(ind);
    return true;
  }

  return false;
}

//*****************************************************************************

//Finds closest point within a theta/phi range (approximate closest)
unsigned int SphericalBins::findClosestTo(const vnl_double_3 &pt, const vcl_vector<vnl_double_3> &points, double thresh) const
{
  unsigned int index = (unsigned int)points.size();
  double distsq = DBL_MAX;

  double x = pt[0], y = pt[1], z = pt[2];
  double r = vcl_sqrt(x*x+y*y+z*z);
  double theta = 0.0;
  double phi = 0.0;

  if(r > 0)  
  phi = vcl_acos(z/r);

  if(x*x+y*y > 0)  
   theta = vcl_atan2(y, x);

  while (theta < m_theta_origin) theta += twoPi;

  vcl_vector<unsigned int> bin;
  int u = static_cast<int>(vcl_floor((theta - m_theta_origin)/m_theta_spacing));
  int v = static_cast<int>(vcl_floor((phi - m_phi_origin)/m_phi_spacing));

  if (u < 0 || v < 0 || u >= (int)m_bins_width || v >= (int)m_bins_height)
    return index;

  int bin_index;
  if (thresh > 0.0) 
  {
    bin_index = m_bins_width*v + u;
    if (!m_bins[bin_index].empty())
    {
      index = m_bins[bin_index][0];
      distsq = vnl_vector_ssd(pt, points[index]);

      if (distsq < thresh)
        return index;
    }
  }
  
  for (int j = v - 2; j <= v + 2; j++)
  {
    if (j < 0 || j >= m_bins_height) continue;
    for (int i = u - 2; i <= u + 2; i++)
    {
      if (i < 0 || i >= m_bins_width) continue;
      bin_index = m_bins_width*j + i;
      for (unsigned int k = 0; k < m_bins[bin_index].size(); k++)
      {
        double d = vnl_vector_ssd(pt, points[m_bins[bin_index][k]]);
        if (d < distsq)
        {
          index = m_bins[bin_index][k];
          distsq = d; 
        }
      }
    }
  }

  return index;
}

//*****************************************************************************

