#include "pskmultigrid.h"
#include "scan.h"
#include "parameters.h"
#include <vcl_algorithm.h>

#include <rsdl/rsdl_point.h>
#include <rsdl/rsdl_kd_tree.h>

#define MINSCALE .03
#define RADIANS_TO_DEGREES(x) (x)*57.29577951
#define DESCRIPTOR_COMPONENT_MAXIMUM 0.2

const double two_pi = 2.0*vnl_math::pi;  
vcl_vector<double> scales;

struct match
{
  PSKpt *f, *m, *b;
  double ratio;
  bool correct;
};

typedef vcl_vector<vcl_vector<PSKpt *> > KeyVec;

void ComputeKeypoints(PSKMultiGrid *g, KeyVec &keypoints, const vcl_string &dir);
void GetPossibleMatches(const KeyVec &fixed, const KeyVec &moving, vcl_vector<match> &pmatches);
void MatchKeypoints(const KeyVec &fixed, const KeyVec &moving, const char *filename);
void GetBestMatch(const KeyVec &fixed, vcl_vector<match> &pmatches);
void WriteMatches(const vcl_vector<match> &matches, const char *filename);
void WriteKeypoints(const KeyVec &kpts, const char *filename, SimpleRigid *xform);



int main(int argc, char *argv[])
{
  bool load = false;
  //c:/Data/vcc_south/A vcc_south1.ptx c:/Data/vcc_south/B vcc_south2.ptx to_vcc_south1.xform 10 3
  //c:/Data/vcc_south/D vcc_south4.ptx c:/Data/vcc_south/F vcc_south6.ptx to_vcc_south4.xform 3 2
  //c:/Data/2006-08-09/A dccwest1.ptx c:/Data/2006-08-09/B dccwest2.ptx truth.xform 9 16
  //c:/Data/2004-10-27/A bio_side.ptx c:/Data/2004-10-27/B bio_side2.ptx truth.xform 9 14
  //c:/Data/2000-00-00/A vcc_fake1.ptx c:/Data/2000-00-00/B vcc_fake2.ptx truth.xform 6 5
  //c:/Data/2006-07-31/A jecwest1.ptx c:/Data/2006-07-31/B jecwest2.ptx truth.xform 5 3
  //c:/Data/sci_center/A science_center_sw1.ptx c:/Data/sci_center/B jrowl2.ptx truth.xform 8 2
  //c:/Data/parking_lot/A MRCParkinglot1.ptx c:/Data/parking_lot/B MRCParkinglot2.ptx truth.xform 3 6

  vcl_string dir1 = vcl_string(argv[1]);
  vcl_string name1 = vcl_string(argv[2]);
  vcl_string dir2 = vcl_string(argv[3]);
  vcl_string name2 = vcl_string(argv[4]);

  vcl_string xformfile = dir2 + "/" + vcl_string(argv[5]);

  int img1 = atoi(argv[6]);
  int img2 = atoi(argv[7]);

  Parameters::Instance()->SetVariable(Parameters::LOADRANGEDATA, true);
  SetPriorityClass(GetCurrentProcess(), BELOW_NORMAL_PRIORITY_CLASS);
  SimpleRigid xform(xformfile.c_str());

  KeyVec fixedkpts, movingkpts;

  scales.push_back(MINSCALE);
  double k = pow(2.0, 1.0/2.0);
  while (scales.back() < .15) 
    scales.push_back(k * scales.back());
  scales.push_back(0.05);
  scales.push_back(0.07);
  scales.push_back(0.09);
  scales.push_back(0.12);
  scales.push_back(0.17);

  fixedkpts.resize(scales.size());
  movingkpts.resize(scales.size());

  Scan *scan1 = new Scan(true);
  scan1->ReadInScanInfo(dir1.c_str(), name1.c_str()); 
  //PSKMultiGrid *mg1 = new PSKMultiGrid(*scan1, img1, scales[0], 213, 510, 554, 892, NULL, load);
  PSKMultiGrid *mg1 = new PSKMultiGrid(*scan1, img1, scales, -1, -1, -1, -1, NULL, load);
  delete scan1;
  ComputeKeypoints(mg1, fixedkpts, dir1);
  delete mg1;

  Scan *scan2 = new Scan(true);
  scan2->ReadInScanInfo(dir2.c_str(), name2.c_str()); 
  //PSKMultiGrid *mg2 = new PSKMultiGrid(*scan2, img2, scales[0], 332, 776, 261, 800, &xform, load);
  PSKMultiGrid *mg2 = new PSKMultiGrid(*scan2, img2, scales, -1, -1, -1, -1, &xform, load);
  delete scan2;
  ComputeKeypoints(mg2, movingkpts, dir2);
  delete mg2;
  

  //vcl_vector<match> pmatches;
  //GetPossibleMatches(fixedkpts, movingkpts, pmatches);
  //GetBestMatch(fixedkpts, pmatches);
  //vcl_string matchname = dir2 + "/matches.dat"; 
  //WriteMatches(pmatches, matchname.c_str());
  WriteKeypoints(fixedkpts, (dir1 + "/pskpts.dat").c_str(), NULL);
  WriteKeypoints(movingkpts, (dir2 + "/pskpts.dat").c_str(), &xform);

  MatchKeypoints(fixedkpts, movingkpts, dir2.c_str());
  return 0;
}

//*****************************************************************************

void ComputeKeypoints(PSKMultiGrid *mg, KeyVec &keypoints, const vcl_string &dir)
{
  char buf[32];

  int count = 0;
  unsigned int i = mg->minscaleindex;
  mg->DetectExtrema(keypoints[i], "blah.vtp"); // detect for minscale
  count += keypoints[i].size();
  sprintf(buf, "%.3fm-mesh.vtp", scales[i]);
  vcl_string filename = dir + "/" + vcl_string(buf);
  mg->WriteVTP(filename.c_str());
  i++;
  for ( ; i < scales.size(); i++)
  {
    mg->DecimateToScale(scales[i]);
    mg->DetectExtrema(keypoints[i], "blah.vtp");
    count += keypoints[i].size();
    sprintf(buf, "%.3fm-mesh.vtp", scales[i]);
    vcl_string filename = dir + "/" + vcl_string(buf);
    mg->WriteVTP(filename.c_str());
  }

  vcl_cout << "Detected " << count << " keypoints.\n";
}

//*****************************************************************************

void GetPossibleMatches(const KeyVec &fixed, const KeyVec &moving, vcl_vector<match> &pmatches)
{
  int count = 0;
  for (unsigned int s = 0; s < fixed.size(); s++)
  {
    vcl_vector<rsdl_point> kdpts;

    for(unsigned int i = 0; i < fixed[s].size(); i++)
    {
      rsdl_point point(fixed[s][i]->location);
      kdpts.push_back(point);
    }

    if (kdpts.empty())
      continue;

    rsdl_kd_tree kdtree(kdpts);

    for (unsigned int i = 0; i < moving[s].size(); i++)
    {
      vcl_vector<rsdl_point> pts;
      vcl_vector<int> inds;
      rsdl_point pt(moving[s][i]->location);
      kdtree.points_in_radius(pt, 2.0*moving[s][i]->scale, pts, inds);
	    for (unsigned int j = 0; j < pts.size(); j++)
	    {		    
        match m;
		    m.m = moving[s][i];
		    m.f = fixed[s][inds[j]];
        count++;
	      if (RADIANS_TO_DEGREES(acos(dot_product(m.m->direction, m.f->direction))) > 15.0) continue;

		    pmatches.push_back(m);
	    }
    }
  }

  vcl_vector<int> scalecount(scales.size(), 0);
  vcl_map<double, int> scalemap;
  for (unsigned int i = 0; i < scales.size(); i++)
    scalemap[scales[i]] = i;

  for (unsigned int i = 0; i < pmatches.size(); i++)  
    scalecount[scalemap[pmatches[i].f->scale]]++;  

  for (unsigned int i = 0; i < scalecount.size(); i++)
    vcl_cout << scalecount[i] << vcl_endl;  
  vcl_cout << pmatches.size() << " possible matches.\n";
  vcl_cout << count << " location matches.\n";
}

//*****************************************************************************

void MatchKeypoints(const KeyVec &fixed, const KeyVec &moving, const char *filename)
{
  vcl_cout << "Matching keypoints...\n";
  vcl_vector<match> matches;
  int totalcorrect = 0;
  int falsepositives = 0;
  for (unsigned int s = 0; s < fixed.size(); s++)
  {
    if (fixed[s].empty() || moving[s].empty())
      continue;

    vcl_cout << "matching at scale: " << scales[s] << " ";

    vcl_vector<rsdl_point> kdpts;

    for(unsigned int i = 0; i < fixed[s].size(); i++)
    {
      rsdl_point point(fixed[s][i]->feature_vector);
      kdpts.push_back(point);
    }

    rsdl_kd_tree kdtree(kdpts);

    int correct = 0;
    for (unsigned int i = 0; i < moving[s].size(); i++)
    {
      vcl_vector<rsdl_point> pts;
      vcl_vector<int> inds;
      rsdl_point pt(moving[s][i]->feature_vector);
      kdtree.n_nearest(pt, 2, pts, inds, true, 200);
      if (inds.size() == 2)
      {
        // Compute the ratio of distance to closest and second closest matches.
        double ratio = 1.0;

        PSKpt *first = fixed[s][inds[0]];
        PSKpt *second = fixed[s][inds[1]];
        double dist_to_first = (first->feature_vector - moving[s][i]->feature_vector).two_norm();
        double dist_to_second = (second->feature_vector - moving[s][i]->feature_vector).two_norm();

        ratio = dist_to_first/dist_to_second;
        if (ratio > 0.8)
          continue;
        
        match m;
        m.m = moving[s][i];
        m.f = first;
        m.b = second;
        m.ratio = ratio;

        if ((moving[s][i]->location - first->location).two_norm() <= 4.0*scales[s] && 
          RADIANS_TO_DEGREES(acos(dot_product(moving[s][i]->direction, first->direction))) <= 15.0) {
            correct++;
            m.correct = true;
            
        }
        else  {
          falsepositives++;
          m.correct = false;
        }
        matches.push_back(m);
      }
    }
    vcl_cout << correct << "\n";
    totalcorrect += correct;
  }

  vcl_vector<match> correct;
  for (unsigned int i = 0; i < matches.size(); i++)
    if (matches[i].correct)
      correct.push_back(matches[i]);

  WriteMatches(correct, (vcl_string(filename)+"/matches.dat").c_str());
  WriteMatches(matches, (vcl_string(filename)+"/matchesall.dat").c_str());


  vcl_sort(matches.begin(), matches.end());
  int firstcorrect = -1;
  for (unsigned int i = 0; i < matches.size(); i++) {
    vcl_cout << matches[i].correct << " ";
    if (firstcorrect == -1 && matches[i].correct)
      firstcorrect = i + 1;
  }
  vcl_cout << vcl_endl;

  unsigned int to = std::min<unsigned int>(50, matches.size());
  int intop50 = 0;
  for (unsigned int i = 0; i < to; i++)
  {
    if (matches[i].correct)
      intop50++;
  }

  vcl_cout << totalcorrect << " correct matches.\n";
  vcl_cout << falsepositives << " false positives.\n";
  vcl_cout << intop50 << " correct in top 50.\n";
  vcl_cout << firstcorrect << " first correct match.\n";


}

//*****************************************************************************

void GetBestMatch(const KeyVec &fixed, vcl_vector<match> &pmatches)
{
  int correct = 0;
  vcl_cout << "Getting Best Matches...\n";
  for (unsigned int s = 0; s < fixed.size(); s++)
  {
    vcl_vector<rsdl_point> kdpts;

    for(unsigned int i = 0; i < fixed[s].size(); i++)
    {
      rsdl_point point(fixed[s][i]->feature_vector);
      kdpts.push_back(point);
    }

    if (kdpts.empty())
      continue;

    rsdl_kd_tree kdtree(kdpts);

    for (unsigned int i = 0; i < pmatches.size(); i++)
    {
      PSKpt *m = pmatches[i].m;
      if (m->scale != scales[s]) continue;
      vcl_vector<rsdl_point> pts;
      vcl_vector<int> inds;
      rsdl_point pt(m->feature_vector);
      kdtree.n_nearest(pt, 1, pts, inds, true, 200);
      if (inds.size() == 1)
      {
        PSKpt *first = fixed[s][inds[0]];
        pmatches[i].b = first;
        if ((m->location - first->location).two_norm() <= 2.0*scales[s] && 
            RADIANS_TO_DEGREES(acos(dot_product(m->direction, first->direction))) <= 15.0)
            correct++;
      }
    }
  }

   // vcl_cout << correct << " correct matches.\n";
}

//*****************************************************************************

void WriteMatches(const vcl_vector<match> &matches, const char *filename)
{
  FILE *file = fopen(filename, "wb");
  unsigned int size = (unsigned int)matches.size();
 
  fwrite(&size, sizeof(unsigned int), 1, file);
  for (unsigned int i = 0; i < size; i++)
  {
    WriteKeypoint(file, matches[i].f);
    WriteKeypoint(file, matches[i].m);
    WriteKeypoint(file, matches[i].b);
    double ratio = 1.0;
    double dist_to_best = (matches[i].b->feature_vector - matches[i].m->feature_vector).two_norm();
    double dist_to_correct = (matches[i].f->feature_vector - matches[i].m->feature_vector).two_norm();
    if (dist_to_correct > 0)      
      ratio = dist_to_best/dist_to_correct;
    double normal_dist_to_correct = (matches[i].f->normal_offsets - matches[i].m->normal_offsets).two_norm();
    double normal_dist_to_best = (matches[i].b->normal_offsets - matches[i].m->normal_offsets).two_norm();

    fwrite(&dist_to_best, sizeof(double), 1, file);
    fwrite(&matches[i].ratio, sizeof(double), 1, file);
    fwrite(&normal_dist_to_correct, sizeof(double), 1, file);
    fwrite(&normal_dist_to_best, sizeof(double), 1, file);
  }

  fclose(file);
  vcl_cout << "Wrote " << size << " matches to " << filename << vcl_endl;
}

//*****************************************************************************

void WriteKeypoints(const KeyVec &kpts, const char *filename, SimpleRigid *xform)
{
  FILE *file = fopen(filename, "wb");
  unsigned int size = 0;
  for (unsigned int i = 0; i < kpts.size(); i++)
    size += kpts[i].size();
 
  fwrite(&size, sizeof(unsigned int), 1, file);

  for (unsigned int i = 0; i < kpts.size(); i++)
    for (unsigned int j = 0; j < kpts[i].size(); j++)    
  {
    PSKpt *key;
    PSKpt k;
    if (xform) {      
      k.location = xform->inverse().map_point(kpts[i][j]->location);
      k.direction = xform->inverse().map_direction(kpts[i][j]->direction);
      k.normal = xform->inverse().map_direction(kpts[i][j]->normal);
      k.scale = kpts[i][j]->scale;
      k.feature_vector = kpts[i][j]->feature_vector;
      key = &k;
    }
    else 
      key = kpts[i][j];

    WriteKeypoint(file, key);
  }

  fclose(file);
  vcl_cout << "Wrote " << size << " keypoints to " << filename << vcl_endl;
}

bool operator<(const match &l, const match &r)
{
  return l.ratio < r.ratio;
}