#include "mesh.h"
#include <vcl_iostream.h>
#include <vnl/vnl_math.h>
#include <vcl_map.h>

#include "vtkSmartPointer.h"
#include "vtkPolyData.h"
#include "vtkPointData.h"
#include "vtkXMLPolyDataWriter.h"
#include "vtkFloatArray.h"
#include "vtkCellArray.h"

//*****************************************************************************

Mesh::Mesh(MeshRangeData *rangedata) : rd(rangedata)
{
  ComputeThreshold();
  Triangulate();
  RemoveUnstableTriangles();
  ConsolidatePoints();
}

//*****************************************************************************

Mesh::~Mesh()
{
  delete rd;
}

//******************************************************************************

void Mesh::ComputeThreshold()
{
  const vbl_array_2d<int> &grid = rd->get_grid();
  const vcl_vector<vnl_double_3> &points = rd->get_points();
  int gridWidth = rd->getGridWidth()-1;
  int gridHeight = rd->getGridHeight()-1;

  double M = 0.0; int count = 0;
  for (int i = 1; i < gridHeight; i++)
  {
    for (int j = 1; j < gridWidth; j++)
    {
      //Check if index is a non-boundary point
      bool broke = false;
      for (int k = i - 1; k <= i + 1; k++)        {
        for (int m = j - 1; m <= j + 1; m++)          {
          if (grid[k][m] < 0) {
            broke = true;
            goto breakloop;
          }
        }
      } 

      breakloop:;
      if (broke) continue;

      const vnl_double_3 &a = points[grid[i][j]];       // c  b  o
      const vnl_double_3 &b = points[grid[i-1][j]];     // d  a  o
      const vnl_double_3 &c = points[grid[i-1][j-1]];   // o  o  o
      const vnl_double_3 &d = points[grid[i][j-1]];

      if ((a - c).two_norm() <  (b - d).two_norm())
      {
        double L1 = ConsiderTriangle(d, a, c);   // c\ b
        double L2 = ConsiderTriangle(a, b, c);   // d \a
        if (L1 > 0.0) {M += L1; count++;}
        if (L2 > 0.0) {M += L2; count++;}
      }
      else
      {
        double L1 = ConsiderTriangle(d, b, c);   // c /b
        double L2 = ConsiderTriangle(a, b, d);   // d/ a
        if (L1 > 0.0) {M += L1; count++;}
        if (L2 > 0.0) {M += L2; count++;}
      }         
    }
  }

  D = 1.4 * M / count;
  vcl_cout << "Distance threshold: " << D << vcl_endl;
}

//*****************************************************************************

double Mesh::ConsiderTriangle(const vnl_double_3 &a, const vnl_double_3 &b, const vnl_double_3 &c)
{
  vnl_double_3 n = vnl_cross_3d(b-a,c-a).normalize();
  vnl_double_3 los = ((a + b + c)/3.0).normalize();
  double angle = acos(dot_product(n, los));
  if (angle >= (8.0/9.0)*vnl_math::pi && angle < vnl_math::pi)
  {
    double ab = (a - b).one_norm(), bc = (b - c).one_norm(), ac = (a - c).one_norm();
    double longest = ab > bc ? ab : bc;
    return sqrt(longest > ac ? longest : ac);
  }

  return -1.0;
}

//*****************************************************************************

void Mesh::Triangulate()
{
  const vbl_array_2d<int> &grid = rd->get_grid();
  const vcl_vector<vnl_double_3> &points = rd->get_points();
  int gridWidth = rd->getGridWidth()-1;
  int gridHeight = rd->getGridHeight()-1;

  for (int i = 1; i < gridHeight; i++)
  {
    for (int j = 1; j < gridWidth; j++)
    {
      vcl_vector<int> pts;

      //Check if index is a non-boundary point
      if (grid[i][j] >= 0) pts.push_back(grid[i][j]);          // c  b  o
      if (grid[i-1][j] >= 0) pts.push_back(grid[i-1][j]);      // d  a  o
      if (grid[i-1][j-1] >= 0) pts.push_back(grid[i-1][j-1]);  // o  o  o
      if (grid[i][j-1] >= 0) pts.push_back(grid[i][j-1]);

      if (pts.size() <= 2)
        continue;

      if (pts.size() == 3)
      { 
        Triangle t(pts[0], pts[1], pts[2]);
        CheckTriangle(t);
      }
      else
      {
        if ((points[pts[0]] - points[pts[2]]).two_norm() <  (points[pts[1]] - points[pts[3]]).two_norm())
        {
          Triangle t1(pts[0], pts[1], pts[2]);
          Triangle t2(pts[3], pts[0], pts[2]);
          CheckTriangle(t1);  CheckTriangle(t2);
        }
        else
        {
          Triangle t1(pts[3], pts[1], pts[2]);
          Triangle t2(pts[0], pts[1], pts[3]);
          CheckTriangle(t1);  CheckTriangle(t2);
        }        
      }
    }
  }

  vcl_cout << "Created " << triangles.size() << " triangles." << vcl_endl;
}

//*****************************************************************************

bool Mesh::CheckTriangle(Triangle &t)
{
  const vcl_vector<vnl_double_3> &points = rd->get_points();
  const vnl_double_3 &a = points[t.tri[0]], &b = points[t.tri[1]], &c = points[t.tri[2]];
  if ((a - b).two_norm() > D || (a - c).two_norm() > D || (b - c).two_norm() > D) 
    return false;  
      
  vnl_double_3 n = vnl_cross_3d(b-a, c-a).normalize();
  vnl_double_3 los = ((a + b + c)/3.0).normalize();

  if (dot_product(n, los) > 0)
  {
   int temp = t.tri[0];
   t.tri[0] = t.tri[2];
   t.tri[2] = temp;
  }

  triangles.push_back(t);

  return true;
}

////*****************************************************************************

void Mesh::RemoveUnstableTriangles()
{
  vcl_vector<unsigned char> ptincl(rd->get_points().size(), 0);
  for (vcl_list<Triangle>::const_iterator i = triangles.begin(); i != triangles.end(); i++)
  {
    for (unsigned int j = 0; j < 3; j++)    
      ptincl[i->tri[j]]++;    
  }

  bool done = false;  int removed = 0;
  while (!done)
  {
    done = true;   
    for (vcl_list<Triangle>::iterator i = triangles.begin(); i != triangles.end(); )
    {
      bool kept = true;
      for (unsigned int j = 0; j < 3; j++)
      {
        if (ptincl[i->tri[j]] == 1)
        {
          for (unsigned int k = 0; k < 3; k++)        
            ptincl[i->tri[k]]--;
          i = triangles.erase(i);
          kept = done = false;
          removed++;
          break;
        }
      }  
      if (kept)
       i++;
    }
  }

  vcl_cout << "Removed " << removed << " triangles." << vcl_endl;

}

//*****************************************************************************

void Mesh::ConsolidatePoints()
{
  vcl_vector<vnl_double_3> &points = rd->get_points();
  vcl_vector<vnl_vector_fixed<unsigned char, 3> > &colors = rd->get_colors();

  int oldsize = (int)points.size();
  vcl_vector<unsigned char> ptincl(oldsize, 0);
  for (vcl_list<Triangle>::const_iterator i = triangles.begin(); i != triangles.end(); i++)
  {
    for (unsigned int j = 0; j < 3; j++)    
      ptincl[i->tri[j]]++;    
  }

  vcl_vector<vnl_double_3> pts;
  vcl_vector<vnl_double_3> norms;
  vcl_vector<vnl_vector_fixed<unsigned char, 3> > clrs;
  vcl_vector<int> mapping(oldsize,-1);
  for (int i = oldsize-1; i >= 0; i--)
  {
    if (ptincl[i] > 0)
    {
      pts.push_back(points[i]);
      clrs.push_back(colors[i]);
      mapping[i] = (int)pts.size()-1;
    }
    points.pop_back();    
    colors.pop_back();
  }

  points = pts;
  colors = clrs;

  for (vcl_list<Triangle>::iterator i = triangles.begin(); i != triangles.end(); i++)
  {
    for (unsigned int j = 0; j < 3; j++)  
    {
      assert(mapping[i->tri[j]] != -1);
      i->tri[j] = mapping[i->tri[j]];          
    }
  }

  //grid is now invalid so clear it
  rd->get_grid().clear(); 

  vcl_cout << "Consolidation removed " << oldsize - points.size() << " points.\n";
}

//*****************************************************************************

void Mesh::WriteToVTP(const char *filename)
{
  vtkSmartPointer<vtkPoints> pts = vtkSmartPointer<vtkPoints>::New(); 
  vtkSmartPointer<vtkCellArray> tris = vtkSmartPointer<vtkCellArray>::New();
  vtkSmartPointer<vtkUnsignedCharArray> crs = vtkUnsignedCharArray::New();

  crs->SetName("Colors");
  crs->SetNumberOfComponents(3);

  const vcl_vector<vnl_double_3> &points = rd->get_points();
  const vcl_vector<vnl_vector_fixed<unsigned char, 3> > &colors = rd->get_colors();

  for (unsigned int i = 0; i < points.size(); i++)  
    pts->InsertNextPoint(points[i].data_block());  

  vtkIdType triangle[3];
  for (vcl_list<Triangle>::const_iterator i = triangles.begin(); i != triangles.end(); i++)  {
  triangle[0] = i->tri[0];
  triangle[1] = i->tri[1];
  triangle[2] = i->tri[2];
    tris->InsertNextCell(3, triangle);    
  }

  for (unsigned int i = 0; i < colors.size(); i++)  
    crs->InsertNextTuple3(colors[i][0], colors[i][1], colors[i][2]);  

  vtkSmartPointer<vtkPolyData> polydata = vtkSmartPointer<vtkPolyData>::New();

  polydata->SetPoints(pts);
  polydata->SetPolys(tris);
  polydata->GetPointData()->SetVectors(crs);
  vtkSmartPointer<vtkXMLPolyDataWriter> writer = vtkSmartPointer<vtkXMLPolyDataWriter>::New();
  writer->SetInput(polydata);
  writer->SetFileName(filename);
  writer->Write(); 

  vcl_cout << "Wrote " << filename << vcl_endl;
}

//*****************************************************************************