#include "mesh_range_data.h"

#include <vcl_string.h>
#include <vcl_iostream.h>
#include <vnl/vnl_math.h>

#include <vul/vul_file_iterator.h>
#include <vul/vul_file.h>

#include "vtkPoints.h"
#include "vtkPolyData.h"
#include "vtkPointData.h"
#include "vtkSmartPointer.h"
#include "vtkXMLPolyDataWriter.h"

#include "utilities.h"

//*****************************************************************************

MeshRangeData::MeshRangeData() : gridWidth(0), gridHeight(0)                         
{

} 

//*****************************************************************************

MeshRangeData::~MeshRangeData()
{

}

//*****************************************************************************

bool MeshRangeData::ReadInScanInfo(const char *directory, const char *scan_name)
{
  // Construct globbing expression.
  if(!directory) 
    directory = "."; 

  vcl_string dir = vcl_string(directory);

  vcl_string expr;
  if (scan_name == NULL)
    expr = dir + "/*.ptx";
  else
    expr = dir + "/" + scan_name;

  vul_file_iterator filei = expr;
  
  if (filei)  
    this->processData(filei());      
    
  return true;
}

//*****************************************************************************

bool MeshRangeData::processData(const char *filename)
{  
  vcl_ifstream infile(filename);
  if (!infile) return false;
  vcl_string line;
  // Read grid width and height.
  int tempwidth, tempheight;
  if(!(vcl_getline(infile, line) && sscanf(line.c_str(), "%d", &tempwidth) == 1))
    return false; 
  if(!(vcl_getline(infile, line) && sscanf(line.c_str(), "%d", &tempheight) == 1))
    return false;
  this->gridWidth = tempwidth;   this->gridHeight = tempheight;

  //Skip transform lines.  
  for(int i = 0; i < 8; i++)    
    vcl_getline(infile, line);

  //Read points.  Scan points are organized with bottom-to-top index
  //varying fastest, and then left-to-right.  Fill in point grid for normals/scales calc  
  //Also calculate an approximate spherical grid for the points to fit into
  vbl_array_2d< vnl_vector_fixed<unsigned char, 3>  > color_grid(this->gridHeight, this->gridWidth);
  for(int i = 0; i < this->gridWidth; ++i)
  {
    for(int j = this->gridHeight - 1; j >= 0; --j)
    {
      // Read this point.
      double tmp;
      vnl_double_3 point; vnl_vector_fixed<unsigned int, 3>  c;
      if(!(vcl_getline(infile, line) && sscanf(line.c_str(), "%lg %lg %lg %lg %d %d %d", &point[0], &point[1], &point[2], &tmp, &c[0], &c[1], &c[2]) == 7))
        return false;
      
      vnl_vector_fixed<unsigned char, 3>  color((unsigned char)c[0], (unsigned char)c[1], (unsigned char)c[2]);
      color_grid.put(j, i, color);
    }
  }

  grid = vbl_array_2d< int >(this->gridHeight, this->gridWidth);

  vbl_array_2d< PointRecord > point_grid;
  load_point_grid(filename, point_grid);

  numpoints = 0;
  for (int i = 0; i < this->gridHeight; i++)
  {
    for (int j = 0; j < this->gridWidth; j++)
    {    
      const PointRecord &pr = point_grid(i,j);
      vnl_double_3 pt = pr.estimated_distance * pr.direction;
      if (pr.is_valid_return && pt.magnitude() >= 1.0)     
      {
        normals.push_back(pr.surface_normal);
        points.push_back(pt);
        colors.push_back(color_grid(i, j));
        grid[i][j] = numpoints;
        numpoints++;
      }
      else
        grid[i][j] = -1;
    }
  }

  return true;
}

//*****************************************************************************

void MeshRangeData::load_point_grid(const char *filename, vbl_array_2d<PointRecord> &point_grid)
{
  vcl_string pgname(filename);
  pgname.erase(pgname.size()-3, 3);
  pgname += "pg";
  
  vcl_cout << "Loading point grid.\n";
  FILE *file = fopen(pgname.c_str(), "rb");

  int rows, cols;
  fread(&rows, sizeof(int), 1, file);
  fread(&cols, sizeof(int), 1, file);

  double std_dev;
  fread(&std_dev, sizeof(double), 1, file);

  point_grid.resize(rows, cols);
  for (int r = 0; r < rows; r++)
  {
    for (int c = 0; c < cols; c++)
    {
      PointRecord &pr = point_grid[r][c];
      fread(&pr.is_valid_return, sizeof(bool), 1, file);
      if (!pr.is_valid_return)
        continue;
      fread(&pr.measured_distance, sizeof(double), 1, file);
      fread(pr.direction.data_block(), sizeof(double)*3, 1, file);
      fread(&pr.is_initialized, sizeof(bool), 1, file);
      fread(&pr.is_done, sizeof(bool), 1, file);
      fread(&pr.point_spacing_normal_to_los, sizeof(double), 1, file);
      fread(&pr.priority, sizeof(double), 1, file);
      fread(&pr.pq_index, sizeof(int), 1, file);
      fread(&pr.done_index, sizeof(int), 1, file);
      fread(&pr.surface_type, sizeof(PointRecord::SurfaceType), 1, file);
      fread(&pr.estimated_scale, sizeof(double), 1, file);
      fread(&pr.estimated_angle_scale, sizeof(double), 1, file);
      fread(&pr.estimated_distance, sizeof(double), 1, file);
      fread(pr.surface_normal.data_block(), sizeof(double)*3, 1, file);
      for (int i = 0; i < 4; i++)
      {
        fread(pr.corner_directions[i].data_block(), sizeof(double)*3, 1, file);
        fread(&pr.corner_distances, sizeof(double), 1, file);
      }
    }
  }

  fclose(file);
}

//*****************************************************************************

void MeshRangeData::writePolyData(const char *filename) const
{
  vtkSmartPointer<vtkPoints> pts = vtkSmartPointer<vtkPoints>::New(); 

  for (unsigned int i = 0; i < points.size(); i++)  
    pts->InsertNextPoint(points[i].data_block());
  
  vtkSmartPointer<vtkPolyData> pdata = vtkSmartPointer<vtkPolyData>::New();
  pdata->SetPoints(pts);

  vtkSmartPointer<vtkXMLPolyDataWriter> writer = vtkSmartPointer<vtkXMLPolyDataWriter>::New();
  writer->SetInput(pdata);
  writer->SetFileName(filename);
  writer->Write();  
}

//*****************************************************************************
